//    NEO Translator 1.0 for R4
//    Copyright (C) The Bitstreamers, 1999
//

#include <AppKit.h>
#include <TranslationKit.h>
#include <translation/TranslatorAddOn.h>
#include <stdio.h>
#include <Screen.h>


//////////////////////////////////////////////////////////////////////////////
// Misc. definitions
//////////////////////////////////////////////////////////////////////////////

// The following constants are not part of the NEO file format,
// this represents how this program categorizes the NEO files it can read.
const int NEO_FLAG_SIZE = 2;
const int NEO_RESOLUTION_SIZE = 2;
const int NEO_PALETTE_SIZE = 16;
const int NEO_PALETTE_DOUBLE_SIZE = 32;

const int NEO_READ_HALF_BUFFER_SIZE = 16000;
const int NEO_READ_BUFFER_SIZE = 32000;
const int NEO_BUFFER_SIZE = 64000;
const int NEO_QUAD_BUFFER_SIZE = 256000;

const int NEO_TYPE = 'NEO';
#define NEO_MIME "image/x-neo"
#define NEO_NAME "NEO bitmap format"

const int NEO_SIZE_FORMAT = 32128;
const float NEO_INPUT_QUALITY = 0.3; 
const float NEO_INPUT_CAPABILITY = 0.6;

//////////////////////////////////////////////////////////////////////////////
// Prototypes
//////////////////////////////////////////////////////////////////////////////

status_t DecodeNeo( BPositionIO *inSource, BPositionIO *outDestination);

//////////////////////////////////////////////////////////////////////////////
// main() does nothing
//////////////////////////////////////////////////////////////////////////////

int main( int argc, char *argv[] )
{
        return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Exported data, used by the translation kit
//////////////////////////////////////////////////////////////////////////////

char translatorName[] = "NEOTranslator";
char translatorInfo[] = "NEOTranslator 1.0 Copyright (C) The Bitstreamers, 1999";
int32 translatorVersion = 100;  //  Version 1.00

translation_format inputFormats[] =
{
    { NEO_TYPE, B_TRANSLATOR_BITMAP, NEO_INPUT_QUALITY, NEO_INPUT_CAPABILITY, NEO_MIME, NEO_NAME },
    { 0, 0, 0, 0, {0}, {0} } 
};

translation_format outputFormats[] =
{
    { B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/x-be-bitmap", "Be Bitmap format" },
        { 0, 0, 0, 0, {0}, {0} } 
};

//////////////////////////////////////////////////////////////////////////////
// Translation functions
//////////////////////////////////////////////////////////////////////////////

// Called to identify the type of a file if supported
// Identification should be as accurate as possible especially to determine
// whether we'll be able to translate it.

status_t Identify( BPositionIO * inSource,
                                const translation_format * inFormat,
                                BMessage * ioExtension,
                                translator_info * outInfo,
                                uint32 outType)
{
        /////////// VERIFY IF OUTPUT TYPE IS SUPPORTED //////////////
        
        if( outType != 0 && outType != B_TRANSLATOR_BITMAP )
                return B_NO_TRANSLATOR; 

        /////////// CHECK IF NEO FILE FLAG IS VALID //////////////

        int8 neoFlagType[NEO_FLAG_SIZE];
		int8 neoResolution[NEO_RESOLUTION_SIZE];
		
        if( NEO_FLAG_SIZE != inSource->Read( neoFlagType, NEO_FLAG_SIZE ) )
                return B_NO_TRANSLATOR; 

		if ( NEO_RESOLUTION_SIZE != inSource->Read( neoResolution, NEO_RESOLUTION_SIZE ) )
				return B_NO_TRANSLATOR;
				
        /////////// GIVE INFO ABOUT THE TRANSLATION //////////////

		inSource->Seek(0, SEEK_SET);
		uchar *file_buffer = new uchar[40000];
	    ssize_t file_length = inSource->Read(file_buffer, 40000);
	    delete[] file_buffer;
		
		if (file_length == NEO_SIZE_FORMAT)
		{	
          outInfo->type = NEO_TYPE;
          outInfo->group = B_TRANSLATOR_BITMAP; 
          outInfo->quality = NEO_INPUT_QUALITY; 
          outInfo->capability = NEO_INPUT_CAPABILITY; 
          strcpy( outInfo->name, NEO_NAME ); 
          strcpy( outInfo->MIME, NEO_MIME ); 
     
          return B_OK;
        }
        
        return B_NO_TRANSLATOR;
}

//
// Performs the actual translation work.
//

status_t Translate( BPositionIO * inSource, const translator_info * inInfo,
                                    BMessage * ioExtension, uint32 outType, BPositionIO * outDestination )
{
        /////////// VERIFY IF OUTPUT TYPE IS SUPPORTED //////////////
 
    status_t ret_val = B_OK;

	//	Check that we handle input and output types

	if (outType == 0)
		outType = B_TRANSLATOR_BITMAP;

	if (outType != B_TRANSLATOR_BITMAP)
		ret_val = B_NO_TRANSLATOR;
        
        /////////// CHECK IF NEOs FILE HEADER IS VALID //////////////

        int8 neoResolution[NEO_RESOLUTION_SIZE];
        int16 iNEO_X_Resolution = 0;
        int16 iNEO_Y_Resolution = 0;
        float fNEO_X_Resolution = 0.0;
        float fNEO_Y_Resolution = 0.0;

		// Check voor the right size of the image
		inSource->Seek(2, SEEK_SET);
        if( NEO_RESOLUTION_SIZE != inSource->Read( neoResolution, NEO_RESOLUTION_SIZE ) )
           ret_val = B_NO_TRANSLATOR; 

        // Tests if the file is a valid NEO file and determines its resolution
        if( neoResolution[0] == 0 && neoResolution[1] == 0 )
        {
                iNEO_X_Resolution = 320; 
                iNEO_Y_Resolution = 200;
                fNEO_X_Resolution = 319.0; 
                fNEO_Y_Resolution = 199.0; 
        }
        else
                ret_val = B_NO_TRANSLATOR;


        /////////// WRITE B_TRANSLATOR_BITMAP HEADER DATA //////////////
        
        TranslatorBitmap bitmapHeader;
        bitmapHeader.magic = B_TRANSLATOR_BITMAP;
        bitmapHeader.colors = B_RGBA32;
        bitmapHeader.bounds = BRect(0.0, 0.0, fNEO_X_Resolution, fNEO_Y_Resolution);
        bitmapHeader.rowBytes = iNEO_X_Resolution * 4;
        bitmapHeader.dataSize = (iNEO_X_Resolution * 4) * iNEO_Y_Resolution;

        /*
         * B_TRANSLATOR_BITMAP is always in BIG ENDIAN format.
         * Convert the header to the correct endianess.
         */
		swap_data(B_UINT32_TYPE, &(bitmapHeader.magic), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_UINT32_TYPE, &(bitmapHeader.colors), sizeof(color_space), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_RECT_TYPE, &(bitmapHeader.bounds), sizeof(BRect), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_UINT32_TYPE, &(bitmapHeader.rowBytes), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_UINT32_TYPE, &(bitmapHeader.dataSize), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);

        if (outDestination->Write( &bitmapHeader, sizeof(TranslatorBitmap) ) != sizeof(TranslatorBitmap))
        {
          ret_val = B_NO_TRANSLATOR;
        } else
          {
             ret_val = DecodeNeo( inSource, outDestination);
          }
        
    return ret_val;
}

status_t DecodeNeo( BPositionIO *inSource, BPositionIO *outDestination)
{
        /////////// READ AND CONVERT NEO PALETTE //////////////

		status_t return_value = B_OK;
        
        // reads the pcx palette
        uint16 paletteBuffer[NEO_PALETTE_SIZE];
        uint32 Red, Green, Blue;
        
        inSource->Seek(4, SEEK_SET);
        inSource->Read(&paletteBuffer, NEO_PALETTE_DOUBLE_SIZE );

        // converts the palette into big endian B_RGB32 format
        uint32 paletteRGB[NEO_PALETTE_SIZE];
       
        for( int i=0; i < NEO_PALETTE_SIZE; i++)
        {
           Red   = (paletteBuffer[i] & 0xF) << 5;
           Green = (paletteBuffer[i] & 0xF000) >> 7; 
           Blue  = (paletteBuffer[i] & 0xF00) >> 3;
     
           paletteRGB[i] = (0xFF << 24) | (Red << 16) | (Green << 8) | Blue;
         }
  
        /////////// READ AND CONVERT NEO IMAGE //////////////

        uint16 *inBuffer     = new uint16[NEO_READ_HALF_BUFFER_SIZE];
        uint16 *middleBuffer = new uint16[NEO_BUFFER_SIZE]; 
        uint32 *outBuffer    = new uint32[NEO_BUFFER_SIZE];
       
        // Seek to the beginning of data & Reads a first block of the data
        inSource->Seek( 128, SEEK_SET);
        inSource->Read(inBuffer, NEO_READ_BUFFER_SIZE );
        uint16 Atari_Plane_0, Atari_Plane_1, Atari_Plane_2, Atari_Plane_3;
        
        for( int i = 0, j = 0; i < NEO_READ_BUFFER_SIZE; i+=4, j++ )
        {
			Atari_Plane_0 = ((inBuffer[i] & 0xff00) >> 8)   | ((inBuffer[i] & 0x0ff) << 8);
		    Atari_Plane_1 = ((inBuffer[i+1] & 0xff00) >> 8) | ((inBuffer[i+1] & 0x0ff) << 8);
		    Atari_Plane_2 = ((inBuffer[i+2] & 0xff00) >> 8) | ((inBuffer[i+2] & 0x0ff) << 8);
		    Atari_Plane_3 = ((inBuffer[i+3] & 0xff00) >> 8) | ((inBuffer[i+3] & 0x0ff) << 8); 

			// Pixel_0
			middleBuffer[j++] = ((Atari_Plane_3 >> 12) & 8) | ((Atari_Plane_2 >> 13) & 4) |
		    					((Atari_Plane_1 >> 14) & 2) | ((Atari_Plane_0 >> 15) & 1);
				
			// Pixel_1
			middleBuffer[j++] = ((Atari_Plane_3 >> 11) & 8) | ((Atari_Plane_2 >> 12) & 4) |
			           			((Atari_Plane_1 >> 13) & 2) | ((Atari_Plane_0 >> 14) & 1);
		
			//Pixel_2
			middleBuffer[j++] = ((Atari_Plane_3 >> 10) & 8) | ((Atari_Plane_2 >> 11) & 4) |
			           			((Atari_Plane_1 >> 12) & 2) | ((Atari_Plane_0 >> 13) & 1);

			//Pixel_3
			middleBuffer[j++] = ((Atari_Plane_3 >>  9) & 8) | ((Atari_Plane_2 >> 10) & 4) |
			           			((Atari_Plane_1 >> 11) & 2) | ((Atari_Plane_0 >> 12) & 1);
		
			//Pixel_4
			middleBuffer[j++] = ((Atari_Plane_3 >>  8) & 8) | ((Atari_Plane_2 >> 9) & 4) |
			           			((Atari_Plane_1 >> 10) & 2) | ((Atari_Plane_0 >> 11) & 1);
		
			//Pixel_5
			middleBuffer[j++] = ((Atari_Plane_3 >>  7) & 8) | ((Atari_Plane_2 >> 8) & 4) |
			           			((Atari_Plane_1 >>  9) & 2) | ((Atari_Plane_0 >> 10) & 1);
		
			//Pixel_6
			middleBuffer[j++] = ((Atari_Plane_3 >>  6) & 8) | ((Atari_Plane_2 >> 7) & 4) |
			           			((Atari_Plane_1 >>  8) & 2) | ((Atari_Plane_0 >> 9) & 1);
		
			//Pixel_7
			middleBuffer[j++] = ((Atari_Plane_3 >>  5) & 8) | ((Atari_Plane_2 >> 6) & 4) |
			           			((Atari_Plane_1 >>  7) & 2) | ((Atari_Plane_0 >> 8) & 1);
		
			//Pixel_8
			middleBuffer[j++] = ((Atari_Plane_3 >>  4) & 8) | ((Atari_Plane_2 >> 5) & 4) |
			           			((Atari_Plane_1 >>  6) & 2) | ((Atari_Plane_0 >> 7) & 1);
		
			//Pixel_9
			middleBuffer[j++] = ((Atari_Plane_3 >>  3) & 8) | ((Atari_Plane_2 >> 4) & 4) |
			           			((Atari_Plane_1 >>  5) & 2) | ((Atari_Plane_0 >> 6) & 1);
		
			//Pixel_10
			middleBuffer[j++] = ((Atari_Plane_3 >>  2) & 8) | ((Atari_Plane_2 >> 3) & 4) |
			           			((Atari_Plane_1 >>  4) & 2) | ((Atari_Plane_0 >> 5) & 1);
		
			//Pixel_11
			middleBuffer[j++] = ((Atari_Plane_3 >>  1) & 8) | ((Atari_Plane_2 >> 2) & 4) |
			           			((Atari_Plane_1 >>  3) & 2) | ((Atari_Plane_0 >> 4) & 1);
	
			//Pixel_12
			middleBuffer[j++] = ((Atari_Plane_3 >>  0) & 8) | ((Atari_Plane_2 >> 1) & 4) |
			           			((Atari_Plane_1 >>  2) & 2) | ((Atari_Plane_0 >> 3) & 1);
		
			//Pixel_13
			middleBuffer[j++] = ((Atari_Plane_3 <<  1) & 8) | ((Atari_Plane_2 >> 0) & 4) |
			           			((Atari_Plane_1 >>  1) & 2) | ((Atari_Plane_0 >> 2) & 1);
		
			//Pixel_14
			middleBuffer[j++] = ((Atari_Plane_3 <<  2) & 8) | ((Atari_Plane_2 << 1) & 4) |
			           			((Atari_Plane_1 >>  0) & 2) | ((Atari_Plane_0 >> 1) & 1);
		
			//Pixel_15
			middleBuffer[j] = ((Atari_Plane_3 <<  3) & 8) | ((Atari_Plane_2 << 2) & 4) |
			           		  ((Atari_Plane_1 <<  1) & 2) | ((Atari_Plane_0 >> 0) & 1);
        }

        for ( int i = 0; i < NEO_BUFFER_SIZE; i++)
        {
           outBuffer[i] = (paletteRGB[middleBuffer[i]]);
        }

		// write the bitmap         
        int ret = outDestination->Write(outBuffer, NEO_QUAD_BUFFER_SIZE);
        if (ret != NEO_QUAD_BUFFER_SIZE)
		{
		  return_value = B_NO_TRANSLATOR;
		} else
		  {
		    return_value = B_OK;
          }
           
        // clean-up
        delete inBuffer;
        delete middleBuffer;
        delete outBuffer;
        
        return return_value;
}
