//    PI1 Translator 1.0 for R4
//    Copyright (C) The Bitstreamers, 1999
//

#include <AppKit.h>
#include <TranslationKit.h>
#include <translation/TranslatorAddOn.h>
#include <stdio.h>
#include <Screen.h>

//////////////////////////////////////////////////////////////////////////////
// Misc. definitions
//////////////////////////////////////////////////////////////////////////////

// The following constants are not part of the PCX file format,
// this represents how this program categorizes the PCX files it can read.
typedef enum { PI1_16_COLORS, PI1_4_COLORS, PI1_2_COLORS } PI1_Type;

const int PI1_HEADER_SIZE = 2;
const int PI1_PALETTE_SIZE = 16;
const int PI1_PALETTE_DOUBLE_SIZE = 32;

const int PI1_READ_HALF_BUFFER_SIZE = 16000;
const int PI1_READ_BUFFER_SIZE = 32000;
const int PI1_BUFFER_SIZE = 64000;
const int PI1_QUAD_BUFFER_SIZE = 256000;

const int PI1_TYPE = 'PI1';
#define PI1_MIME "image/x-pi1"
#define PI1_NAME "PI1 bitmap format"

const int PI1_OLD_FORMAT = 32034;
const int PI1_NEW_FORMAT = 32066;
const float PI1_INPUT_QUALITY = 0.3; 
const float PI1_INPUT_CAPABILITY = 0.6;

//////////////////////////////////////////////////////////////////////////////
// Prototypes
//////////////////////////////////////////////////////////////////////////////

status_t Decode16( BPositionIO *inSource, BPositionIO *outDestination);

//////////////////////////////////////////////////////////////////////////////
// main() does nothing
//////////////////////////////////////////////////////////////////////////////

int main( int argc, char *argv[] )
{
        return 0;
}

//////////////////////////////////////////////////////////////////////////////
// Exported data, used by the translation kit
//////////////////////////////////////////////////////////////////////////////

char translatorName[] = "PI1Translator";
char translatorInfo[] = "PI1Translator 1.0 Copyright (C) The Bitstreamers, 1999";
int32 translatorVersion = 100;  //  Version 1.00

translation_format inputFormats[] =
{
    { PI1_TYPE, B_TRANSLATOR_BITMAP, PI1_INPUT_QUALITY, PI1_INPUT_CAPABILITY, PI1_MIME, PI1_NAME },
    { 0, 0, 0, 0, {0}, {0} } 
};

translation_format outputFormats[] =
{
    { B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/x-be-bitmap", "Be Bitmap format" },
        { 0, 0, 0, 0, {0}, {0} } 
};

//////////////////////////////////////////////////////////////////////////////
// Translation functions
//////////////////////////////////////////////////////////////////////////////

// Called to identify the type of a file if supported
// Identification should be as accurate as possible especially to determine
// whether we'll be able to translate it.

status_t Identify( BPositionIO * inSource,
                                const translation_format * inFormat,
                                BMessage * ioExtension,
                                translator_info * outInfo,
                                uint32 outType)
{
        /////////// VERIFY IF OUTPUT TYPE IS SUPPORTED //////////////
        
        if( outType != 0 && outType != B_TRANSLATOR_BITMAP )
                return B_NO_TRANSLATOR; 

        /////////// CHECK IF PI1 FILE HEADER IS VALID //////////////

        int8 pi1Header[PI1_HEADER_SIZE];

        if( PI1_HEADER_SIZE != inSource->Read( pi1Header, PI1_HEADER_SIZE ) )
                return B_NO_TRANSLATOR; 

        PI1_Type type;
        // Tests if the file is a valid PI1 file and determines its type
        if( pi1Header[0] == 0 && pi1Header[1] == 0 )
                type = PI1_16_COLORS;
        else
                return B_NO_TRANSLATOR;

        /////////// GIVE INFO ABOUT THE TRANSLATION //////////////

		inSource->Seek(0, SEEK_SET);
		uchar *file_buffer = new uchar[40000];
	    ssize_t file_length = inSource->Read(file_buffer, 40000);
	    delete[] file_buffer;
		
		if ((file_length == PI1_OLD_FORMAT) || (file_length == PI1_NEW_FORMAT))
		{	
          outInfo->type = PI1_TYPE;
          outInfo->group = B_TRANSLATOR_BITMAP; 
          outInfo->quality = PI1_INPUT_QUALITY; 
          outInfo->capability = PI1_INPUT_CAPABILITY; 
          strcpy( outInfo->name, PI1_NAME ); 
          strcpy( outInfo->MIME, PI1_MIME ); 
     
          return B_OK;
        }
        
        return B_NO_TRANSLATOR;
}

//
// Performs the actual translation work.
//

status_t Translate( BPositionIO * inSource, const translator_info * inInfo,
                                    BMessage * ioExtension, uint32 outType, BPositionIO * outDestination )
{
        /////////// VERIFY IF OUTPUT TYPE IS SUPPORTED //////////////
 
    status_t ret_val = B_OK;

	//	Check that we handle input and output types

	if (outType == 0)
		outType = B_TRANSLATOR_BITMAP;

	if (outType != B_TRANSLATOR_BITMAP)
		ret_val = B_NO_TRANSLATOR;
        
        /////////// CHECK IF PCX FILE HEADER IS VALID //////////////

        int8 pi1Header[PI1_HEADER_SIZE];
        int16 iPI1_X_Resolution = 0;
        int16 iPI1_Y_Resolution = 0;
        float fPI1_X_Resolution = 0.0;
        float fPI1_Y_Resolution = 0.0;

		// Check voor the right size of the image
        if( PI1_HEADER_SIZE != inSource->Read( pi1Header, PI1_HEADER_SIZE ) )
           ret_val = B_NO_TRANSLATOR; 

        PI1_Type type = PI1_16_COLORS;

        // Tests if the file is a valid PCX file and determines its type
        if( pi1Header[0] == 0 && pi1Header[1] == 0 )
        {
                type = PI1_16_COLORS;
                iPI1_X_Resolution = 320; 
                iPI1_Y_Resolution = 200;
                fPI1_X_Resolution = 319.0; 
                fPI1_Y_Resolution = 199.0; 
        }
        else
                ret_val = B_NO_TRANSLATOR;


        /////////// WRITE B_TRANSLATOR_BITMAP HEADER DATA //////////////
        
        TranslatorBitmap bitmapHeader;
        bitmapHeader.magic = B_TRANSLATOR_BITMAP;
        bitmapHeader.colors = B_RGBA32;
        bitmapHeader.bounds = BRect(0.0, 0.0, fPI1_X_Resolution, fPI1_Y_Resolution);
        bitmapHeader.rowBytes = iPI1_X_Resolution * 4;
        bitmapHeader.dataSize = (iPI1_X_Resolution * 4) * iPI1_Y_Resolution;

        /*
         * B_TRANSLATOR_BITMAP is always in BIG ENDIAN format.
         * Convert the header to the correct endianess.
         */
		swap_data(B_UINT32_TYPE, &(bitmapHeader.magic), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_UINT32_TYPE, &(bitmapHeader.colors), sizeof(color_space), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_RECT_TYPE, &(bitmapHeader.bounds), sizeof(BRect), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_UINT32_TYPE, &(bitmapHeader.rowBytes), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
		swap_data(B_UINT32_TYPE, &(bitmapHeader.dataSize), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);

        if (outDestination->Write( &bitmapHeader, sizeof(TranslatorBitmap) ) != sizeof(TranslatorBitmap))
        {
          ret_val = B_NO_TRANSLATOR;
        }

        switch( type )
        {
                case PI1_16_COLORS:
                {
                  ret_val = Decode16( inSource, outDestination);
                  break;
                }
                default:
                    ret_val = B_NO_TRANSLATOR;
        }
        
    return ret_val;
}

status_t Decode16( BPositionIO *inSource, BPositionIO *outDestination)
{
        /////////// READ AND CONVERT PCX PALETTE //////////////

		status_t return_value = B_OK;
        
        // reads the pcx palette
        uint16 paletteBuffer[PI1_PALETTE_SIZE];
        uint32 Red, Green, Blue;
        
        inSource->Seek(2, SEEK_SET);
        inSource->Read(&paletteBuffer, PI1_PALETTE_DOUBLE_SIZE );

        // converts the palette into big endian B_RGB32 format
        uint32 paletteRGB[PI1_PALETTE_SIZE];
       
        for( int i=0; i < PI1_PALETTE_SIZE; i++)
        {
           Red   = (paletteBuffer[i] & 0xF) << 5;
           Green = (paletteBuffer[i] & 0xF000) >> 7; 
           Blue  = (paletteBuffer[i] & 0xF00) >> 3;
     
           paletteRGB[i] = (0xFF << 24) | (Red << 16) | (Green << 8) | Blue;
         }
  
        /////////// READ AND CONVERT PCX IMAGE //////////////

        uint16 *inBuffer     = new uint16[PI1_READ_HALF_BUFFER_SIZE];
        uint16 *middleBuffer = new uint16[PI1_BUFFER_SIZE]; 
        uint32 *outBuffer    = new uint32[PI1_BUFFER_SIZE];
       
        // Seek to the beginning of data & Reads a first block of the data
        inSource->Seek( 34, SEEK_SET);
        inSource->Read(inBuffer, PI1_READ_BUFFER_SIZE );
        uint16 Atari_Plane_0, Atari_Plane_1, Atari_Plane_2, Atari_Plane_3;
        
        for( int i = 0, j = 0; i < PI1_READ_BUFFER_SIZE; i+=4, j++ )
        {
			Atari_Plane_0 = ((inBuffer[i] & 0xff00) >> 8)   | ((inBuffer[i] & 0x0ff) << 8);
		    Atari_Plane_1 = ((inBuffer[i+1] & 0xff00) >> 8) | ((inBuffer[i+1] & 0x0ff) << 8);
		    Atari_Plane_2 = ((inBuffer[i+2] & 0xff00) >> 8) | ((inBuffer[i+2] & 0x0ff) << 8);
		    Atari_Plane_3 = ((inBuffer[i+3] & 0xff00) >> 8) | ((inBuffer[i+3] & 0x0ff) << 8); 

			// Pixel_0
			middleBuffer[j++] = ((Atari_Plane_3 >> 12) & 8) | ((Atari_Plane_2 >> 13) & 4) |
		    					((Atari_Plane_1 >> 14) & 2) | ((Atari_Plane_0 >> 15) & 1);
				
			// Pixel_1
			middleBuffer[j++] = ((Atari_Plane_3 >> 11) & 8) | ((Atari_Plane_2 >> 12) & 4) |
			           			((Atari_Plane_1 >> 13) & 2) | ((Atari_Plane_0 >> 14) & 1);
		
			//Pixel_2
			middleBuffer[j++] = ((Atari_Plane_3 >> 10) & 8) | ((Atari_Plane_2 >> 11) & 4) |
			           			((Atari_Plane_1 >> 12) & 2) | ((Atari_Plane_0 >> 13) & 1);

			//Pixel_3
			middleBuffer[j++] = ((Atari_Plane_3 >>  9) & 8) | ((Atari_Plane_2 >> 10) & 4) |
			           			((Atari_Plane_1 >> 11) & 2) | ((Atari_Plane_0 >> 12) & 1);
		
			//Pixel_4
			middleBuffer[j++] = ((Atari_Plane_3 >>  8) & 8) | ((Atari_Plane_2 >> 9) & 4) |
			           			((Atari_Plane_1 >> 10) & 2) | ((Atari_Plane_0 >> 11) & 1);
		
			//Pixel_5
			middleBuffer[j++] = ((Atari_Plane_3 >>  7) & 8) | ((Atari_Plane_2 >> 8) & 4) |
			           			((Atari_Plane_1 >>  9) & 2) | ((Atari_Plane_0 >> 10) & 1);
		
			//Pixel_6
			middleBuffer[j++] = ((Atari_Plane_3 >>  6) & 8) | ((Atari_Plane_2 >> 7) & 4) |
			           			((Atari_Plane_1 >>  8) & 2) | ((Atari_Plane_0 >> 9) & 1);
		
			//Pixel_7
			middleBuffer[j++] = ((Atari_Plane_3 >>  5) & 8) | ((Atari_Plane_2 >> 6) & 4) |
			           			((Atari_Plane_1 >>  7) & 2) | ((Atari_Plane_0 >> 8) & 1);
		
			//Pixel_8
			middleBuffer[j++] = ((Atari_Plane_3 >>  4) & 8) | ((Atari_Plane_2 >> 5) & 4) |
			           			((Atari_Plane_1 >>  6) & 2) | ((Atari_Plane_0 >> 7) & 1);
		
			//Pixel_9
			middleBuffer[j++] = ((Atari_Plane_3 >>  3) & 8) | ((Atari_Plane_2 >> 4) & 4) |
			           			((Atari_Plane_1 >>  5) & 2) | ((Atari_Plane_0 >> 6) & 1);
		
			//Pixel_10
			middleBuffer[j++] = ((Atari_Plane_3 >>  2) & 8) | ((Atari_Plane_2 >> 3) & 4) |
			           			((Atari_Plane_1 >>  4) & 2) | ((Atari_Plane_0 >> 5) & 1);
		
			//Pixel_11
			middleBuffer[j++] = ((Atari_Plane_3 >>  1) & 8) | ((Atari_Plane_2 >> 2) & 4) |
			           			((Atari_Plane_1 >>  3) & 2) | ((Atari_Plane_0 >> 4) & 1);
	
			//Pixel_12
			middleBuffer[j++] = ((Atari_Plane_3 >>  0) & 8) | ((Atari_Plane_2 >> 1) & 4) |
			           			((Atari_Plane_1 >>  2) & 2) | ((Atari_Plane_0 >> 3) & 1);
		
			//Pixel_13
			middleBuffer[j++] = ((Atari_Plane_3 <<  1) & 8) | ((Atari_Plane_2 >> 0) & 4) |
			           			((Atari_Plane_1 >>  1) & 2) | ((Atari_Plane_0 >> 2) & 1);
		
			//Pixel_14
			middleBuffer[j++] = ((Atari_Plane_3 <<  2) & 8) | ((Atari_Plane_2 << 1) & 4) |
			           			((Atari_Plane_1 >>  0) & 2) | ((Atari_Plane_0 >> 1) & 1);
		
			//Pixel_15
			middleBuffer[j] = ((Atari_Plane_3 <<  3) & 8) | ((Atari_Plane_2 << 2) & 4) |
			           		  ((Atari_Plane_1 <<  1) & 2) | ((Atari_Plane_0 >> 0) & 1);
        }

        for ( int i = 0; i < PI1_BUFFER_SIZE; i++)
        {
           outBuffer[i] = (paletteRGB[middleBuffer[i]]);
        }

		// write the bitmap         
        int ret = outDestination->Write(outBuffer, PI1_QUAD_BUFFER_SIZE);
        if (ret != PI1_QUAD_BUFFER_SIZE)
		{
		  return_value = B_NO_TRANSLATOR;
		} else
		  {
		    return_value = B_OK;
          }
           
        // clean-up
        delete inBuffer;
        delete middleBuffer;
        delete outBuffer;
        
        return return_value;
}
