#include <Locker.h>
#include <Path.h>
#include <Entry.h>
#include <FindDirectory.h>
#include <View.h>
#include <Screen.h>
#include <GraphicsDefs.h>
#include <InterfaceDefs.h>
#include <CheckBox.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <MenuField.h>
#include <Message.h>
#include <StringView.h>
#include <Roster.h>
#include <Alert.h>
#include <stdio.h>
#include <Bitmap.h>
#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>

#define BUF_SIZE 4096
#define BMP_CODE 'BMP '
#define MAX_COLORMAP 256
#define BI_RGB       0
#define BI_RLE8      1
#define BI_RLE4      2
#define BMP_INFO_HDR 0
#define BMP_CORE_HDR 1

#if defined (__POWERPC__)
#  define COLOR_MASK	0xFFFFFF00
#  define ALPHA_MASK	0x000000FF
#  define BLUE(x)     ((x) >> 24)
#  define GREEN(x)   (((x) >> 16) & 0xFF)
#  define RED(x)     (((x) >>  8) & 0xFF)
#  define ALPHA(x)    ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(b) << 24) & 0xFF000000) | ((int(g) << 16) & 0xFF0000) | ((int(r) << 8) & 0xFF00) | (int(a) & 0xFF))
#else	// IA
#  define COLOR_MASK	0x00FFFFFF
#  define ALPHA_MASK	0xFF000000
#  define ALPHA(x)    ((x) >> 24)
#  define RED(x)     (((x) >> 16) & 0xFF)
#  define GREEN(x)   (((x) >>  8) & 0xFF)
#  define BLUE(x)     ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(a) << 24) & 0xFF000000) | ((int(r) << 16) & 0xFF0000) | ((int(g) << 8) & 0xFF00) | (int(b) & 0xFF))
#endif

struct bmp_settings
{
	BPoint  origin;
	int		bit_depth;	// 1, 4, 8, or 24
	int		compress;
	bool	settings_touched;
};
BLocker g_settings_lock ("BMP Settings Lock");
bmp_settings g_settings;

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Bit Depth = %d\n", g_settings.bit_depth);
	printf ("Compress = %d\n", g_settings.compress);
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
	 PrefsLoader ()
	 {
	 	// printf ("PrefsLoader... "); fflush (stdout);
	 	g_settings_lock.Lock();
	 	// defaults:
		g_settings.origin = B_ORIGIN;
		g_settings.bit_depth = 24;
		g_settings.compress = 1;
		g_settings.settings_touched = false;
		
		BPath path;
		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
		{
			path.SetTo ("/tmp");
		}
		path.Append ("BMPTranslator_Settings");
		FILE *fp = fopen (path.Path(), "r");
		if (fp)
		{
			char line[1024];
			char name[32];
			char *ptr;
			while (true)
			{
				line[0] = 0;
				fgets (line, 1024, fp);
				if (!line[0])
					break;
				ptr = line;
				while (isspace (*ptr))
					ptr++;
				if (*ptr == '#' || !*ptr)	// Comment line
					continue;
				if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
				{
					fprintf (stderr, "Strange BMPTranslator settings line: %s\n", name);
				}
				else
				{
					if (!strcmp (name, "window_origin"))
					{
						while (*ptr != '=')
							ptr++;
						ptr++;
						if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
							fprintf (stderr, "Illegal window origin in BMP Translator settings: %s\n", ptr);
					}
					else if (!strcmp (name, "bit_depth"))
					{
						while (*ptr != '=')
							ptr++;
						ptr++;
						if ((sscanf (ptr, "%d", &g_settings.bit_depth) != 1) || ((g_settings.bit_depth != 1) && (g_settings.bit_depth != 4) && (g_settings.bit_depth != 8) && (g_settings.bit_depth != 24)))
						{
							fprintf (stderr, "Illegal bit depth in BMP Translator settings: %s\n", ptr);
							g_settings.bit_depth = 24;
						}
					}
					else if (!strcmp (name, "compress"))
					{
						while (*ptr != '=')
							ptr++;
						ptr++;
						if ((sscanf (ptr, "%d", &g_settings.compress) != 1) || ((g_settings.compress != 1) && (g_settings.compress != 0)))
						{
							fprintf (stderr, "Illegal compression in BMP Translator settings: %s\n", ptr);
							g_settings.compress = 0;
						}
					}
					else
					{
						fprintf (stderr, "Unknown BMPTranslator setting: %s\n", name);
					}
				}
			}
			fclose (fp);
		}
		g_settings_lock.Unlock();
		// printf ("Done.\n");
	 }
	 
	~PrefsLoader ()
	 {
	 	// printf ("~PrefsLoader... "); fflush (stdout);
	 	g_settings_lock.Lock();
	 	if (g_settings.settings_touched)
	 	{
	 		BPath path;
	 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
	 		{
	 			path.SetTo ("/tmp");
	 		}
	 		path.Append ("BMPTranslator_Settings");
	 		FILE *fp = fopen (path.Path(), "w");
	 		if (fp)
	 		{
	 			fprintf (fp, "# BMPTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
	 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
	 			fprintf (fp, "bit_depth = %d\n", g_settings.bit_depth);
	 			fprintf (fp, "compress = %d\n", g_settings.compress);
	 			fclose (fp);
	 		}
	 	}
	 	g_settings_lock.Unlock();
	 	// printf ("Done.\n");
	 }
};

PrefsLoader g_prefsloader;


class BMPDecoder
{
public:
				 BMPDecoder (BPositionIO *inSource, BPositionIO *outDestination);
	status_t	 Decode ();

private:
	status_t	 Decode1 ();
	status_t	 Decode4 ();
	status_t	 Decode8 ();
	status_t	 Decode24 ();
	status_t	 Decode32 ();

	BPositionIO	*in, *out;
	
	uint32		 fWidth;
	uint32		 fHeight;
	
	uint32		 fColorMap[MAX_COLORMAP];
	uint32		 fNumColors;
	uint32		 fCompression;
	uint32		*fBitmap;
};


BMPDecoder::BMPDecoder (BPositionIO *inSource, BPositionIO *outDestination)
: in (inSource), out (outDestination)
{
}


// Note: Upon entry, the 2-byte 'BM' header has already been read
status_t BMPDecoder::Decode ()
{
	char bmfileheader[12];
	if (in->Read (bmfileheader, 12L) != 12)
	{
		fprintf (stderr, "BMPDecoder:  Couldn't read bitmap file header\n");
		return B_ERROR;
	}
	
//	uint32 bfSize = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (bmfileheader)));
	uint32 bfOffBits = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (bmfileheader + 8)));
	
//	printf ("bfSize = %ld, bfOffBits = %ld\n", bfSize, bfOffBits);
	
	// We now might have either a BITMAPCOREHEADER or a BITMAPHEADER.
	// In both cases, the next DWORD is the size of the header.
	
	char hdrsize[4];
	if (in->Read (hdrsize, 4L) != 4L)
	{
		fprintf (stderr, "BMPDecoder:  Couldn't read header size\n");
		return B_ERROR;
	}
	
	uint32 hdrSize = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) hdrsize));

	uint32 bitCount;	// skip "Planes" - it's always 1
	uint32 sizeImage = 0, clrUsed = 256, clrImportant = 0;
	
	uint32 hdrtype;

	if (hdrSize == 12)	// BITMAPCOREHEADER
	{
		hdrtype = BMP_CORE_HDR;
		char coreheader[8];
		if (in->Read (coreheader, 8L) != 8)
		{
			fprintf (stderr, "BMPDecoder:  Couldn't read CORE header\n");
			return B_ERROR;
		}
		fWidth   = B_LENDIAN_TO_HOST_INT16 (*((uint16 *) (coreheader)));
		fHeight  = B_LENDIAN_TO_HOST_INT16 (*((uint16 *) (coreheader + 2)));
		bitCount = B_LENDIAN_TO_HOST_INT16 (*((uint16 *) (coreheader + 6)));
		fCompression = 0;
	}
	else				// BITMAPINFOHEADER
	{
		hdrtype = BMP_INFO_HDR;
		char *header = new char [hdrSize - 4];
		if (in->Read (header, hdrSize - 4L) != hdrSize - 4)
		{
			fprintf (stderr, "BMPDecoder:  Couldn't read info header\n");
			return B_ERROR;
		}
		fWidth        = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (header)));
		fHeight       = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (header + 4)));
		bitCount      = B_LENDIAN_TO_HOST_INT16 (*((uint16 *) (header + 10)));
		
		fCompression  = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (header + 12)));
		sizeImage     = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (header + 16)));
		clrUsed       = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (header + 28)));
		clrImportant  = B_LENDIAN_TO_HOST_INT32 (*((uint32 *) (header + 32)));
		
		delete [] header;
	}
	
//	printf ("(%ld, %ld); Bit Depth = %ld, %sompressed\n", fWidth, fHeight, bitCount, fCompression ? "C" : "Unc");
//	printf ("%s; colors used %ld, colors important %ld\n", hdrtype == BMP_INFO_HDR ? "BMP_INFO_HDR" : "BMP_CORE_HDR", clrUsed, clrImportant);
	
	if (bitCount == 1 || bitCount == 4 || bitCount == 8)	// color mapped formats
	{
		uint32 num_colors = 1 << bitCount;
		uint8 *colormap;
		in->Seek (14 + hdrSize, SEEK_SET);	// just to be sure
		if (hdrtype == BMP_INFO_HDR)	// we have RGB quads
		{
			colormap = new uint8 [num_colors*4];
			if (in->Read (colormap, num_colors*4L) != num_colors*4)
			{
				fprintf (stderr, "BMPDecoder:  Couldn't read RGB quad colormap\n");
				delete [] colormap;
				return B_ERROR;
			}
			uint8 *cur = colormap;
			for (uint32 i = 0; i < num_colors; i++)
			{
				uint32 blue = *cur++;
				uint32 green = *cur++;
				uint32 red = *cur++;
				cur++;	// skip A
				fColorMap[i] = PIXEL (red, green, blue, 255);
				// printf ("color %ld: (%ld, %ld, %ld)\n", i, red, green, blue);
			}
		}
		else							// we have RGB tiplets
		{
			colormap = new uint8 [num_colors*3];
			if (in->Read (colormap, num_colors*4L) != num_colors*3)
			{
				fprintf (stderr, "BMPDecoder:  Couldn't read RGB triplet colormap\n");
				delete [] colormap;
				return B_ERROR;
			}
			uint8 *cur = colormap;
			for (uint32 i = 0; i < num_colors; i++)
			{
				uint32 red = *cur++;
				uint32 green = *cur++;
				uint32 blue = *cur++;
				fColorMap[i] = PIXEL (red, green, blue, 255);
			}
		}
		delete [] colormap;
		if (clrUsed == 256)
			fNumColors = num_colors;
		else
			fNumColors = clrUsed;
	}
	else if (bitCount != 24 && bitCount != 32)	// Hmm, I've seen 32bits BMPs too...
	{
		fprintf (stderr, "BMPDecoder:  Invalid bit count (%ld)\n", bitCount);
		return B_ERROR;
	}

	fBitmap = new uint32 [fWidth*fHeight];
	status_t err = B_ERROR;
	in->Seek (bfOffBits, SEEK_SET);
	switch (bitCount)
	{
		case 1:
			err = Decode1();
			break;
		case 4:
			err = Decode4();
			break;
		case 8:
			err = Decode8();
			break;
		case 24:
			err = Decode24();
			break;
		case 32:
			err = Decode32();
			break;
		default:
			fprintf (stderr, "BMPDecoder:  Invalid bit count (%ld)\n", bitCount);
	}
//	printf ("Decoded\n");
	
	if (err != B_OK)
	{
		fprintf (stderr, "BMPDecoder: Error decoding bitmap data\n");
		delete [] fBitmap;
		return err;
	}
	
	TranslatorBitmap tbitmap;
	tbitmap.magic    = B_TRANSLATOR_BITMAP;
	tbitmap.bounds   = BRect (0, 0, fWidth - 1, fHeight - 1);
	tbitmap.rowBytes = fWidth*4;
	tbitmap.colors   = B_RGB_32_BIT;
	tbitmap.dataSize = fHeight*fWidth*4;
	
	// tbitmap.bounds.PrintToStream();

	swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	
	err = out->SetSize (sizeof (TranslatorBitmap) + fWidth*fHeight*4);
	if (err < 0)
	{
		fprintf (stderr, "BMPTranslatore: SetSize returns %ld\n", err);
		delete [] fBitmap;
		return err;
	}
	err = out->Write (&tbitmap, sizeof (TranslatorBitmap));
	if (err < 0)
	{
		fprintf (stderr, "BMPTranslator: Write (header) returns %ld\n", err);
		delete [] fBitmap;
		return err;
	}
	err = out->Write (fBitmap, fWidth*fHeight*4);
	if (err < 0)
	{
		fprintf (stderr, "BMPTranslator: Write (bitmap) returns %ld\n", err);
		delete [] fBitmap;
		return err;
	}

	return B_OK;
}

status_t BMPDecoder::Decode1 ()
{
	uint32 padwidth = fWidth/8;
	if (fWidth % 8)
		padwidth++;
	if (padwidth % 4)
		padwidth += 4 - (padwidth % 4);
	uint8 *pixels = new uint8 [padwidth];
	for (uint32 y = fHeight - 1; y > 0; y--)
	{
		uint32 *dline = fBitmap + y*fWidth;
		if (in->Read (pixels, padwidth) != padwidth)
		{
			delete [] pixels;
			return B_ERROR;
		}
		uint8 *sline = pixels;
		for (uint32 x = 0; x < fWidth/8; x++)
		{
			uint8 eightpix = *sline++;
			for (uint32 i = 0; i < 8; i++)
				*dline++ = fColorMap [(eightpix >> (7 - i)) & 0x1];
		}
		if (fWidth % 8)
		{
			uint8 lastpix = *sline;
			for (uint32 i = 0; i < fWidth % 8; i++)
				*dline++ = fColorMap [(lastpix >> (7 - i)) & 0x1];
		}
	}
	// handle last line separately - needn't be padded :-(
	uint32 *dline = fBitmap;
	uint32 lwidth = fWidth/8;
	if (fWidth % 8) lwidth++;
	if (in->Read (pixels, lwidth) != lwidth)
	{
		delete [] pixels;
		return B_ERROR;
	}
	uint8 *sline = pixels;
	for (uint32 x = 0; x < fWidth/8; x++)
	{
		uint8 eightpix = *sline++;
		for (uint32 i = 0; i < 8; i++)
			*dline++ = fColorMap [(eightpix >> (7 - i)) & 0x1];
	}
	if (fWidth % 8)
	{
		uint8 lastpix = *sline;
		for (uint32 i = 0; i < fWidth % 8; i++)
			*dline++ = fColorMap [(lastpix >> (7 - i)) & 0x1];
	}
	delete [] pixels;
	return B_OK;
}

status_t BMPDecoder::Decode4 ()
{
	if (fCompression)
	{
		int32 y = fHeight - 1;
		uint32 *sline = fBitmap + y*fWidth;
		while (y >= 0)
		{
			uint8 cmd[2];
			if (in->Read (cmd, 2L) != 2L)
			{
				fprintf (stderr, "BMPDecode8: Couldn't read cmd block\n");
				return B_ERROR;
			}
			if (cmd[0] == 0)	// escape
			{
				if (cmd[1] == 1)	// end of bitmap
					return B_OK;
				if (cmd[1] == 0)	// end of line
				{
					y--;
					sline = fBitmap + y*fWidth;
				}
				else if (cmd[1] == 2)	// delta
				{
					if (in->Read (cmd, 2L) != 2L)
					{
						fprintf (stderr, "BMPDecode8: Couldn't read delta\n");
						return B_ERROR;
					}
					sline += cmd[0];
					y -= cmd[1];
					sline -= cmd[1]*fWidth;
				}
				else	// uncompressed run
				{
					uint32 unclen = cmd[1];
					if (unclen % 2) unclen++;
					char *unc = new char [unclen];
					if (in->Read (unc, unclen) != unclen)
					{
						fprintf (stderr, "BMPDecode8: Couldn't read uncompressed run\n");
						delete [] unc;
						return B_ERROR;
					}
					for (uint32 i = 0; i < cmd[1]; i++)
					{
						int val1 = unc[i] >> 4;
						int val2 = unc[i] & 0xF;
						*sline++ = fColorMap[val1];
						*sline++ = fColorMap[val2];
					}
					delete [] unc;
				}
			}
			else
			{
				uint32 val1 = fColorMap[cmd[1] >> 4];
				uint32 val2 = fColorMap[cmd[1] & 0xF];
				for (uint32 i = 0; i < cmd[0]; i++)
				{
					*sline++ = val1;
					*sline++ = val2;
				}
			}
		}
		return B_OK;	// but don't trust this...
	}
	else
	{
		uint32 padwidth = fWidth/2;
		if (fWidth % 2)	// odd width
			padwidth++;
		if (padwidth % 4)
			padwidth += 4 - (padwidth % 4);
//		printf ("width = %ld, padwitdh = %ld\n", fWidth, padwidth);
		uint8 *pixels = new uint8 [padwidth];
		for (uint32 y = fHeight - 1; y > 0; y--)
		{
			uint32 *dline = fBitmap + y*fWidth;
			if (in->Read (pixels, padwidth) != padwidth)
			{
				delete [] pixels;
				return B_ERROR;
			}
			uint8 *sline = pixels;
			for (uint32 x = 0; x < fWidth/2; x++)
			{
				uint8 twopix = *sline++;
				*dline++ = fColorMap [twopix >> 4];
				*dline++ = fColorMap [twopix & 0x0F];
			}
			if (fWidth % 2)
			{
				uint8 lastpix = *sline;
				*dline = fColorMap [lastpix >> 4];
			}
		}
		// handle last line separately - needn't be padded :-(
		uint32 *dline = fBitmap;
		uint32 lwidth = fWidth/2;
		if (fWidth % 2) lwidth++;
		if (in->Read (pixels, lwidth) != lwidth)
		{
			delete [] pixels;
			return B_ERROR;
		}
		uint8 *sline = pixels;
		for (uint32 x = 0; x < fWidth/2; x++)
		{
			uint8 twopix = *sline++;
			*dline++ = fColorMap [twopix >> 4];
			*dline++ = fColorMap [twopix & 0x0F];
		}
		if (fWidth % 2)
		{
			uint8 lastpix = *sline;
			*dline = fColorMap [lastpix >> 4];
		}
		delete [] pixels;
	}
	return B_OK;
}

status_t BMPDecoder::Decode8 ()
{
	if (fCompression)
	{
		int32 y = fHeight - 1;
		uint32 *sline = fBitmap + y*fWidth;
		while (y >= 0)
		{
			// printf ("%ld: ", y); fflush (stdout);
			uint8 cmd[2];
			if (in->Read (cmd, 2L) != 2L)
			{
				fprintf (stderr, "BMPDecode8: Couldn't read cmd block\n");
				return B_ERROR;
			}
			if (cmd[0] == 0)	// escape
			{
				if (cmd[1] == 1)	// end of bitmap
					return B_OK;
				if (cmd[1] == 0)	// end of line
				{
					y--;
					sline = fBitmap + y*fWidth;
				}
				else if (cmd[1] == 2)	// delta
				{
					if (in->Read (cmd, 2L) != 2L)
					{
						fprintf (stderr, "BMPDecode8: Couldn't read delta\n");
						return B_ERROR;
					}
					// fprintf (stderr, "(%d, %d) ", cmd[0], cmd[1]); fflush (stderr);
					sline += cmd[0];
					y -= cmd[1];
					sline -= cmd[1]*fWidth;
				}
				else	// uncompressed run
				{
					// printf ("%d ", cmd[1]); fflush (stdout);
					uint32 unclen = cmd[1];
					if (unclen % 2) unclen++;
					uint8 *unc = new uint8 [unclen];
					if (in->Read (unc, unclen) != unclen)
					{
						fprintf (stderr, "BMPDecode8: Couldn't read uncompressed run\n");
						delete [] unc;
						return B_ERROR;
					}
					for (uint32 i = 0; i < cmd[1]; i++)
					{
						*sline++ = fColorMap[unc[i]];
					}
					delete [] unc;
				}
			}
			else
			{
				// printf ("[%d, %d] ", cmd[0], cmd[1]); fflush (stdout);
				uint32 val = fColorMap[cmd[1]];
				for (uint32 i = 0; i < cmd[0]; i++)
				{
					*sline++ = val;
				}
			}
		}
		return B_OK;	// but don't trust this...
	}
	else
	{
		uint32 padwidth = fWidth;
		if (padwidth % 4)
			padwidth += 4 - (padwidth % 4);
//		printf ("width = %ld, padwitdh = %ld\n", fWidth, padwidth);
		uint8 *pixels = new uint8 [padwidth];
		for (uint32 y = fHeight - 1; y > 0; y--)
		{
			uint32 *dline = fBitmap + y*fWidth;
			if (in->Read (pixels, padwidth) != padwidth)
			{
				delete [] pixels;
				return B_ERROR;
			}
			uint8 *sline = pixels;
			for (uint32 x = 0; x < fWidth; x++)
			{
				*dline++ = fColorMap [*sline++];
			}
		}
		// handle last line separately - needn't be padded :-(
		uint32 *dline = fBitmap;
		if (in->Read (pixels, padwidth) != padwidth)
		{
			delete [] pixels;
			return B_ERROR;
		}
		uint8 *sline = pixels;
		for (uint32 x = 0; x < fWidth; x++)
		{
			*dline++ = fColorMap [*sline++];
		}
		delete [] pixels;
	}
	return B_OK;
}

status_t BMPDecoder::Decode24 ()
{
	uint32 padwidth = fWidth*3;
	if (padwidth % 4)
		padwidth += 4 - (padwidth % 4);
	uint8 *pixels = new uint8 [padwidth];
	for (uint32 y = fHeight - 1; y > 0; y--)
	{
		uint32 *dline = fBitmap + y*fWidth;
		if (in->Read (pixels, padwidth) != padwidth)
		{
			fprintf (stderr, "BMPDecode24 - couldn't read data\n");
			delete [] pixels;
			return B_ERROR;
		}
		uint8 *sline = pixels;
		for (uint32 x = 0; x < fWidth; x++)
		{
			uint32 blue = *sline++;
			uint32 green = *sline++;
			uint32 red = *sline++;
			*dline++ = PIXEL (red, green, blue, 255);
		}
	}
	// Handle last line separately - needn't be padded :-(
	uint32 *dline = fBitmap;
	if (in->Read (pixels, fWidth*3) != fWidth*3)
	{
		fprintf (stderr, "BMPDecode24 - couldn't read data\n");
		delete [] pixels;
		return B_ERROR;
	}
	uint8 *sline = pixels;
	for (uint32 x = 0; x < fWidth; x++)
	{
		uint32 blue = *sline++;
		uint32 green = *sline++;
		uint32 red = *sline++;
		*dline++ = PIXEL (red, green, blue, 255);
	}
	delete [] pixels;
	return B_OK;
}

status_t BMPDecoder::Decode32 ()
{
	uint32 padwidth = fWidth*4;
	uint8 *pixels = new uint8 [padwidth];
	for (uint32 y = fHeight - 1; y > 0; y--)
	{
		uint32 *dline = fBitmap + y*fWidth;
		if (in->Read (pixels, padwidth) != padwidth)
		{
			fprintf (stderr, "BMPDecode32 - couldn't read data\n");
			delete [] pixels;
			return B_ERROR;
		}
		uint8 *sline = pixels;
		for (uint32 x = 0; x < fWidth; x++)
		{
			uint32 blue = *sline++;
			uint32 green = *sline++;
			uint32 red = *sline++;
			uint32 alpha = *sline++;
			*dline++ = PIXEL (red, green, blue, alpha);
		}
	}
	// Handle last line separately - needn't be padded :-(
	uint32 *dline = fBitmap;
	if (in->Read (pixels, fWidth*4) != fWidth*4)
	{
		fprintf (stderr, "BMPDecode32 - couldn't read last line\n");
		delete [] pixels;
		return B_ERROR;
	}
	uint8 *sline = pixels;
	for (uint32 x = 0; x < fWidth; x++)
	{
		uint32 blue = *sline++;
		uint32 green = *sline++;
		uint32 red = *sline++;
		uint32 alpha = *sline++;
		*dline++ = PIXEL (red, green, blue, alpha);
	}
	delete [] pixels;
	return B_OK;
}

class BMPEncoder
{
public:
				 BMPEncoder (BPositionIO *inSource, BPositionIO *outDestination);
	status_t	 Encode ();

private:
	status_t	 Encode1 ();
	status_t	 Encode4 ();
	status_t	 Encode8 ();
	status_t	 Encode24 ();
	status_t	 Encode32 ();
	int			 construct_colormap ();
	int			 index_for (uint32 pixel);
	uint32		 cdiff (uint32 a, uint32 b);
	status_t	 find_run4 (uint32 *s, uint32 x, uint32 *start, uint32 *len);
	status_t	 find_run8 (uint32 *s, uint32 x, uint32 *start, uint32 *len);

	BPositionIO	*in, *out;
	int			 fNumColors;
	BBitmap		*fSrc;
	uint32		 fWidth;
	uint32		 fHeight;
	bool		 fCompress;
	uint32		 fColorMap[MAX_COLORMAP];
	uint8		*fScratch;
	uint32		 fSize;
};


BMPEncoder::BMPEncoder (BPositionIO *inSource, BPositionIO *outDestination)
: in (inSource), out (outDestination)
{
}

uint32 BMPEncoder::cdiff (uint32 a, uint32 b)
{
	int32 rdiff = RED(a) - RED(b);
	int32 gdiff = GREEN(a) - GREEN(b);
	int32 bdiff = BLUE(a) - BLUE(b);
	return (9*rdiff*rdiff + 36*gdiff+gdiff + bdiff+bdiff)/7;
}

int BMPEncoder::index_for (uint32 pixel)
{
	for (int i = 0; i < fNumColors; i++)
	{
		if ((pixel & COLOR_MASK) == fColorMap[i])
			return i;
	}
	// Palette didn't contain enough colors - pick the one with the smallest visual distance.
	// Do a stupid calculation of the color distance.
	uint32 sofar = 256*256*256;
	int idx = 0;
	for (int i = 0; i < fNumColors; i++)
	{
		uint32 thisdiff = cdiff ((pixel & COLOR_MASK), fColorMap[i]);
		if (thisdiff < sofar)
		{
			idx = i;
			sofar = thisdiff;
		}
	}
	return idx;
}

int BMPEncoder::construct_colormap ()
{
	int count = 1;
	
	uint32 *s = (uint32 *) fSrc->Bits();
	fColorMap[0] = *s & COLOR_MASK;

	for (int32 i = fSrc->BitsLength()/4; i > 0; i--)
	{
		uint32 pixel = *(++s);
		uint32 pc = pixel & COLOR_MASK;
		int j = 0;
		while (pc != fColorMap[j] && j < count)
			j++;
		if (j == count)	// We have a new color
		{
			fColorMap[count] = pc;
			count++;
			if (count > MAX_COLORMAP)
				return count - 1;
		}
	}
	return count - 1;
}

status_t BMPEncoder::find_run4 (uint32 *s, uint32 x, uint32 *start, uint32 *len)
{
	return B_OK;
}

// finds the first occurrence of a 3-or-more-the-same run
status_t BMPEncoder::find_run8 (uint32 *s, uint32 x, uint32 *start, uint32 *len)
{
retry:
	if (x > fWidth - 3)
		return B_ERROR;

	*start = x;
	*len = 1;
	uint8 val = *s++;
	while (++x < fWidth)
	{
		if (*s++ == val)
			(*len)++;
		else
		{
			if (*len > 2)
				return B_OK;
			else
				goto retry;
		}
	}
	if (*len > 2)
		return B_OK;
	else
		return B_ERROR;
}

status_t BMPEncoder::Encode ()
{
	// Get info on the bitmap
	TranslatorBitmap tbitmap;
	in->Seek (0L, SEEK_SET);
	if (in->Read (&tbitmap, sizeof (TranslatorBitmap)) < sizeof (TranslatorBitmap))
	{
		fprintf (stderr, "BMPTranslator: Couldn't read header from inStream\n");
		return B_ERROR;
	}
	swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	
	BRect bounds = tbitmap.bounds;

	fSrc = new BBitmap (bounds, B_RGB32);

	fWidth  = bounds.IntegerWidth() + 1;
	fHeight = bounds.IntegerHeight() + 1;
	// bounds.PrintToStream();
	// printf ("(%ld x %ld)\n", fWidth, fHeight);
	switch (tbitmap.colors)
	{
		case B_RGB32:
		case B_RGBA32:
		{
			if (in->Read (fSrc->Bits(), tbitmap.dataSize) != tbitmap.dataSize)
			{
				fprintf (stderr, "BMPTranslator:  Error reading bitmap data (stream truncated?)\n");
				delete fSrc;
				return B_ERROR;
			}
			break;
		}
		default:
			fprintf (stderr, "BMPTranslator: Unsupported input color_space (only 32bit please)\n");
			delete fSrc;
			return B_NO_TRANSLATOR;
	}
	
	g_settings_lock.Lock();
	uint32 bit_depth = g_settings.bit_depth;
	fCompress = g_settings.compress;
	g_settings_lock.Unlock();

	uint32 palette_size;
	if (bit_depth == 24)
		palette_size = 1;	// just a NULL entry in this case
	else
		palette_size = 1 << bit_depth;

	if (bit_depth != 24)
	{
//		printf ("Quantize!\n");
//		We do a poor job of quantizing here: No dithering, for example.
//		You should provide a properly dithered/quantize image yourself.
		BBitmap *src8 = new BBitmap (bounds, B_CMAP8, true);
		src8->Lock();
		BView *v = new BView (bounds, "tmp cq view", 0, 0);
		src8->AddChild (v);
		v->DrawBitmap (fSrc);	// Use the BeOS to render down to 256 colors max
		src8->RemoveChild (v);
		delete v;
		uint32 *dst = (uint32 *) fSrc->Bits() - 1;
		const rgb_color *spalette = system_colors()->color_list;
		for (int i = 0; i < MAX_COLORMAP; i++)
		{
			fColorMap[i] = PIXEL (spalette[i].red, spalette[i].green, spalette[i].blue, 0);
		}
		for (int y = 0; y < fHeight; y++)
		{
			uint8 *row8 = (uint8 *) src8->Bits() + y*src8->BytesPerRow() - 1;
			for (int x = 0; x < fWidth; x++)
			{
				uint32 pixel = *(++dst);
				*dst = (fColorMap[*(++row8)] & COLOR_MASK) | (pixel & ALPHA_MASK);
			}
		}
		delete src8;
		fNumColors = construct_colormap();
		// printf ("Need %d colors\n", fNumColors);
		if (fNumColors > palette_size)
		{
			fprintf (stderr, "BMPEncoder: Need %d colors, only got %ld - truncating\n", fNumColors, palette_size);
			fNumColors = palette_size;
		}
	}
	
	// Unfortunately, BMP wants the file size in the header,
	// so we'll have to encode in a scratch buffer and then write out the size.
	
	fScratch = new uint8 [uint32 (fSrc->BitsLength() + 256)];	// be on the safe side here.
	fSize = 0;	// The routines below adjust this.
	
	status_t err = B_ERROR;
	switch (bit_depth)
	{
		case 1:
			err = Encode1();
			break;
				
		case 4:
			err = Encode4();
			break;
				
		case 8:
			err = Encode8();
			break;
				
		case 24:
			err = Encode24();
			break;
				
		default:
			fprintf (stderr, "BMPEncode:  Unsupported bit depth (%ld)\n", bit_depth);
	}

	if (err != B_OK)
	{
		fprintf (stderr, "BMPEncoder: Error during encoding\n");
		delete [] fScratch;
		delete fSrc;
		return err;
	}
	
	delete fSrc;	// we have the encoded data in fScratch now

	// Make the BITMSPFILEHEADER and write it out
	
	// printf ("fSize = %ld\n", fSize);
	uint32 padSize = 14 + 40 + palette_size*4;

	char bmFileHeader[14];
	bmFileHeader[0] = 'B';	// bfType
	bmFileHeader[1] = 'M';
	*((uint32 *) (bmFileHeader + 2)) = B_HOST_TO_LENDIAN_INT32 (fSize + padSize);	// bfSize;
	bmFileHeader[6] = 0;	// bfReserved1
	bmFileHeader[7] = 0;
	bmFileHeader[8] = 0;	// bfReserved2
	bmFileHeader[9] = 0;
	*((uint32 *) (bmFileHeader + 10)) = B_HOST_TO_LENDIAN_INT32 (padSize);	// bfOffBits;

	if (out->Write (bmFileHeader, 14L) != 14L)
	{
		fprintf (stderr, "BMPEncoder: Couldn't write file header\n");
		delete [] fScratch;
		return B_ERROR;
	}
	
	// Make the BITMAPINFOHEADER and write it out
	uint32 biCompression = BI_RGB;
	if (fCompress)
	{
		if (bit_depth == 4)
			biCompression = BI_RLE4;
		if (bit_depth == 8)
			biCompression = BI_RLE8;
	}
	
	char bmInfoHeader[40];
	*((uint32 *) (bmInfoHeader))	  = B_HOST_TO_LENDIAN_INT32 (40);			// biSize
	*((uint32 *) (bmInfoHeader +  4)) = B_HOST_TO_LENDIAN_INT32 (fWidth);		// biWidth
	*((uint32 *) (bmInfoHeader +  8)) = B_HOST_TO_LENDIAN_INT32 (fHeight);		// biHeight
	*((uint32 *) (bmInfoHeader + 12)) = B_HOST_TO_LENDIAN_INT16 (1);			// biPlanes
	*((uint32 *) (bmInfoHeader + 14)) = B_HOST_TO_LENDIAN_INT16 (bit_depth);	// biBitCount
	*((uint32 *) (bmInfoHeader + 16)) = B_HOST_TO_LENDIAN_INT32 (biCompression);// biCompression
	*((uint32 *) (bmInfoHeader + 20)) = B_HOST_TO_LENDIAN_INT32 (fSize);		// biSizeImage
	*((uint32 *) (bmInfoHeader + 24)) = B_HOST_TO_LENDIAN_INT32 (2835);			// biXPelsPerMeter: 72 dpi
	*((uint32 *) (bmInfoHeader + 28)) = B_HOST_TO_LENDIAN_INT32 (2835);			// biYPelsPerMeter: 72 dpi
	*((uint32 *) (bmInfoHeader + 32)) = B_HOST_TO_LENDIAN_INT32 (fNumColors);	// biClrUsed;
	*((uint32 *) (bmInfoHeader + 36)) = B_HOST_TO_LENDIAN_INT32 (fNumColors);	// biClrImportant;
	
	if (out->Write (bmInfoHeader, 40L) != 40L)
	{
		fprintf (stderr, "BMPEncoder: Couldn't write out bitmap info header\n");
		delete [] fScratch;
		return B_ERROR;
	}
	
	// Write out the color palette (if in non-24 bit space)
	if (bit_depth == 24)
	{
		char palette[4];	// NULL
		if (out->Write (palette, 4L) != 4L)
		{
			fprintf (stderr, "BMPEncoder: Couldn't write out NULL palette\n");
			delete [] fScratch;
			return B_ERROR;
		}
	}
	else
	{
		uint8 *palette = new uint8 [palette_size*4];
		uint8 *pentry = palette;
		for (uint32 i = 0; i < fNumColors; i++)
		{
			uint32 color = fColorMap[i];
			*pentry++ = BLUE(color);
			*pentry++ = GREEN(color);
			*pentry++ = RED(color);
			*pentry++ = 0;
		}
		// printf ("Writing out palette (%ld*4 = %ld bytes)\n", palette_size, palette_size*4);
		if (out->Write (palette, palette_size*4L) != palette_size*4L)
		{
			fprintf (stderr, "BMPEncoder: Couldn't write out palette\n");
			delete [] fScratch;
			delete [] palette;
			return B_ERROR;
		}
		delete [] palette;
	}
	
	err = B_OK;
	// Now we write out the (possibly compressed) image data in our scratch buffer
	if (out->Write (fScratch, fSize) != fSize)
	{
		fprintf (stderr, "BMPEncoder: Couldn't write out bitmap data\n");
		err = B_ERROR;
	}
	
	delete [] fScratch;
	return err;
}

status_t BMPEncoder::Encode1 ()
{
	uint8 *dest = fScratch;
	for (uint32 y = 0; y < fHeight; y++)
	{
		uint32 *sline = (uint32 *) fSrc->Bits() + (fHeight - y - 1)*fSrc->BytesPerRow()/4;
		for (uint32 x = 0; x < fWidth/8; x++)
		{
			// get eight pixels
			uint8 p;
			uint8 d = 0;
			for (uint32 i = 0; i < 8; i++)
			{
				p = index_for (*sline++);
				d |= (p << (7 - i));
			}
			*dest++ = d;
		}
		fSize += fWidth/8;
		if (fWidth % 8)
		{
			// get the remaining pixels
			uint8 p;
			uint8 d = 0;
			for (uint32 i = 0; i < fWidth % 8; i++)
			{
				p = index_for (*sline++);
				d |= (p << (7 - 1));
			}
			*dest++ = d;
			fSize++;
		}
		if ((uint32) dest % 4)	// padding required
		{
			uint32 pad = 4 - ((uint32) dest % 4);
			dest += pad;
			fSize += pad;
		}
	}
	return B_OK;
}

status_t BMPEncoder::Encode4 ()
{
	if (fCompress)
	{
		uint8 *dest = fScratch;
		for (uint32 y = 0; y < fHeight; y++)
		{
			uint32 *sline = (uint32 *) fSrc->Bits() + (fHeight - y - 1)*fSrc->BytesPerRow()/4;
			uint32 x = 0;
			while (x < fWidth/2)
			{
				uint8 val1 = index_for (*sline++);
				uint8 val2 = index_for (*sline++);
				x++;
				uint32 n = 1;
				bool ok = true;
				while (ok)
				{
					ok &= (x < fWidth/2);
					ok &= (val1 == index_for (*sline++));
					ok &= (val2 == index_for (*sline++));
					if (ok)
						ok &= (n++ < 256);
					if (ok)
						x++;
				}
				*dest++ = n;
				*dest++ = (val1 << 4) | val2;
				sline -= 2;
				fSize += 2;
			}
			*dest++ = 0;
			if (y < fHeight)
				*dest++ = 0;	// signal end of line
			else
				*dest = 1;		// end of bitmap
			fSize += 2;
		}
	}
	else
	{
		uint8 *dest = fScratch;
		for (uint32 y = 0; y < fHeight; y++)
		{
			uint32 *sline = (uint32 *) fSrc->Bits() + (fHeight - y - 1)*fSrc->BytesPerRow()/4;
			for (uint32 x = 0; x < fWidth/2; x++)
			{
				// get two pixels
				uint8 p1 = index_for (*sline++);
				uint8 p2 = index_for (*sline++);
				// combine into one byte
				*dest++ = (p1 << 4) | p2;
			}
			fSize += fWidth/2;
			if (fWidth % 2)
			{
				// get the remaining odd pixel
				uint8 p1 = index_for (*sline++);
				*dest++ = (p1 << 4);
				fSize++;
			}
			if ((uint32) dest % 4)	// padding required
			{
				uint32 pad = 4 - ((uint32) dest % 4);
				dest += pad;
				fSize += pad;
			}
		}
	}
	return B_OK;
}

status_t BMPEncoder::Encode8 ()
{
	if (fCompress)
	{
		uint8 *dest = fScratch;
		for (uint32 y = 0; y < fHeight; y++)
		{
			uint32 *sline = (uint32 *) fSrc->Bits() + (fHeight - y - 1)*fSrc->BytesPerRow()/4;
			uint32 x = 0;
			while (x < fWidth)
			{
				uint8 val = index_for (*sline++);
				x++;
				uint32 n = 1;
				while (x < fWidth && val == index_for (*sline++) && n++ < 256)
					x++;
				*dest++ = n;
				*dest++ = val;
				--sline;
				fSize += 2;
			}
			*dest++ = 0;
			if (y < fHeight)
				*dest++ = 0;	// signal end of line
			else
				*dest = 1;		// end of bitmap
			fSize += 2;
		}
	}
	else
	{
		uint8 *dest = fScratch;
		for (uint32 y = 0; y < fHeight; y++)
		{
			uint32 *sline = (uint32 *) fSrc->Bits() + (fHeight - y - 1)*fSrc->BytesPerRow()/4;
			for (uint32 x = 0; x < fWidth; x++)
			{
				*dest++ = index_for (*sline++);
			}
			fSize += fWidth;
			if ((uint32) dest % 4)	// padding required
			{
				uint32 pad = 4 - ((uint32) dest % 4);
				dest += pad;
				fSize += pad;
			}
		}
	}
	return B_OK;
}

status_t BMPEncoder::Encode24 ()
{
	uint8 *dest = fScratch;
	for (uint32 y = 0; y < fHeight; y++)
	{
		uint32 *sline = (uint32 *) fSrc->Bits() + (fHeight - y - 1)*fSrc->BytesPerRow()/4;
		for (uint32 x = 0; x < fWidth; x++)
		{
			uint32 color = *sline++;
			*dest++ = BLUE(color);
			*dest++ = GREEN(color);
			*dest++ = RED(color);
		}
		fSize += fWidth*4;
		if ((uint32) dest % 4)	// padding required
		{
			uint32 pad = 4 - ((uint32) dest % 4);
			for (uint32 j = 0; j < pad; j++)
				*dest++ = 0;
			fSize += pad;
		}
	}
	return B_OK;
}

status_t BMPEncoder::Encode32 ()
{
	uint8 *dest = fScratch;
	for (uint32 y = 0; y < fHeight; y++)
	{
		uint32 *sline = (uint32 *) fSrc->Bits() + (fHeight - y - 1)*fSrc->BytesPerRow()/4;
		for (uint32 x = 0; x < fWidth; x++)
		{
			uint32 color = *sline++;
			*dest++ = BLUE(color);
			*dest++ = GREEN(color);
			*dest++ = RED(color);
			*dest++ = ALPHA(color);
		}
		fSize += fWidth*4;
	}
	return B_OK;
}

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage *ioExtension, translator_info *outInfo, uint32 outType)
{
	inSource->Seek (0L, SEEK_SET);
	char header[4];
	size_t number_read = 0;
	number_read = inSource->Read (header, 4L);
	if (number_read == 4)
	{
		if (header[0] == 'B' && header[1] == 'M')
		// We have a BMP file in *inSource!
		{
			if (outType == BMP_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= BMP_CODE;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.5;	// BMP _can_ be 24 bit, but has an often-used 16 color (! - not bit) mode :-(
				outInfo->capability	= 0.6;	// Let's be modest for now.
				strcpy (outInfo->MIME, "image/bmp");
				strcpy (outInfo->name, "BMP Image");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == BMP_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= B_TRANSLATOR_BITMAP;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;
				outInfo->capability	= 0.1;
				strcpy (outInfo->MIME, "image/x-be-bitmap");
				strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage *ioExtension, uint32 outType, BPositionIO *outDestination)
{
	inSource->Seek (0L, SEEK_SET);
	char header[4];
	ssize_t number_read = 0;
	number_read = inSource->Read (header, 4L);
	if (number_read == 4)
	{
		if (header[0] == 'B' && header[1] == 'M')
		// We have a BMP file in *inSource!
		{
			if (outType == BMP_CODE)
			// "translate" it to BMP...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// decode the BMP to a bitmap
			{
				outType = B_TRANSLATOR_BITMAP;
				inSource->Seek (2L, SEEK_SET);
				BMPDecoder *decoder = new BMPDecoder (inSource, outDestination);
				status_t err = decoder->Decode ();
				return err;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == BMP_CODE)
			// encode the bitmap to BMP...
			{
				BMPEncoder *encoder = new BMPEncoder (inSource, outDestination);
				status_t err = encoder->Encode();
				return err;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// "translate" the bitmap...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

class BMPConfigView : public BView
{
public:
			 BMPConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
				SetViewColor (218, 218, 218, 255);
				
				colorsPU = new BPopUpMenu ("Output Colors");
				colorsPU->AddItem (new BMenuItem ("2 (duotone)", new BMessage ('pP01')));
				colorsPU->AddItem (new BMenuItem ("16", new BMessage ('pP04')));
				colorsPU->AddItem (new BMenuItem ("256", new BMessage ('pP08')));
				colorsPU->AddItem (new BMenuItem ("16.7 Million", new BMessage ('pP24')));
				colorsPU->AddItem (new BMenuItem ("16.7 Million, Alpha", new BMessage ('pP32')));
				colorsMF = new BMenuField (BRect (8, 70, 192, 90), "colorsMF", "Output Colors", colorsPU);
				colorsMF->SetDivider (80);
				AddChild (colorsMF);
				
				compressionCB = new BCheckBox (BRect (8, 94, 192, 112), "compressionCB", "Compress", new BMessage ('pCmp'));
				AddChild (compressionCB);
				
				g_settings_lock.Lock();
				switch (g_settings.bit_depth)
				{
					case 1:
						colorsPU->ItemAt(0)->SetMarked (true);
						break;
					case 4:
						colorsPU->ItemAt(1)->SetMarked (true);
						break;
					case 8:
						colorsPU->ItemAt(2)->SetMarked (true);
						break;
					case 24:
						colorsPU->ItemAt(3)->SetMarked (true);
						break;
					case 32:
						colorsPU->ItemAt(4)->SetMarked (true);
						break;
					default:
						fprintf (stderr, "BMPTranslator: Illegal bit depth\n");;
				}
				
				if (g_settings.bit_depth == 4 || g_settings.bit_depth == 8)
					compressionCB->SetEnabled (true);
				else
					compressionCB->SetEnabled (false);
				
				compressionCB->SetValue (g_settings.compress);
				
				g_settings_lock.Unlock();
			 }

virtual		~BMPConfigView ()
			 {
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
			 	BMessenger messenger (this);
				for (int i = 0; i < colorsPU->CountItems(); i ++)
				{
			 		BMenuItem *item = colorsPU->ItemAt (i);
			 		if (item)
			 			item->SetTarget (messenger);
				}
				compressionCB->SetTarget (messenger);
			 }
virtual void Draw (BRect update)
			 {
			 	SetFont (be_bold_font);
			 	SetDrawingMode (B_OP_OVER);
			 	DrawString ("BMP Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 2000 ∑ Sum Software", BPoint (8, 48));
			 	BView::Draw (update);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
			 	case 'pP01':
			 		g_settings_lock.Lock();
			 		g_settings.bit_depth = 1;
			 		g_settings_lock.Unlock();
			 		compressionCB->SetEnabled (false);
			 		break;
			 		
			 	case 'pP04':
			 		g_settings_lock.Lock();
			 		g_settings.bit_depth = 4;
			 		g_settings_lock.Unlock();
			 		compressionCB->SetEnabled (true);
			 		break;
			 		
			 	case 'pP08':
			 		g_settings_lock.Lock();
			 		g_settings.bit_depth = 8;
			 		g_settings_lock.Unlock();
			 		compressionCB->SetEnabled (true);
			 		break;
			 		
			 	case 'pP24':
			 		g_settings_lock.Lock();
			 		g_settings.bit_depth = 24;
			 		g_settings_lock.Unlock();
			 		compressionCB->SetEnabled (false);
			 		break;
			 		
				case 'pCmp':
					g_settings_lock.Lock();
					g_settings.compress = compressionCB->Value();
					g_settings_lock.Unlock();
					break;
					
				default:
					inherited::MessageReceived (msg);
				}
				g_settings_lock.Lock();
				g_settings.settings_touched = true;
				g_settings_lock.Unlock();
			 }
			 
private:
typedef BView inherited;
BPopUpMenu	*colorsPU;
BMenuField	*colorsMF;
BCheckBox	*compressionCB;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new BMPConfigView (BRect (0, 0, 200, 116), "BMPTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}

char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "BMP Translator";

int32 translatorVersion = 10;

translation_format inputFormats[] = {
	{ BMP_CODE, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/bmp", "Windows Device Independent Bitmap Format (BMP)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ BMP_CODE, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/bmp", "Windows Device Independent Bitmap Format (BMP)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
