#include <Locker.h>
#include <Path.h>
#include <FindDirectory.h>
#include <View.h>
#include <Screen.h>
#include <GraphicsDefs.h>
#include <InterfaceDefs.h>
#include <Slider.h>
#include <StringView.h>
#include <stdio.h>
#include <Bitmap.h>
#include <CheckBox.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <PopUpMenu.h>
#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>

status_t init_lib();

#define GS_BEOS			0x01
#define GS_ANTI_ALIAS	0x02
#define GS_AA_TEXT_ONLY	0x04

uint32 lib_cap();
static uint32 s_lib_cap = 0;
static bool g_drawn = false;

extern "C"
{
	#include "gsdll.h"
	void s_gsdll_draw (unsigned char *device, BView *view, BRect dest, BRect src);

	int GSDLLAPI s_gsdll_revision(char GSFAR * GSFAR * product, char GSFAR * GSFAR * copyright, long GSFAR * gs_revision, long GSFAR * gs_revisiondate);
	int GSDLLAPI s_gsdll_init(GSDLL_CALLBACK callback, HWND hwnd, int argc, char GSFAR * GSFAR * argv);
	int GSDLLAPI s_gsdll_execute_begin(void);
	int GSDLLAPI s_gsdll_execute_cont(const char GSFAR * str, int len);
	int GSDLLAPI s_gsdll_execute_end(void);
	int GSDLLAPI s_gsdll_exit(void);
	int GSDLLAPI s_gsdll_lock_device(unsigned char *device, int flag);
}

#define BUF_SIZE 4096
#define EPS_CODE 'EPS '

#if defined (__POWERPC__)
#  define COLOR_MASK	0xFFFFFF00
#  define ALPHA_MASK	0x000000FF
#  define BLUE(x)     ((x) >> 24)
#  define GREEN(x)   (((x) >> 16) & 0xFF)
#  define RED(x)     (((x) >>  8) & 0xFF)
#  define ALPHA(x)    ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(b) << 24) & 0xFF000000) | ((int(g) << 16) & 0xFF0000) | ((int(r) << 8) & 0xFF00) | (int(a) & 0xFF))
#else	// IA
#  define COLOR_MASK	0x00FFFFFF
#  define ALPHA_MASK	0xFF000000
#  define ALPHA(x)    ((x) >> 24)
#  define RED(x)     (((x) >> 16) & 0xFF)
#  define GREEN(x)   (((x) >>  8) & 0xFF)
#  define BLUE(x)     ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(a) << 24) & 0xFF000000) | ((int(r) << 16) & 0xFF0000) | ((int(g) << 8) & 0xFF00) | (int(b) & 0xFF))
#endif

struct eps_settings
{
	BPoint  origin;
	int		anti_alias;
	int		aa_text_only;
	bool	settings_touched;
};
BLocker g_settings_lock ("EPS Settings Lock");
eps_settings g_settings;

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();
status_t ReadEPS (BPositionIO *inSource, BPositionIO *outDestination);
status_t WriteEPS (BPositionIO *inSource, BPositionIO *outDestination);

unsigned char *Device;

class PSView : public BView
{
public:
		PSView (BRect frame);
		virtual void Draw (BRect area);
};

PSView::PSView (BRect frame)
: BView (frame, "GSView", B_FOLLOW_ALL_SIDES, B_WILL_DRAW)
{
}

void PSView::Draw (BRect area)
{
	s_gsdll_lock_device (Device, 1);
	s_gsdll_draw (Device, this, area, area);
	s_gsdll_lock_device (Device, 0);
}

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Anti-Aliasing resample size = %d\n", g_settings.anti_alias);
	printf ("Anti-Alias %s\n", g_settings.aa_text_only ? "text only" : "text and graphics");
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
			 PrefsLoader ()
			 {
			 	// printf ("PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	// defaults:
				g_settings.origin = B_ORIGIN;
				g_settings.anti_alias = 1;
				g_settings.aa_text_only = 0;
				g_settings.settings_touched = false;
				
				BPath path;
				if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
				{
					path.SetTo ("/tmp");
				}
				path.Append ("EPSTranslator_Settings");
				FILE *fp = fopen (path.Path(), "r");
				if (fp)
				{
					char line[1024];
					char name[32];
					char *ptr;
					while (true)
					{
						line[0] = 0;
						fgets (line, 1024, fp);
						if (!line[0])
							break;
						ptr = line;
						while (isspace (*ptr))
							ptr++;
						if (*ptr == '#' || !*ptr)	// Comment line
							continue;
						if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
						{
							fprintf (stderr, "Strange EPSTranslator settings line: %s\n", name);
						}
						else
						{
							if (!strcmp (name, "window_origin"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
									fprintf (stderr, "Illegal window origin in EPS Translator settings: %s\n", ptr);
							}
							else if (!strcmp (name, "resample"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%d", &g_settings.anti_alias) != 1) || (g_settings.anti_alias != 0 && g_settings.anti_alias != 2 && g_settings.anti_alias != 4))
								{
									fprintf (stderr, "Illegal anti-alias resample size in EPS Translator settings: %s\n", ptr);
									g_settings.anti_alias = 0;
								}
							}
							else if (!strcmp (name, "text_only"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%d", &g_settings.aa_text_only) != 1) || (g_settings.aa_text_only < 0) || (g_settings.aa_text_only > 1))
								{
									fprintf (stderr, "Illegal anti-alias type in EPS Translator settings: %s\n", ptr);
									g_settings.aa_text_only = 0;
								}
							}
							else
							{
								fprintf (stderr, "Unknown EPSTranslator setting: %s\n", name);
							}
						}
					}
					fclose (fp);
				}
				g_settings_lock.Unlock();
				// printf ("Done.\n");
			 }
			 
			~PrefsLoader ()
			 {
			 	// printf ("~PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	if (g_settings.settings_touched)
			 	{
			 		BPath path;
			 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
			 		{
			 			path.SetTo ("/tmp");
			 		}
			 		path.Append ("EPSTranslator_Settings");
			 		FILE *fp = fopen (path.Path(), "w");
			 		if (fp)
			 		{
			 			fprintf (fp, "# EPSTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
			 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
			 			fprintf (fp, "resample = %d\n", g_settings.anti_alias);
			 			fprintf (fp, "text_only = %d\n", g_settings.aa_text_only);
			 			fclose (fp);
			 		}
			 	}
			 	g_settings_lock.Unlock();
			 	// printf ("Done.\n");
			 }
};

PrefsLoader g_prefsloader;

BPositionIO *g_in;
BPositionIO *g_out;
int g_h, g_w;
BRect g_frame;
PSView *g_view = 0;
float g_ty, g_by;

int gsdll_callback (int message, char *str, ulong count)
{
//	char *p;
	switch (message)
	{
		case GSDLL_STDIN:
//			printf ("STDIN\n");
//			return g_in->Read (str, count);
//			p = fgets (str, count, stdin);
//			if (p)
//				return (strlen (p));
//			else
//				return 0;
			*str = 13;
			return 1;
				
		case GSDLL_STDOUT:
//			printf ("STDOUT\n");
//			if (str != (char *)NULL)
//				return g_out->Write (str, count);
//				fwrite (str, 1, count, stdout);
			return count;
		
		case GSDLL_DEVICE:
//			printf ("Callback: DEVICE %p %s\n", str, count ? "open" : "close");
			if (count)
				Device = (unsigned char *) str;
			break;
			
		case GSDLL_SYNC:
//			printf ("Callback: SYNC %p\n", str);
		case GSDLL_PAGE:
//			printf ("Callback: PAGE %p\n", str);
			if (g_view && !g_drawn)
			{
//				printf ("Invalidating\n");
				g_view->Invalidate ();
				g_view->Draw (g_frame);
				g_view->Sync();
				g_drawn = true;
			}
			break;
		
		case GSDLL_SIZE:
			g_h = (int) ((count >> 16) & 0xFFFF);
			g_w = (int) (count & 0xFFFF);
//			printf ("Callback: SIZE %p width = %d, height = %d\n", str,
//				g_w, g_h);
			break;
		
		case GSDLL_POLL:
			return 0;	// no error
		
		default:
			printf ("EPSTranslator: Callback: Unknown message %d\n", message);
			break;
	}
	return 0;
}

BRect get_bounding_box ();
status_t get_line (char *buf, BPositionIO *in);
status_t put_line (char *buf, BPositionIO *out);
status_t showpage_hack (char *buf, size_t len);

status_t showpage_hack (char *buf, size_t len)
{
	char *p = buf + len;
	for (int i = 1; i < 40; i++)	// check only last 40 lines
	{
		while (--p > buf && *p != 10 && *p != 13)
			;
		if (!strncmp (p + 1, "showpage", 8))	// Ah!  Found it...
		{
//			printf ("Found showpage, disabling it.\n");
			*(p + 1) = '%';	// disable it.
			return 1;
		}
	}
	return 0;
}

status_t get_line (char *buf, BPositionIO *in)
// read a single, \r or \n delimited line
{
	for (int i = 0; i < 1023; i++)
	{
		if (in->Read (buf + i, 1) != 1)
		{
			return B_ERROR;
		}
		if ((buf[i] == 13) || (buf[i] == 10))
		{
			buf[i + 1] = 0;
			return B_OK;
		}
	}
	return B_ERROR;
}

status_t put_line (char *buf, BPositionIO *out)
// write a single buffer line out
{
	size_t len = strlen (buf);
	if (out->Write (buf, len) != len)
		return B_ERROR;
	return B_OK;
}

BRect get_bounding_box ()
// Warning: In PS coords, i.e. bottom-up!
{
	char line[1024];	// should be enough...
	BRect bb = g_frame;
	for (int i = 0; i < 40; i++)	// Check only the first 40 lines
	{
		if (get_line (line, g_in) == B_OK)
		{
			// puts (line);
			if (!strcasecmp (line, "%%endcomments"))
				return bb;	// Nothing will be found below this
				
			if (!strncasecmp (line, "%%boundingbox", 13))
			{
				char *endp = line + 14;
				bb.left = strtod (endp, &endp);
				g_by = strtod (endp, &endp);
				bb.right = strtod (endp, &endp) - 1;
				g_ty = strtod (endp, &endp);
				bb.top = g_frame.bottom - g_ty;
				// printf ("bb.top = %f - %f = %f\n", g_frame.bottom, g_ty, bb.top);
				bb.bottom = g_frame.bottom - g_by - 1;
				break;
			}
		}
	}
	return bb;
}

uint32 lib_cap ()
{
	status_t err;
	static const int  GSargc = 7;
	uint32 cap = 0;
	char *GSargv[GSargc + 1] =
	{
		"gs",
		"-sDEVICE=beos",
		"-dDEVICEXRESOLUTION=72     ",
		"-dDEVICEYRESOLUTION=72     ",
		"-dNOPAUSE",
		"-dSAFER",
		"-q",
		NULL
	};
	if ((err = s_gsdll_init (gsdll_callback, NULL, GSargc, (char **) GSargv)) != 0)
		return 0;
	s_gsdll_exit();
	cap |= GS_BEOS;
	
	char *GSargvA2[GSargc + 1] =
	{
		"gs",
		"-sDEVICE=bealpha4",
		"-dDEVICEXRESOLUTION=72     ",
		"-dDEVICEYRESOLUTION=72     ",
		"-dNOPAUSE",
		"-dSAFER",
		"-q",
		NULL
	};
	if ((err = s_gsdll_init (gsdll_callback, NULL, GSargc, (char **) GSargvA2)) != 0)
		return cap;
	s_gsdll_exit();
	
	char *GSargvA4[GSargc + 1] =
	{
		"gs",
		"-sDEVICE=bealpha4",
		"-dDEVICEXRESOLUTION=72     ",
		"-dDEVICEYRESOLUTION=72     ",
		"-dNOPAUSE",
		"-dSAFER",
		"-q",
		NULL
	};
	if ((err = s_gsdll_init (gsdll_callback, NULL, GSargc, (char **) GSargvA4)) != 0)
		return cap;
	s_gsdll_exit();
	cap |= GS_ANTI_ALIAS;

	char *GSargvT2[GSargc + 1] =
	{
		"gs",
		"-sDEVICE=bealpht2",
		"-dDEVICEXRESOLUTION=72     ",
		"-dDEVICEYRESOLUTION=72     ",
		"-dNOPAUSE",
		"-dSAFER",
		"-q",
		NULL
	};
	if ((err = s_gsdll_init (gsdll_callback, NULL, GSargc, (char **) GSargvT2)) != 0)
		return cap;
	s_gsdll_exit();

	char *GSargvT4[GSargc + 1] =
	{
		"gs",
		"-sDEVICE=bealpht4",
		"-dDEVICEXRESOLUTION=72     ",
		"-dDEVICEYRESOLUTION=72     ",
		"-dNOPAUSE",
		"-dSAFER",
		"-q",
		NULL
	};
	if ((err = s_gsdll_init (gsdll_callback, NULL, GSargc, (char **) GSargvT4)) != 0)
		return cap;
	s_gsdll_exit();
	cap |= GS_AA_TEXT_ONLY;
	
	return cap;
}

status_t ReadEPS (BPositionIO *inSource, BPositionIO *outDestination)
{
//	printf ("Read EPS\n");

	g_settings_lock.Lock();
	int32 aa = g_settings.anti_alias;
	int32 ato = g_settings.aa_text_only;
	g_settings_lock.Unlock();
	
	static const int  GSargc = 7;
	const char *GSargv[GSargc + 1] =
	{
		"gs",
		aa ? (aa == 2 ? (ato ? "-sDEVICE=bealpht2" : "-sDEVICE=bealpha2") 
					  : (ato ? "-sDEVICE=bealpht4" : "-sDEVICE=bealpha4")) : "-sDEVICE=beos",
		"-dDEVICEXRESOLUTION=96     ",
		"-dDEVICEYRESOLUTION=96     ",
		"-dNOPAUSE",
		"-dSAFER",
		"-q",
		NULL
	};
	
	g_in = inSource;
	g_out = outDestination;
	g_w = 0;
	g_h = 0;
	g_ty = 0;
	g_by = 0;
	
	status_t err;
	if ((err = s_gsdll_init (gsdll_callback, NULL, GSargc, (char **) GSargv)) != 0)
	{
		fprintf (stderr, "EPSTranslator: Can't initialize GhostScript! (err = %ld)\n", err);
		return B_ERROR;
	}
	
	if (!g_h || !g_w)
	{
		fprintf (stderr, "EPSTranslator: Couldn't find image size\n");
		return B_ERROR;
	}
	
	g_frame = BRect (0, 0, g_w - 1, g_h - 1);

//	printf ("Page dimensions:\n");
//	g_frame.PrintToStream();
	
	g_in->Seek (0L, SEEK_SET);	// restore to start of EPS data
	BRect bb = get_bounding_box();
//	bb.PrintToStream();
	uint32 len = g_in->Seek (0L, SEEK_END);	// len now contains the size of the EPS data
	
	int w = bb.IntegerWidth() + 1;
	int h = bb.IntegerHeight() + 1;
	
	g_frame.right = w - 1;
	g_frame.bottom = h - 1;

//	g_frame.PrintToStream();
	g_view = new PSView (g_frame);
	BBitmap *b = new BBitmap (g_frame, B_RGB32, true);
	b->Lock();
	b->AddChild (g_view);
	s_gsdll_execute_begin();
	g_in->Seek (0L, SEEK_SET);
//	printf ("Allocating %ld bytes...\n", len);
	char *EPSbuffer = new char[len + 1];
	if (!EPSbuffer)
		fprintf (stderr, "EPSTranslator: Couldn't allocate %ld bytes!\n", len);
	
	if (g_in->Read (EPSbuffer, len) != len)
		fprintf (stderr, "EPSTranslator: Error reading EPS data - truncated file?");
	
	// Take care of stupid bounding boxes (hello, CorelDRAW!)
	char translate_buf[128];
	sprintf (translate_buf, "%f %f translate\n", -bb.left, bb.top);
//	printf ("Translating by (%f, %f)\n",  -bb.left, bb.top);
	err = s_gsdll_execute_cont (translate_buf, strlen (translate_buf));
	// The rest of the image will now be aligned at a normal (i.e. top left) origin.
	
	// Another thing to take care of is a "showpage" command at the end
	// of the image.  This messes up our purpose as we will always get an
	// empty page back.  Hack: Just overwrite it.
	// Note: This took major head-scratching to find out.
//	showpage_hack (EPSbuffer, len);

	// 19mar2001: Changed to a different (more robust) method of only syncing the first page
	// and flagging the fact it's already drawn.  Makes EPSTranslator work with PS files too
	// (displaying the first page).

	g_drawn = false;
	
#if 1

	// Please don't ask me why the following code works and the
	// version with one call doesn't.
	// What's even more mysterious:  Changing the 32768 below to,
	// say, 65536 breaks it again!
	
	// Sander, Sander.  The call takes an _int_, which is probably defined as a _short_
	// in libgs.so...
	// (But then again, the fact that this still needs flushing is rather strange...)
	
	uint32 offset = 0;
	uint32 tlen = 32767;
	
	while (offset < len)
	{
//		printf ("offset = %ld\n", offset);
		err = s_gsdll_execute_cont (EPSbuffer + offset, tlen < (len - offset) ? tlen : (len - offset));
//		printf ("err = %ld\n", err);
		offset += tlen;
	}

#else

	err = s_gsdll_execute_cont (EPSbuffer, len);
//	printf ("err = %ld\n", err);

#endif

//	printf ("gsdll_execute_end()\n");
	s_gsdll_execute_end();
//	printf ("gsdll_exit()\n");
	s_gsdll_exit();
	b->RemoveChild (g_view);
	delete [] EPSbuffer;
	delete g_view;
	g_view = 0;

	// The bitmap we get from GS has its alpha channel zero, so pass over
	// all the pixels and fix that.
	uint32 *bp = (uint32 *) b->Bits() - 1;
	for (uint32 i = 0; i < h*w; i++)
		*(++bp) |= ALPHA_MASK;

	TranslatorBitmap tbitmap;
	tbitmap.magic    = B_TRANSLATOR_BITMAP;
	tbitmap.bounds   = g_frame;
	tbitmap.rowBytes = w*4;
	tbitmap.colors   = B_RGB_32_BIT;
	tbitmap.dataSize = h*w*4;
	
//	tbitmap.bounds.PrintToStream();
	swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	
	// OK, we have the image in a bitmap now, let's write it out!
//	printf ("SetSize\n");
	err = g_out->SetSize (sizeof (TranslatorBitmap) + w*h*4);
	if (err < 0)
	{
		fprintf (stderr, "EPSTranslatore: SetSize returns %ld\n", err);
		return err;
	}
	// printf ("Writing out the bitmap header\n");
	err = g_out->Write (&tbitmap, sizeof (TranslatorBitmap));
	if (err < 0)
	{
		fprintf (stderr, "EPSTranslator: Write (header) returns %ld\n", err);
		return err;
	}
//	printf ("Writing out the data.\n");
	err = g_out->Write (b->Bits(), w*h*4);
	if (err < 0)
	{
		fprintf (stderr, "EPSTranslator: Write (data) returns %ld\n", err);
		return err;
	}

	delete b;
	return B_OK;
}

status_t WriteEPS (BPositionIO *inSource, BPositionIO *outDestination)
{
	int32 off_x = 0;
	int32 off_y = 0;
	char line[1024];
	// printf ("Writing EPS...\n");

	// Get info on the bitmap
	TranslatorBitmap tbitmap;
	inSource->Seek (0L, SEEK_SET);
	if (inSource->Read (&tbitmap, sizeof (TranslatorBitmap)) < sizeof (TranslatorBitmap))
	{
		fprintf (stderr, "EPSTranslator: Couldn't read header from inStream\n");
		return B_ERROR;
	}
	swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	
	BRect bounds = tbitmap.bounds;
	int32 width  = bounds.IntegerWidth() + 1;
	int32 height = bounds.IntegerHeight() + 1;
	switch (tbitmap.colors)
	{
		case B_RGB32:
		case B_RGBA32:
			break;
		case B_CMAP8:
		case B_GRAY1:
		default:
			fprintf (stderr, "EPSTranslator: Unsupported color_space\n");
			return B_ERROR;
	}

	put_line ("%!PS-Adobe-2.0 EPSF-2.0\n", outDestination);
	sprintf (line, "%%%%Creator: EPSTranslator Version %.2f by Sum Software\n", translatorVersion/100.0);
	put_line (line, outDestination);
	sprintf (line, "%%%%BoundingBox: %ld %ld %ld %ld\n", off_x, off_y, off_x + width, off_y + height);
	put_line (line, outDestination);
	put_line ("%%Pages: 1\n", outDestination);
	put_line ("%%EndComments\n", outDestination);
	put_line ("%%EndProlog\n", outDestination);
	sprintf (line, "%ld %ld translate\n", off_x, off_y + height + 1);
	put_line (line, outDestination);
	put_line ("/SumDict 8 dict def\n", outDestination);
	put_line ("SumDict begin\n", outDestination);
	put_line ("    /ImageType 1 def\n", outDestination);
	sprintf (line, "    /Width %ld def\n    /Height %ld def\n", width, height);
	put_line (line, outDestination);
	put_line ("    /BitsPerComponent 8 def\n", outDestination);
//	sprintf (line, "    /ImageMatrix [%ld 0 0 -%ld 0 %ld] def\n", width, height, height);
	sprintf (line, "    /ImageMatrix [1 0 0 -1 0 0] def\n");
	put_line (line, outDestination);
	put_line ("    /DataSource currentfile /ASCIIHexDecode filter def\n", outDestination);
	put_line ("    /MultipleDataSources false def\n", outDestination);
	put_line ("    /Decode [ 0 1 0 1 0 1 ] def\n", outDestination);
	put_line ("end\n", outDestination);
	put_line ("/DeviceRGB setcolorspace\n", outDestination);
	put_line ("SumDict image\n", outDestination);
	// Image data here
	uint32 pixels[4*4*12];
	for (uint32 i = 0; i <= tbitmap.dataSize/(4*4*12); i++)
	{
		size_t numread = inSource->Read (pixels, 4*4*12);
		uint32 numpix = numread/4;
		if (numpix > 0)
		{
			char hexbuf[6*4*12 + 5];
			for (uint32 j = 0; j < numpix; j++)
				sprintf (hexbuf + 6*j + (j/12), "%02X%02X%02X", int (RED(pixels[j])), int (GREEN(pixels[j])), int (BLUE(pixels[j])));
			
			if (numpix > 12)	// did we read more than one line?
				hexbuf[6*12] = '\n';
			if (numpix > 2*12)	// ...more than two?
				hexbuf[12*12 + 1] = '\n';
			if (numpix > 3*12) // ...more than three?
				hexbuf[18*12 + 2] = '\n';
			
			hexbuf[6*numpix + (numpix - 1)/12] = '\n';
			hexbuf[6*numpix + (numpix - 1)/12 + 1] = 0;
			
			put_line (hexbuf, outDestination);
		}
	}
	put_line ("\nshowpage\n", outDestination);
	put_line ("%%EOF\n", outDestination);

	// printf ("Done.\n");
	return B_OK;
}

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage */*ioExtension*/, translator_info *outInfo, uint32 outType)
{
	inSource->Seek (0L, SEEK_SET);
	char header[4];
	size_t number_read = 0;
	number_read = inSource->Read (header, 4L);
	if (number_read == 4)
	{
		if (!strncmp (header, "%!", 2) || !strncmp (header, "%PDF", 4))
		// We have a PS file in *inSource
		{
			// printf ("Identify: PS detected\n");
			if (outType == EPS_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				extern image_id libgs_image;
				if (libgs_image <= 0 && init_lib() == B_ERROR)
				{
					// printf ("EPS detected but libgs unavailable\n");
					return B_NO_TRANSLATOR;
				}
				outInfo->type		= EPS_CODE;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.6;
				outInfo->capability	= 0.6;	// Let's be modest for now.
				strcpy (outInfo->MIME, "application/postscript");
				strcpy (outInfo->name, "(Encapsulated) PostScript");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource
		{
			if (outType == EPS_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= B_TRANSLATOR_BITMAP;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;
				outInfo->capability	= 0.1;
				strcpy (outInfo->MIME, "image/x-be-bitmap");
				strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		{
		// Sorry, we can't handle this thing.
			// printf ("Hmmm.\n");
			return B_NO_TRANSLATOR;
		}
	}
	return B_NO_TRANSLATOR;
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage */*ioExtension*/, uint32 outType, BPositionIO *outDestination)
{
	inSource->Seek (0L, SEEK_SET);
	char header[4];
	ssize_t number_read = 0;
	number_read = inSource->Read (header, 4L);
	if (number_read == 4)
	{
		if (!strncmp (header, "%!", 2) || !strncmp (header, "%PDF", 4))
		// We have a PS file in *inSource
		{
			// printf ("Translate: PS detected\n");
			if (outType == EPS_CODE)
			// "translate" it to EPS...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// decode the EPS to a bitmap
			{
				extern image_id libgs_image;
				if (libgs_image <= 0 && init_lib() == B_ERROR)
				{
					printf ("EPS detected but libgs unavailable\n");
					return B_NO_TRANSLATOR;
				}
				outType = B_TRANSLATOR_BITMAP;
				status_t err = ReadEPS (inSource, outDestination);
				return err;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == EPS_CODE)
			// encode the bitmap to EPS...
			{
				status_t err = WriteEPS (inSource, outDestination);
				return err;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// "translate" the bitmap...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

class EPSConfigView : public BView
{
public:
			 EPSConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
				SetViewColor (ui_color (B_PANEL_BACKGROUND_COLOR));
				if (init_lib() == B_OK)
				{
					if (!s_lib_cap)
						s_lib_cap = lib_cap();
				}
				else
					s_lib_cap = 0;
						
				g_settings_lock.Lock();
				bool aa = false;
				if (s_lib_cap & GS_ANTI_ALIAS)
				{
					aaPU = new BPopUpMenu ("Anti-Alias");
					aaPU->AddItem (new BMenuItem ("None", new BMessage ('aaa0')));
					aaPU->AddItem (new BMenuItem ("2x2 Resampling", new BMessage ('aaa2')));
					aaPU->AddItem (new BMenuItem ("4x4 Resampling", new BMessage ('aaa4')));
					aaMF = new BMenuField (BRect (8, 56, 192, 80), "aaMF", "Anti-Aliasing", aaPU);
					aaMF->SetDivider (70);
					
					switch (g_settings.anti_alias)
					{
					case 2:
						aaPU->ItemAt(1)->SetMarked (true);
						aa = true;
						break;
					case 4:
						aaPU->ItemAt(2)->SetMarked (true);
						aa = true;
						break;
					default:
						aaPU->ItemAt(0)->SetMarked (true);
						break;
					}
					AddChild (aaMF);
				}
				else
				{
					aaPU = 0;
					aaMF = 0;
				}
				if (s_lib_cap & GS_AA_TEXT_ONLY)
				{
					aaCB = new BCheckBox (BRect (8, 82, 192, 106), "aaCB", "Anti-alias text only", new BMessage ('aaTO'));
					aaCB->SetEnabled (aa);
					aaCB->SetValue (g_settings.aa_text_only);
					AddChild (aaCB);
				}
				else
					aaCB = 0;
				g_settings_lock.Unlock();
			 }

virtual		~EPSConfigView ()
			 {
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
			 	BMessenger messenger (this);
			 	if (aaPU)
			 	{
					for (int i = 0; i < aaPU->CountItems(); i++)
				 	{
				 		BMenuItem *item = aaPU->ItemAt (i);
				 		if (item)
				 			item->SetTarget (messenger);
				 	}
				}
				if (aaCB)
					aaCB->SetTarget (this);
			 }

virtual void Draw (BRect update)
			 {
			 	extern image_id libgs_image;
			 	SetFont (be_bold_font);
			 	SetDrawingMode (B_OP_OVER);
			 	DrawString ("EPS Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 1999 - 2001 ∑ Sum Software", BPoint (8, 48));
			 	if ((libgs_image > 0) && s_lib_cap)
			 	{
					// DrawString ("libgs.so detected", BPoint (8, 98));
					// DrawString ("(E)PS input available", BPoint (8, 114));
			 	}
			 	else
			 	{
					DrawString ("libgs.so not detected", BPoint (8, 98));
					DrawString ("(E)PS input not available", BPoint (8, 114));
				}
			 	BView::Draw (update);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
			 	case 'aaa0':
			 		aaCB->SetEnabled (false);
			 		g_settings_lock.Lock();
			 		g_settings.anti_alias = 0;
					g_settings.settings_touched = true;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'aaa2':
			 		aaCB->SetEnabled (true);
			 		g_settings_lock.Lock();
			 		g_settings.anti_alias = 2;
					g_settings.settings_touched = true;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'aaa4':
			 		aaCB->SetEnabled (true);
			 		g_settings_lock.Lock();
			 		g_settings.anti_alias = 4;
					g_settings.settings_touched = true;
			 		g_settings_lock.Unlock();
			 		break;
			 	
			 	case 'aaTO':
			 		g_settings_lock.Lock();
			 		g_settings.aa_text_only = aaCB->Value();
					g_settings.settings_touched = true;
			 		g_settings_lock.Unlock();
			 		break;
			 		
				default:
					inherited::MessageReceived (msg);
				}
			 }
			 
private:
typedef BView inherited;

BCheckBox	*aaCB;
BMenuField	*aaMF;
BPopUpMenu	*aaPU;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new EPSConfigView (BRect (0, 0, 200, 122), "EPSTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}

char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "EPS Translator";

int32 translatorVersion = 31;

translation_format inputFormats[] = {
	{ EPS_CODE, B_TRANSLATOR_BITMAP, 0.6, 0.6, "application/postscript", "(Encapsulated) PostScript" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ EPS_CODE, B_TRANSLATOR_BITMAP, 0.6, 0.6, "application/postscript", "(Encapsulated) PostScript" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
