#include <stdio.h>
#include <Window.h>
#include <Alert.h>
#include <Screen.h>
#include <Application.h>
#include <TranslatorAddOn.h>
#include <View.h>

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

class GIFWindow : public BWindow
{
public:
			 GIFWindow (BRect frame)
			 : BWindow (frame, "GIFTranslator", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
			 {
			 }
			 
			~GIFWindow ()
			 {
			 	// printf ("~GIFWindow... "); fflush (stdout);
			 	BPoint origin (0, 0);
			 	ConvertToScreen (&origin);
			 	set_window_origin (origin);
			 	be_app->PostMessage (B_QUIT_REQUESTED);
			 	// printf ("Done.\n");
			 }
};

int main (int argc, char *argv[])
{
	if (argc > 1)
	{
		bool sensible = false;
		for (int i = 1; i < argc; i++)
		{
			if (!strcmp (argv[i], "-v") || !strcmp (argv[i], "--version"))
			{
				// extern int32 translatorVersion;
				// extern char translatorName[];
				printf ("%s version %f\n", translatorName, (float) translatorVersion/100.0);
				sensible = true;
			}
			else if (!strcmp (argv[i], "-i") || !strcmp (argv[i], "--info"))
			{
				// extern char translatorName[];
				// extern char translatorInfo[];
				printf ("%s\n%s\n", translatorName, translatorInfo);
				sensible = true;
			}
			else if (!strcmp (argv[i], "-s") || !strcmp (argv[i], "--settings"))
			{
				show_settings();
				sensible = true;
			}
			else
			{
				printf ("Usage: %s [options]\n", argv[0]);
				printf ("    -v, --version   Print version information and exit\n");
				printf ("    -i, --info      Print copyright information and exit\n");
				printf ("    -s, --settings  Print current settings and exit\n");
				printf ("    -h, --help      Print this text and exit\n");
				printf ("    <none>          Open Settings Window\n");
			}
		}
	}
	else
	{
		BApplication app ("application/x-vnd.sum-GIFTranslator");
		BView *view = NULL;
		BRect rect (0, 0, 200, 62);
		if (MakeConfig (NULL, &view, &rect))
		{
			(new BAlert ("Error", "Something is wrong with the GIFTranslator!", "OK"))->Go();
			return B_ERROR;
		}
		GIFWindow *win = new GIFWindow (rect);
		view->ResizeTo (rect.Width(), rect.Height());
		win->AddChild (view);
		BPoint origin = get_window_origin();
		{
			BScreen screen;
			BRect frame = screen.Frame();
			frame.InsetBy (10, 23);
			if (!frame.Contains (origin))
			{
				uint32 buttons;
				view->GetMouse (&origin, &buttons);
				origin.x -= rect.Width()/2;
				origin.y -= rect.Height()/2;
				origin.x = max_c (origin.x, frame.left);
				origin.y = max_c (origin.y, frame.top);
				origin.x = min_c (origin.x, frame.right);
				origin.y = min_c (origin.y, frame.bottom);
			}
		}
		win->MoveTo (origin);
		win->Show();
		app.Run();
	}
	return B_NO_ERROR;
}