#include <Locker.h>
#include <Path.h>
#include <Entry.h>
#include <FindDirectory.h>
#include <View.h>
#include <Screen.h>
#include <GraphicsDefs.h>
#include <InterfaceDefs.h>
#include <Slider.h>
#include <Button.h>
#include <Message.h>
#include <StringView.h>
#include <Roster.h>
#include <Alert.h>
#include <stdio.h>
#include <Bitmap.h>
#include <Autolock.h>
#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>

#define BUF_SIZE 4096
#define GIF_CODE 'GIF '
#define MAX_LZW_BITS 12
#define MAX_COLORMAP 256
#define INTERLACE  0x40
#define LOCAL_CMAP 0x80
#define bit(byte, bit) (((byte) & (bit)) == (bit))
#define MAXCODE(n_bits)	((1 << (n_bits)) - 1)

#if defined (__POWERPC__)
#  define COLOR_MASK	0xFFFFFF00
#  define ALPHA_MASK	0x000000FF
#  define BLUE(x)     ((x) >> 24)
#  define GREEN(x)   (((x) >> 16) & 0xFF)
#  define RED(x)     (((x) >>  8) & 0xFF)
#  define ALPHA(x)    ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(b) << 24) & 0xFF000000) | ((int(g) << 16) & 0xFF0000) | ((int(r) << 8) & 0xFF00) | (int(a) & 0xFF))
#else	// IA
#  define COLOR_MASK	0x00FFFFFF
#  define ALPHA_MASK	0xFF000000
#  define ALPHA(x)    ((x) >> 24)
#  define RED(x)     (((x) >> 16) & 0xFF)
#  define GREEN(x)   (((x) >>  8) & 0xFF)
#  define BLUE(x)     ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(a) << 24) & 0xFF000000) | ((int(r) << 16) & 0xFF0000) | ((int(g) << 8) & 0xFF00) | (int(b) & 0xFF))
#endif

struct gif_settings
{
	BPoint  origin;
	int		threshold;
	bool	settings_touched;
};
BLocker g_settings_lock ("GIF Settings Lock");
BLocker g_lock ("GIF Serialization Lock");
gif_settings g_settings;

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Transpacency Threshold [1..255] = %d\n", g_settings.threshold);
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
			 PrefsLoader ()
			 {
			 	// printf ("PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	// defaults:
				g_settings.origin = B_ORIGIN;
				g_settings.threshold = 255;
				g_settings.settings_touched = false;
				
				BPath path;
				if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
				{
					path.SetTo ("/tmp");
				}
				path.Append ("GIFTranslator_Settings");
				FILE *fp = fopen (path.Path(), "r");
				if (fp)
				{
					char line[1024];
					char name[32];
					char *ptr;
					while (true)
					{
						line[0] = 0;
						fgets (line, 1024, fp);
						if (!line[0])
							break;
						ptr = line;
						while (isspace (*ptr))
							ptr++;
						if (*ptr == '#' || !*ptr)	// Comment line
							continue;
						if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
						{
							fprintf (stderr, "Strange GIFTranslator settings line: %s\n", name);
						}
						else
						{
							if (!strcmp (name, "window_origin"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
									fprintf (stderr, "Illegal window origin in GIF Translator settings: %s\n", ptr);
							}
							else if (!strcmp (name, "threshold"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%d", &g_settings.threshold) != 1) || (g_settings.threshold < 1) || (g_settings.threshold > 255))
								{
									fprintf (stderr, "Illegal transparency threshold in GIF Translator settings: %s\n", ptr);
									g_settings.threshold = 255;
								}
							}
							else
							{
								fprintf (stderr, "Unknown GIFTranslator setting: %s\n", name);
							}
						}
					}
					fclose (fp);
				}
				g_settings_lock.Unlock();
				// printf ("Done.\n");
			 }
			 
			~PrefsLoader ()
			 {
			 	// printf ("~PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	if (g_settings.settings_touched)
			 	{
			 		BPath path;
			 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
			 		{
			 			path.SetTo ("/tmp");
			 		}
			 		path.Append ("GIFTranslator_Settings");
			 		FILE *fp = fopen (path.Path(), "w");
			 		if (fp)
			 		{
			 			fprintf (fp, "# GIFTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
			 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
			 			fprintf (fp, "threshold = %d\n", g_settings.threshold);
			 			fclose (fp);
			 		}
			 	}
			 	g_settings_lock.Unlock();
			 	// printf ("Done.\n");
			 }
};

PrefsLoader g_prefsloader;


class GIFDecoder
{
public:
				 GIFDecoder (BPositionIO *inSource, BPositionIO *outDestination);
	status_t	 Decode (uint32 outType);

private:
	int			 read_cmap (int n, uint32 cmap[MAX_COLORMAP]);
	int			 read_lzw (bool flag, int input_code_size);
	int			 get_code (int code_size, bool flag);
	int			 get_data_block (uint8 *buffer);
	int			 extension (int type);

	BPositionIO	*in, *out;
	bool		 fZeroDB;
	uint8		 fExt[256];
	uint8		 fGetBuf[256];
	int			 curbit, lastbit, done, last_byte;
	int			 fresh;
	int			 code_size, set_code_size;
	int			 max_code, max_code_size;
	int			 firstcode, oldcode;
	int			 clear_code, end_code;
	int			 table[2][(1 << MAX_LZW_BITS)];
	int			 stack[(1 << (MAX_LZW_BITS))*2], *sp;
	int			 transparentIndex;
};

class GIFEncoder
{
public:
				 GIFEncoder (BPositionIO *inSource, BPositionIO *outDestination);
	status_t	 Encode ();

private:
	void		 flush_datablock ();
	void		 output (int code);
	void		 init (int bits);
	int			 index_for (uint32 pixel);
	void		 compress_pixel (int c);
	int			 construct_colormap (BBitmap *image);
	
	BPositionIO	*in, *out;
	int			 n_bits;	// current number of bits/code
	int			 maxcode;	// maximum code, given n_bits
	uint32		 cur_accum;	// holds bits not yet output
	int			 cur_bits;	// # of bits in cur_accum
	int			 code_counter;
	int			 ClearCode;
	int			 EOFCode;
	int			 num_colors;
	uint32		 color_map[MAX_COLORMAP];
	uint8		 threshold;
	int			 transparentIndex;
	uint8		 packet[256];
	int			 bytesinpacket;
};

GIFDecoder::GIFDecoder (BPositionIO *inSource, BPositionIO *outDestination)
: in (inSource), out (outDestination)
{
	curbit = lastbit = done = last_byte = 0;
	fresh = 0;
	code_size = set_code_size = 0;
	max_code = max_code_size = 0;
	firstcode = oldcode = 0;
	clear_code = end_code = 0;
}

int GIFDecoder::get_data_block (unsigned char *buf)
{
	unsigned char   count;
	
	if (in->Read (&count, 1) != 1)
	{
		fprintf (stderr, "GIFTranslator: Error in getting data block size (truncated file?)\n");
		return -1;
	}

	fZeroDB = (count == 0);

	if ((count != 0) && (in->Read (buf, count) != count))
	{
		fprintf (stderr, "GIFTranslator: Error in reading data block (truncated file?)\n");
		return -1;
	}

	return count;
}

int GIFDecoder::get_code (int code_size, bool flag)
{
	int             i, j, ret;
	unsigned char   count;

	if (flag)
	{
		curbit = 0;
		lastbit = 0;
		done = FALSE;
		return 0;
	}

	if ((curbit + code_size) >= lastbit)
	{
		if (done)
		{
			if (curbit >= lastbit)
				fprintf (stderr, "Error while reading lzw code - spilled over bits");
			return -1;
		}
		fGetBuf[0] = fGetBuf[last_byte - 2];
		fGetBuf[1] = fGetBuf[last_byte - 1];

		if ((count = get_data_block (&fGetBuf[2])) <= 0)
			done = true;

		last_byte = 2 + count;
		curbit = (curbit - lastbit) + 16;
		lastbit = (2+count)*8 ;
	}

	ret = 0;
	for (i = curbit, j = 0; j < code_size; ++i, ++j)
		ret |= ((fGetBuf[ i / 8 ] & (1 << (i % 8))) != 0) << j;

	curbit += code_size;

	return ret;
}

int GIFDecoder::read_lzw (bool flag, int input_code_size)
{
	int             code, incode;
	register int    i;

	if (flag)
	{
		set_code_size = input_code_size;
		code_size = set_code_size + 1;
		clear_code = 1 << set_code_size;
		end_code = clear_code + 1;
		max_code_size = 2*clear_code;
		max_code = clear_code + 2;

		get_code (0, true);
		fresh = true;

		for (i = 0; i < clear_code; ++i)
		{
			table[0][i] = 0;
			table[1][i] = i;
		}
		for (; i < (1 << MAX_LZW_BITS); ++i)
			table[0][i] = table[1][0] = 0;

		sp = stack;

		return 0;
	}
	else if (fresh)
	{
		fresh = false;
		do
		{
			firstcode = oldcode = get_code (code_size, false);
		} while (firstcode == clear_code);
		return firstcode;
	}

	if (sp > stack)
		return *--sp;

	while ((code = get_code (code_size, false)) >= 0)
	{
		if (code == clear_code)
		{
			for (i = 0; i < clear_code; ++i)
			{
				table[0][i] = 0;
				table[1][i] = i;
			}
			for (; i < (1 << MAX_LZW_BITS); ++i)
				table[0][i] = table[1][i] = 0;
			
			code_size = set_code_size + 1;
			max_code_size = 2*clear_code;
			max_code = clear_code + 2;
			sp = stack;
			firstcode = oldcode = get_code (code_size, false);
			return firstcode;
		}
		else if (code == end_code)
		{
			int             count;
			unsigned char   buf[260];

			if (fZeroDB)
				return -2;

			while ((count = get_data_block (buf)) > 0)
				;

			if (count != 0)
				fprintf (stderr, "GIFTranslator warning: EOD missing\n");
			return -2;
		}

		incode = code;

		if (code >= max_code)
		{
			*sp++ = firstcode;
			code = oldcode;
		}

		while (code >= clear_code)
		{
			*sp++ = table[1][code];
			if (code == table[0][code])
			{
				fprintf (stderr, "GIFTranslator error: circular table entry\n");
				break;
			}
			code = table[0][code];
		}

		*sp++ = firstcode = table[1][code];

		if ((code = max_code) < (1 << MAX_LZW_BITS))
		{
			table[0][code] = oldcode;
			table[1][code] = firstcode;
			++max_code;
			if ((max_code >= max_code_size) && (max_code_size < (1 << MAX_LZW_BITS)))
			{
				max_code_size *= 2;
				++code_size;
			}
		}

		oldcode = incode;

		if (sp > stack)
			return *--sp;
	}
	return code;
}

int GIFDecoder::extension (int label)
{
	char *str = NULL;

	struct {
		int     transparent;
		int     delayTime;
		int     inputFlag;
		int     disposal;
	} GIF89 = { -1, -1, -1, 0 };


	switch (label) {
	case 0x01:              /* Plain Text Extension */
		str = "Plain Text Extension";
		break;

	case 0xFF:              /* Application Extension */
		str = "Application Extension";
		break;

	case 0xFE:              /* Comment Extension */
		str = "Comment Extension";
		while (get_data_block ((unsigned char*) fExt) > 0)
		{
			fprintf (stderr, "GIFTranslator: GIF comment: %s\n", fExt);
		}
		break;

	case 0xF9:              /* Graphic Control Extension */
		str = "Graphic Control Extension";
		get_data_block ((unsigned char*) fExt);

		GIF89.disposal    = (fExt[0] >> 2) & 0x7;
		GIF89.inputFlag   = (fExt[0] >> 1) & 0x1;
		GIF89.delayTime   = fExt[1] + 256*fExt[2];
		if (fExt[0] & 0x1)
		{
			GIF89.transparent = fExt[3];
			//printf ("Transparent: %d\n", GIF89.transparent);
			transparentIndex = GIF89.transparent;
		}
		break;

	default:
		str = (char *) fExt;
		sprintf ((char *) fExt, "UNKNOWN (0x%02x)\n", label);
		break;
	}

//	fprintf (stderr, "GIFTranslator: Got a '%s' extension\n", str);

//	while (get_data_block ((unsigned char*) fExt) > 0)
//		;

	return 0;
}

status_t GIFDecoder::Decode (uint32 outType)
{
	uint16 width, height;
	int bitp, colorres, bg, aspect;
	uint32 global_colormap[MAX_COLORMAP];
	uint32 local_colormap[MAX_COLORMAP];
	uint32 *colormap;
	
	BAutolock lock (&g_lock);
	
	transparentIndex = -1;
	
	unsigned char screen[7];
	if (in->Read (screen, 7L) != 7)
	{
		fprintf (stderr, "GIFTranslator: Error reading screen descriptor (truncated file?)\n");
		return B_ERROR;
	}
	
	width	 = screen[0] + 256*screen[1];
	height	 = screen[2] + 256*screen[3];
	bitp	 = 2 << (screen[4] & 0x07);
	colorres = ((screen[4] & 0x70) >> 3) + 1;
	bg		 = screen[5];
	aspect	 = screen[6];

//	printf ("w = %d, h = %d, bitp = %d, colorres = %d, bg = %d, aspect = %d\n",
//		width, height, bitp, colorres, bg, aspect);

	TranslatorBitmap tbitmap;
	tbitmap.magic    = B_TRANSLATOR_BITMAP;
	tbitmap.bounds   = BRect (0, 0, width - 1, height - 1);
	tbitmap.rowBytes = width*4;
	tbitmap.colors   = B_RGB_32_BIT;
	tbitmap.dataSize = height*width*4;
	
//	tbitmap.bounds.PrintToStream();
	swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	
	int32 err;
	err = out->SetSize (sizeof (TranslatorBitmap) + width*height*4);
	if (err < 0)
	{
		fprintf (stderr, "GIFTranslatore: SetSize returns %ld\n", err);
		return err;
	}
	err = out->Write (&tbitmap, sizeof (TranslatorBitmap));
	if (err < 0)
	{
		fprintf (stderr, "GIFTranslator: Write (header) returns %ld\n", err);
		return err;
	}

	if (aspect != 0 && aspect != 49)
	{
		float ratio = ((float) aspect + 15)/64.0;
		fprintf (stderr, "GIFTranslator warning:  Non-square pixels (aspect ratio %f)\n", ratio);
	}
	
	if (bit (screen[4], LOCAL_CMAP))	// Global color map
	{
//		printf ("Global color map, %d colors\n", bitp);
		uint8 rgb[3];
		for (int i = 0; i < bitp; i++)
		{
			if (in->Read (rgb, 3) != 3)
			{
				fprintf (stderr, "GIFTranslator error: Bogus global color map\n");
				return B_ERROR;
			}
			
			global_colormap[i] = PIXEL (rgb[0], rgb[1], rgb[2], 255);
		}
	}
	
	// Read in the image
	int iNum = 1, iCount = 0;
	
	for (;;)
	{
		uint8 c;
		if (in->Read (&c, 1L) != 1)
		{
			fprintf (stderr, "GIFTranslator error: Unexpected EOF - truncated file?\n");
			return B_ERROR;
		}
		
		if (c == ';')	// Terminator
		{
			if (iCount < iNum)
			{
				fprintf (stderr, "GIFTranslator warning: Only %d image%s found\n", iCount, (iCount != 1) ? "s" : "");
			}
			return B_OK;
		}
		if (c == '!')	// Extension
		{
			if (in->Read (&c, 1) != 1)
			{
				fprintf (stderr, "GIFTranslator: error reading GIF extension function code\n");
			}
			extension (c);
			continue;
		}
		if (c != ',')	// Should be start char
		{
//			fprintf (stderr, "GIFTranslator warning: Ignoring bogus start char 0x%02x\n", (int) c);
			continue;
		}
		
		iCount++;
		
		unsigned char buffer[16];
		if (in->Read (buffer, 9) != 9)
		{
			fprintf (stderr, "GIFTranslator error: Unexpected EOF - truncated file?\n");
			return B_ERROR;
		}
		
		bool use_local_cmap = bit (buffer[8], LOCAL_CMAP);
		if (use_local_cmap)
		{
			uint8 rgb[3];
			for (int i = 0; i < bitp; i++)
			{
				if (in->Read (rgb, 3) != 3)
				{
					fprintf (stderr, "GIFTranslator error: Bogus local color map\n");
					return B_ERROR;
				}
				
				local_colormap[i] = PIXEL (rgb[0], rgb[1], rgb[2], 255);
			}
			colormap = local_colormap;
		}
		else
		{
			colormap = global_colormap;
		}
		
		if (transparentIndex != -1)
		{
			colormap[transparentIndex] = PIXEL (255, 255, 255, 0);
		}
		
		int l = buffer[0] + 256*buffer[1];
		int t = buffer[2] + 256*buffer[3];
		int w = buffer[4] + 256*buffer[5];
		int h = buffer[6] + 256*buffer[7];
		int interlace = bit (buffer[8], INTERLACE);
		int ign = (iCount != iNum) && (outType == B_TRANSLATOR_BITMAP);
		
		int v;
		int x = l, y = t, pass =0;
		
		// Init decompression routines
		if (in->Read (&c, 1) != 1)
			fprintf (stderr, "GIFTranslator error: Unexpected EOF\n");
		
		if (read_lzw (true, c) < 0)
			fprintf (stderr, "GIFTranslator error reading image\n");
		
		if (ign)	// Ignore this image
		{
			fprintf (stderr, "GIFTranslator warning: Skipping image\n");
			while (read_lzw (false, c) >= 0)
				;
		}
		else
		{
			// fprintf (stderr, "GIFTranslator: reading %d by %d%s GIF image, offset %d, %d\n", w, h, interlace ? " interlaced" : "", l, t);
			uint32 *row_buffer = NULL;
			
			long buf_size = w*4;
			if (interlace)
				buf_size *= 8;
			if ((row_buffer = (uint32 *) malloc (buf_size)) == NULL)
			{
				fprintf (stderr, "GIFTranslator error: out of memory\n");
				return B_ERROR;
			}
			memset (row_buffer, 0, buf_size);

			if (!interlace)
				pass = 3;
	
			uint32 *in_row = row_buffer - 1;
			while ((v = read_lzw (false, c)) >= 0 )
			{
				if (v == transparentIndex)
					*(++in_row) = COLOR_MASK;
				else
					*(++in_row) = colormap[v];
					
				x++;
				if (x == l + w)
				{
					if (interlace)
						out->Seek (sizeof (TranslatorBitmap) + y*4*w, SEEK_SET);
						
					if (out->Write (row_buffer, w*4) < B_OK)
					{
						fprintf (stderr, "GIFTranslator: Error writing out bitmap\n");
						return B_ERROR;
					}

					in_row = row_buffer - 1;
					x = l;
					if (interlace)
					{
						switch (pass)
						{
						case 0:
						case 1:
							y += 8;
							break;
						case 2:
							y += 4;
							break;
						case 3:
							y += 2;
							break;
						}
	
						if (y >= t + h)
						{
							pass++;
							switch (pass)
							{
							case 1:
								y = t + 4;
								break;
							case 2:
								y = t + 2;
								break;
							case 3:
								y = t + 1;
								break;
							default:
								goto end;
							}
						}
					}
					else
					{
						y++;
					}
				}
				if (y == t + h)
					break;
			}

end:
			if (read_lzw (false, c) >= 0)
				fprintf (stderr, "GIFTranslator warning: trailing data ignored\n");
				
			free (row_buffer);
		}
	}
	
	return B_OK;
}


/* The pseudo-compression algorithm.
 *
 * In this module we simply output each pixel value as a separate symbol;
 * thus, no compression occurs.  In fact, there is expansion of one bit per
 * pixel, because we use a symbol width one bit wider than the pixel width.
 *
 * GIF ordinarily uses variable-width symbols, and the decoder will expect
 * to ratchet up the symbol width after a fixed number of symbols.
 * To simplify the logic and keep the expansion penalty down, we emit a
 * GIF Clear code to reset the decoder just before the width would ratchet up.
 * Thus, all the symbols in the output file will have the same bit width.
 * Note that emitting the Clear codes at the right times is a mere matter of
 * counting output symbols and is in no way dependent on the LZW patent.
 *
 * With a small basic pixel width (low color count), Clear codes will be
 * needed very frequently, causing the file to expand even more.  So this
 * simplistic approach wouldn't work too well on bilevel images, for example.
 * But for output of JPEG conversions the pixel width will usually be 8 bits
 * (129 to 256 colors), so the overhead added by Clear symbols is only about
 * one symbol in every 256.
 *
 * This code is heavily based on wrgif.c from the jpeg-6b distribution.
 */

GIFEncoder::GIFEncoder (BPositionIO *inSource, BPositionIO *outDestination)
: in (inSource), out (outDestination)
{
	cur_bits = 0;
	cur_accum = 0;
	bytesinpacket = 0;
}

#define char_out(c) \
{ packet[++bytesinpacket] = (uint8) (c);\
  if (bytesinpacket >= 255) flush_datablock(); }

void GIFEncoder::flush_datablock ()
{	
// A GIF data block consists of 1..255 bytes, preceded by its size.
	if (bytesinpacket > 0)		// Don't write out zero length data blocks
	{
		packet[0] = bytesinpacket++;
		if (out->Write (packet, bytesinpacket) != bytesinpacket)
		{
			fprintf (stderr, "GIFTranslator error while writing data block (disk full?)\n");
		}
		bytesinpacket = 0;
	}
}

void GIFEncoder::output (int code)
{
	cur_accum |= ((uint32) code) << cur_bits;
	cur_bits += n_bits;
	
	while (cur_bits >= 8)
	{
		char_out (cur_accum & 0xFF);
		cur_accum >>= 8;
		cur_bits -= 8;
	}
}

void GIFEncoder::init (int bits)
{
	n_bits = bits;
	maxcode = MAXCODE (bits);
	ClearCode = (1 << (bits - 1));
	EOFCode = ClearCode + 1;
	code_counter = ClearCode + 2;
	
	// GIF Specifies an initial Clear code
	output (ClearCode);
}

void GIFEncoder::compress_pixel (int c)
// "compress" one given pixel; must be less then n_bits wide.
{
	// write the pixel as a symbol;
	output (c);
	// Issue Clear codes often enough to keep the reader from ratcheting up its symbol size
	
	if (code_counter < maxcode)
		code_counter++;
	else
	{
		output (ClearCode);
		code_counter = ClearCode + 2;	// reset the counter;
	}
}

int GIFEncoder::index_for (uint32 pixel)
{
	if (ALPHA (pixel) < threshold)
		return transparentIndex;
		
	for (int i = 0; i < num_colors; i++)
	{
		if ((pixel & COLOR_MASK) == color_map[i])
			return i;
	}
	return 0;	// We shouldn't get here..!
}

int GIFEncoder::construct_colormap (BBitmap *image)
{
	int count = 0;
	bool needtransparent = false;
	
	uint32 *s = (uint32 *) image->Bits();

	for (int32 i = image->BitsLength()/4; i > 0; i--)
	{
		uint32 pixel = *(s++);
		if (ALPHA (pixel) < threshold)
			needtransparent = true;
		else
		{
			uint32 pc = pixel & COLOR_MASK;
			int j = 0;
			while (pc != color_map[j] && j < count)
				j++;
			if (j == count)	// We have a new color
			{
				color_map[count] = pc;
				count++;
				if (count > MAX_COLORMAP)
					return count;
			}
		}
	}
	if (needtransparent)
	{
		if (count < 255)
			transparentIndex = count++;
		else
			transparentIndex = 255;		// Note that we clobber one color now!
	}
	
//	printf ("Transparent Index = %d\n", transparentIndex);
	return count;
}

status_t GIFEncoder::Encode ()
{
	g_settings_lock.Lock();
	threshold = g_settings.threshold;
	g_settings_lock.Unlock();
	transparentIndex = -1;
	
	// Get info on the bitmap
	TranslatorBitmap tbitmap;
	in->Seek (0L, SEEK_SET);
	if (in->Read (&tbitmap, sizeof (TranslatorBitmap)) < sizeof (TranslatorBitmap))
	{
		fprintf (stderr, "GIFTranslator: Couldn't read header from inStream\n");
		return B_ERROR;
	}
	swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
	swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
	
	BRect bounds = tbitmap.bounds;

	BBitmap *src = new BBitmap (bounds, B_RGB32);

	uint16 width  = bounds.IntegerWidth() + 1;
	uint16 height = bounds.IntegerHeight() + 1;
	switch (tbitmap.colors)
	{
		case B_RGB32:
		case B_RGBA32:
		{
			if (in->Read (src->Bits(), tbitmap.dataSize) != tbitmap.dataSize)
			{
				fprintf (stderr, "GIFTranslator:  Error reading bitmap data (stream truncated?)\n");
				return B_ERROR;
			}
			break;
		}
		default:
			fprintf (stderr, "GIFTranslator: Unsupported color_space (only 32bit)\n");
			return B_NO_TRANSLATOR;
	}
	
	num_colors = construct_colormap (src);
	if (num_colors > 256)	// We have to quantize...
	{
//		printf ("Quantize!\n");
		BBitmap *src8 = new BBitmap (bounds, B_CMAP8, true);
		src8->Lock();
		BView *v = new BView (bounds, "tmp cq view", 0, 0);
		src8->AddChild (v);
		v->DrawBitmap (src);	// Use the BeOS to render down to 256 colors
		src8->RemoveChild (v);
		delete v;
		uint32 *dst = (uint32 *) src->Bits() - 1;
		const rgb_color *spalette = system_colors()->color_list;
		for (int i = 0; i < 256; i++)
		{
			color_map[i] = PIXEL (spalette[i].red, spalette[i].green, spalette[i].blue, 0);
		}
		num_colors = 256;
		for (int y = 0; y < height; y++)
		{
			uint8 *row8 = (uint8 *) src8->Bits() + y*src8->BytesPerRow() - 1;
			for (int x = 0; x < width; x++)
			{
				uint32 pixel = *(++dst);
				if (ALPHA (pixel) < threshold)
				{
					*dst = (color_map[255]);
					transparentIndex = 255;
					row8++;
				}
				else
					*dst = (color_map[*(++row8)] & COLOR_MASK) | (pixel & ALPHA_MASK);
				// i.e. preserve original alpha
			}
		}
		delete src8;
	}
	
	int bpp = 1;
	while (num_colors > (1 << bpp))
		bpp++;
	
	int cms = 1 << bpp;
	int ics;
	if (bpp <= 1)
		ics = 2;
	else
		ics = bpp;
	
	
	// Time to write the GIF header.
//	printf ("Writing out the hearder\n");
	out->Write (&"GIF89a", 6L);
	
	// Logical Screen Descriptor
	uint8 buf[16];
	buf[0] = width%256;
	buf[1] = width/256;
	buf[2] = height%256;
	buf[3] = height/256;
	buf[4] = 0x80;				// Global color map
	buf[4] |= (bpp - 1) << 4;	// Color resolution
	buf[4] |= (bpp - 1);		// size of global color map
	buf[5] = 0;					// Background color index
	buf[6] = 0;					// Aspect ratio
	out->Write (buf, 7);

	// Write out the color map
//	printf ("Writing out the color map... "); fflush (stdout);
	uint8 rgb[3];
	for (int i = 0; i < cms; i++)
	{
		if (i < num_colors)
		{
			rgb[0] = RED (color_map[i]);
			rgb[1] = GREEN (color_map[i]);
			rgb[2] = BLUE (color_map[i]);
			out->Write (rgb, 3);
		}
		else	// Fill out the color map to a power of 2
		{
			rgb[0] = rgb[1] = rgb[2] = 0;
			out->Write (rgb, 3);
		}
	}
//	printf ("Done.\n");

	init (ics + 1);
//	printf ("Initialized fake compressor\n");
	
	if (transparentIndex != -1)	// Transparency in image
	{
		// Write out a Graphic Control Extension block with the transparency info
		uint8 ext[7];
		ext[0] = '!';
		ext[1] = 0xF9;	// Graphic Control Extension
		ext[2] = 4;		// data block length
		ext[3] = 1;		// disposal & input flag;
		ext[4] = ext[5] = 0;	// delay time
		ext[6] = transparentIndex;
		out->Write (ext, 7);
		
		// close extension with a zero length data block
		ext[0] = 0;
		out->Write (ext, 1);
	}

	// Write image separator and Image Descriptor
//	printf ("Write image separator and Image Descriptor\n");
	out->Write (&",", 1); 	// separator
	buf[0] = buf[1] = 0;	// left
	buf[2] = buf[3] = 0;	// top offset
	buf[4] = width%256;
	buf[5] = width/256;
	buf[6] = height%256;
	buf[7] = height/256;
	buf[8] = 0x00;			// flag byte: Not interlaced, no local color map
	buf[9] = ics;			// write initial code size byte
	out->Write (buf, 10);

	uint32 *s = (uint32 *) src->Bits() - 1;
	for (int y = 0; y < height; y++)
	{
//		printf ("[%d] ", y); fflush (stdout);
		for (int x = 0; x < width; x++)
		{
			compress_pixel (index_for (*(++s)));
		}
	}
//	printf ("\nClean up at end...\n");
	// Clean up at end
	output (EOFCode);
	if (cur_bits > 0)
		char_out (cur_accum & 0xFF);
	flush_datablock();
	
//	printf ("Mark end\n");
	// Write a zero-length block to end the series, plus a GIF terminator mark
	buf[0] = 0;
	buf[1] = ';';
	out->Write (buf, 2);

	delete src;
	return B_OK;
}

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage */*ioExtension*/, translator_info *outInfo, uint32 outType)
{
	inSource->Seek (0L, SEEK_SET);
	char header[6];
	size_t number_read = 0;
	number_read = inSource->Read (header, 6L);
	if (number_read == 6)
	{
		if (!strncmp (header, "GIF87a", 6) || !strncmp (header, "GIF89a", 6))
		// We have a GIF file in *inSource!
		{
			if (outType == GIF_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= GIF_CODE;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.4;	// GIF is 256 color max; this should do.
				outInfo->capability	= 0.6;	// Let's be modest for now.
				strcpy (outInfo->MIME, "image/gif");
				strcpy (outInfo->name, "GIF Image");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == GIF_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= B_TRANSLATOR_BITMAP;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;
				outInfo->capability	= 0.1;
				strcpy (outInfo->MIME, "image/x-be-bitmap");
				strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage */*ioExtension*/, uint32 outType, BPositionIO *outDestination)
{
	inSource->Seek (0L, SEEK_SET);
	char header[6];
	ssize_t number_read = 0;
	number_read = inSource->Read (header, 6L);
	if (number_read == 6)
	{
		if (!strncmp (header, "GIF87a", 6) || !strncmp (header, "GIF89a", 6))
		// We have a GIF file in *inSource!
		{
			if (outType == GIF_CODE)
			// "translate" it to GIF...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// decode the GIF to a bitmap
			{
				outType = B_TRANSLATOR_BITMAP;
				GIFDecoder *decoder = new GIFDecoder (inSource, outDestination);
				status_t err = decoder->Decode (outType);
				return err;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == GIF_CODE)
			// encode the bitmap to GIF...
			{
				GIFEncoder *encoder = new GIFEncoder (inSource, outDestination);
				status_t err = encoder->Encode();
				return err;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// "translate" the bitmap...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

class MySlider : public BSlider
{
public:
			 MySlider (BRect rect, const char *name, const char *label, BMessage *message, int32 min, int32 max, thumb_style ThumbStyle = B_TRIANGLE_THUMB)
			 : BSlider (rect, name, label, message, min, max, ThumbStyle)
			 {
			 	text = new char[32];
			 }
			~MySlider ()
			 {
			 	delete [] text;
			 }
virtual char *UpdateText (void) const
			 {
			 	sprintf (text, "%ld", Value());
			 	return text;
			 }
private:
char *text;
};

class GIFConfigView : public BView
{
public:
			 GIFConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
				SetViewColor (ui_color (B_PANEL_BACKGROUND_COLOR));
				thresholdS = new MySlider (BRect (8, 54, 192, 106), "thresholdS", "Transparency Threshold", new BMessage ('tsld'), 1, 255);
				g_settings_lock.Lock();
				thresholdS->SetValue (g_settings.threshold);
				g_settings_lock.Unlock();
				thresholdS->SetLimitLabels ("1", "255");
				thresholdS->SetHashMarks (B_HASH_MARKS_BOTTOM);
				thresholdS->SetHashMarkCount (25);
				AddChild (thresholdS);
				
				compressionB = new BButton (BRect (8, 112, 96, 132), "CompressionB", "Compression", new BMessage ('gifc'));
				AddChild (compressionB);
				GIFWizardB = new BButton (BRect (104, 112, 192, 132), "GIFWizardB", "To GIF Wizard", new BMessage ('gifw'));
				AddChild (GIFWizardB);
			 }

virtual		~GIFConfigView ()
			 {
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
			 	BMessenger messenger (this);
				thresholdS->SetTarget (messenger);
				compressionB->SetTarget (messenger);
				GIFWizardB->SetTarget (messenger);
			 }
virtual void Draw (BRect update)
			 {
			 	SetFont (be_bold_font);
			 	SetDrawingMode (B_OP_OVER);
			 	DrawString ("GIF Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 1998 - 2000 ∑ Sum Software", BPoint (8, 48));
			 	BView::Draw (update);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
			 	case 'tsld':
			 		g_settings_lock.Lock();
			 		g_settings.threshold = thresholdS->Value();
			 		g_settings_lock.Unlock();
			 		break;
			 		
				case 'gifw':
					system ("NetPositive http://www.gifwizard.com/pn=43051");
					break;
					
				case 'gifc':
				{
					entry_ref becasso_ref;
					if (be_roster->FindApp ("application/x-sum-Becasso", &becasso_ref) != B_OK)
					{
						(new BAlert ("Not Found", "Can't find the GIF Translator documentation\n(It comes with Becasso)", "Sorry", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT))->Go();
						break;
					}
					// printf ("Found Becasso\n");
					BEntry becassoEntry (&becasso_ref);
					BPath becassoPath (&becassoEntry);
					becassoPath.GetParent (&becassoPath);
					becassoPath.Append ("Documentation/GIF.html");
					// printf ("Doc path: %s\n", becassoPath.Path());
					entry_ref gif_doc_ref;
					if (get_ref_for_path (becassoPath.Path(), &gif_doc_ref) != B_OK)
					{
						(new BAlert ("Not Found", "Can't find the GIF Translator documentation\n(It comes with Becasso)", "Sorry", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT))->Go();
						break;
					}
					be_roster->Launch (&gif_doc_ref);
					break;
				}
					
				default:
					inherited::MessageReceived (msg);
				}
				g_settings_lock.Lock();
				g_settings.settings_touched = true;
				g_settings_lock.Unlock();
			 }
			 
private:
typedef BView inherited;
MySlider	*thresholdS;
BButton		*GIFWizardB;
BButton		*compressionB;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new GIFConfigView (BRect (0, 0, 200, 152), "GIFTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}

char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "GIF Translator";

int32 translatorVersion = 44;

translation_format inputFormats[] = {
	{ GIF_CODE, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/gif", "Graphics Interchange Format (GIF)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ GIF_CODE, B_TRANSLATOR_BITMAP, 0.4, 0.6, "image/gif", "Graphics Interchange Format (GIF)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
