#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <ctype.h>
#include <setjmp.h>
#include <Locker.h>
#include <Autolock.h>
#include <stdio.h>
#include <Path.h>
#include <FindDirectory.h>
#include <Slider.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <MenuField.h>
#include <CheckBox.h>
#include "jpeglib.h"
#include "JPEGdatamgr.h"

#define JPEG_CODE 'JPG '
// I think...

#define BUF_SIZE 4096

#define DEFAULT_QUALITY 95
#define DEFAULT_SMOOTHING 0
#define DEFAULT_COLOR_SPACE JCS_YCbCr
#define DEFAULT_DCT_METHOD JDCT_FLOAT
#define DEFAULT_PROGRESSIVE false

struct jpeg_settings
{
	BPoint  		origin;
	int				quality;	// 0 .. 100
	int				smoothing;	// 0 .. 100
	bool			progressive;
	J_COLOR_SPACE	color_space;
	J_DCT_METHOD	dct_method;
	bool			settings_touched;
};
BLocker g_settings_lock ("JPEG Settings Lock");
BLocker g_lock ("JPEG Translator Lock");
jpeg_settings g_settings;

static int status = 0;

struct my_error_mgr
{
	struct jpeg_error_mgr pub;
	jmp_buf setjmp_buffer;
};

typedef struct my_error_mgr *my_error_ptr;

METHODDEF (void)
my_error_exit (j_common_ptr cinfo)
{
	my_error_ptr myerr = (my_error_ptr) cinfo->err;
	if (status != 1)	// Otherwise it will spew errors during Identify().
	{
		(*cinfo->err->output_message) (cinfo);
	}
	longjmp (myerr->setjmp_buffer, 1);
}

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Quality [0..100] = %d\n", g_settings.quality);
	printf ("Smoothing [0..100] = %d\n", g_settings.smoothing);
	printf ("Output Color Space = %s\n", (g_settings.color_space == JCS_YCbCr) ? "Color" : "Grayscale");
	printf ("Progressive JPEG = %s\n", g_settings.progressive ? "Yes" : "No");
	printf ("DCT Method = %s\n", (g_settings.dct_method == JDCT_ISLOW) ? "Integer (Accurate)" : 
								 (g_settings.dct_method == JDCT_IFAST) ? "Integer (Fast)" : "Floating Point");
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
			 PrefsLoader ()
			 {
//			 	printf ("PrefsLoader...\n");
			 	g_settings_lock.Lock();
			 	// defaults:
				g_settings.origin = B_ORIGIN;
				g_settings.quality = DEFAULT_QUALITY;
				g_settings.smoothing = DEFAULT_SMOOTHING;
				g_settings.progressive = DEFAULT_PROGRESSIVE;
				g_settings.color_space = DEFAULT_COLOR_SPACE;
				g_settings.dct_method = DEFAULT_DCT_METHOD;
				g_settings.settings_touched = false;
				
				BPath path;
				if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
				{
					path.SetTo ("/tmp");
				}
				path.Append ("JPEGTranslator_Settings");
				FILE *fp = fopen (path.Path(), "r");
				// printf ("fopen (%s)\n", path.Path());
				if (fp)
				{
					char line[1024];
					char name[32];
					char *ptr;
					while (true)
					{
						line[0] = 0;
						fgets (line, 1024, fp);
						if (!line[0])
						{
							break;
						}
						ptr = line;
						while (isspace (*ptr))
							ptr++;
						if (*ptr == '#' || !*ptr)	// Comment line
							continue;
						if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
						{
							fprintf (stderr, "Strange JPEGTranslator settings line: %s\n", name);
						}
						else
						{
							if (!strcmp (name, "window_origin"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
									fprintf (stderr, "Illegal window origin in JPEG Translator settings: %s\n", ptr);
							}
							else if (!strcmp (name, "quality"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%d", &g_settings.quality) != 1) || (g_settings.quality < 0) || (g_settings.quality > 100))
								{
									fprintf (stderr, "Illegal quality setting in JPEG Translator settings: %s\n", ptr);
									g_settings.quality = DEFAULT_QUALITY;
								}
							}
							else if (!strcmp (name, "smoothing"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%d", &g_settings.smoothing) != 1) || (g_settings.smoothing < 0) || (g_settings.smoothing > 100))
								{
									fprintf (stderr, "Illegal smoothing setting in JPEG Translator settings: %s\n", ptr);
									g_settings.smoothing = DEFAULT_SMOOTHING;
								}
							}
							else if (!strcmp (name, "progressive"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								int tmp;
								if ((sscanf (ptr, "%d", &tmp) != 1) || (tmp != 0 && tmp != 1))
								{
									fprintf (stderr, "Illegal output type in JPEG Translator settings: %s", ptr);
									g_settings.progressive = DEFAULT_PROGRESSIVE;
								}
								else
									g_settings.progressive = (bool) tmp;
							}
							else if (!strcmp (name, "color_space"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								int tmp;
								if ((sscanf (ptr, "%d", &tmp) != 1) || (tmp != JCS_GRAYSCALE && tmp != JCS_YCbCr))
								{
									fprintf (stderr, "Illegal output color space in JPEG Translator settings: %s", ptr);
									g_settings.color_space = DEFAULT_COLOR_SPACE;
								}
								else
									g_settings.color_space = (J_COLOR_SPACE) tmp;
							}
							else if (!strcmp (name, "DCT_method"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								int tmp;
								if ((sscanf (ptr, "%d", &tmp) != 1) || (tmp != JDCT_IFAST && tmp != JDCT_ISLOW && tmp != JDCT_FLOAT))
								{
									fprintf (stderr, "Illegal DCT method color space in JPEG Translator settings: %s", ptr);
									g_settings.dct_method = DEFAULT_DCT_METHOD;
								}
								else
									g_settings.dct_method = (J_DCT_METHOD) tmp;
							}
							else
							{
								fprintf (stderr, "Unknown JPEGTranslator setting: %s\n", name);
							}
						}
					}
					fclose (fp);
				}
				g_settings_lock.Unlock();
			 }
			 
			~PrefsLoader ()
			 {
//			 	printf ("~PrefsLoader\n");
				Save();
			 }

void		 Save ()
			 {
//			 	printf ("Save\n");
			 	g_settings_lock.Lock();
			 	if (g_settings.settings_touched)
			 	{
			 		BPath path;
			 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
			 		{
			 			path.SetTo ("/tmp");
			 		}
			 		path.Append ("JPEGTranslator_Settings");
			 		FILE *fp = fopen (path.Path(), "w");
			 		if (fp)
			 		{
			 			fprintf (fp, "# JPEGTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
			 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
			 			fprintf (fp, "quality = %d\n", g_settings.quality);
			 			fprintf (fp, "smoothing = %d\n", g_settings.smoothing);
			 			fprintf (fp, "progressive = %d\n", (g_settings.progressive ? 1 : 0));
			 			fprintf (fp, "color_space = %d\n", g_settings.color_space);
			 			fprintf (fp, "DCT_method = %d\n", g_settings.dct_method);
			 			fclose (fp);
			 		}
			 	}
			 	g_settings.settings_touched = false;
			 	g_settings_lock.Unlock();
			 }
};

PrefsLoader g_prefsloader;

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage */*ioExtension*/, translator_info *outInfo, uint32 outType)
{
	BAutolock lock (&g_lock);
	inSource->Seek (0L, SEEK_SET);
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	status = 0;
	
	char header[4];
	inSource->Read (header, 4);
	if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
	// We have a Translator BBitmap in *inSource!
	{
		if (outType == JPEG_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
		{
			outInfo->type		= B_TRANSLATOR_BITMAP;
			outInfo->group		= B_TRANSLATOR_BITMAP;
			outInfo->quality	= 0.9;
			outInfo->capability	= 0.1;
			strcpy (outInfo->MIME, "image/x-be-bitmap");
			strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
			return B_OK;
		}
		else
			return B_NO_TRANSLATOR;
	}
	// Nope.  Try if it's a JPEG file then.
	cinfo.err = jpeg_std_error (&jerr.pub);
	jerr.pub.error_exit = my_error_exit;
	if (setjmp (jerr.setjmp_buffer))
	{
		jpeg_destroy_decompress (&cinfo);
		if (status == 1)
		{
			// We got here during an attempt to read the JPEG header.
			// printf ("Huu\n");
			return B_NO_TRANSLATOR;
		}
		else
		{
			// printf ("Hoo!\n");
			return B_ERROR;
		}
	}

	jpeg_create_decompress (&cinfo);
	jpeg_BPositionIO_src (&cinfo, inSource);
	status = 1;	// Reading header
	inSource->Seek (0L, SEEK_SET);
	(void) jpeg_read_header (&cinfo, true);
	
	// We have a JPEG file in *inSource!
	status = 2;
	if (outType == JPEG_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
	{
		outInfo->type		= JPEG_CODE;
		outInfo->group		= B_TRANSLATOR_BITMAP;
		outInfo->quality	= 0.3;	// JPEG is lossy; this should do.
		outInfo->capability	= 0.1;	// Let's be modest for now.
		strcpy (outInfo->MIME, "image/jpeg");
		strcpy (outInfo->name, "JPEG Image");
		// printf ("JPEG found\n");
		return B_OK;
	}
	else
	{
		// printf ("Not a JPEG\n");
		return B_NO_TRANSLATOR;
	}
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage */*ioExtension*/, uint32 outType, BPositionIO *outDestination)
{
	BAutolock lock (&g_lock);
	inSource->Seek (0L, SEEK_SET);
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	status = 0;
	
	char header[4];
	inSource->Read (header, 4);
	if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
	// We have a Translator BBitmap in *inSource!
	{
		if (outType == JPEG_CODE)
		// Translate it to JPEG (i.e. compress)
		{
			// Get info on the bitmap
			TranslatorBitmap tbitmap;
			inSource->Seek (0L, SEEK_SET);
			if (inSource->Read (&tbitmap, sizeof (TranslatorBitmap)) < sizeof (TranslatorBitmap))
			{
				fprintf (stderr, "JPEGTranslator: Couldn't read header from inStream\n");
				return B_ERROR;
			}
			swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
			swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
			swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
			swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
			swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
			
			BRect bounds = tbitmap.bounds;
			int width  = bounds.IntegerWidth() + 1;
			int height = bounds.IntegerHeight() + 1;
			int bytes_per_sample;
			int in_size;
			J_COLOR_SPACE in_color_space;
			switch (tbitmap.colors)
			{
				// TODO: Add B_CMAP8 and fill the color triplets in yourself.
				case B_RGB32:
				case B_RGBA32:
					in_size = 4;
					bytes_per_sample = 3;
					in_color_space = JCS_RGB;
					break;
				case B_GRAY8:
					in_size = 1;
					bytes_per_sample  = 1;
					in_color_space = JCS_GRAYSCALE;
					break;
				default:
					fprintf (stderr, "JPEGTranslator: Unsupported color_space %d\n", tbitmap.colors);
					return B_ERROR;
			}
			JSAMPLE *scanline = new JSAMPLE[bytes_per_sample*width];
			struct jpeg_compress_struct cinfo;
			struct jpeg_error_mgr jerr;
			JSAMPROW row_pointer[1];
			unsigned char *buffer = new unsigned char [in_size*width];
			int row_stride;
			
			cinfo.err = jpeg_std_error (&jerr);
			jpeg_create_compress (&cinfo);
			jpeg_BPositionIO_dest (&cinfo, outDestination);
			
			cinfo.image_width = width;
			cinfo.image_height = height;
			cinfo.input_components = bytes_per_sample;
			cinfo.in_color_space = in_color_space;
			
			jpeg_set_defaults (&cinfo);
			g_settings_lock.Lock();
			jpeg_set_quality (&cinfo, g_settings.quality, true);
			jpeg_set_colorspace (&cinfo, g_settings.color_space);
			if (g_settings.progressive)
				jpeg_simple_progression (&cinfo);
			cinfo.dct_method = g_settings.dct_method;
			cinfo.smoothing_factor = g_settings.smoothing;
			g_settings_lock.Unlock();
			
			jpeg_start_compress (&cinfo, true);
			row_stride = bytes_per_sample*width;
			row_pointer[0] = scanline;
			
			// Handle the scanlines one by one
			while (cinfo.next_scanline < cinfo.image_height)
			{
				// Read a scanline from inSource;
				inSource->Read (buffer, in_size*width);
				
				// Massage it into the correct format for libjpeg
				switch (tbitmap.colors)
				{
					case B_RGBA32:
					case B_RGB32:
					{
						// Munge BGRA into RGB.
						unsigned char *destptr = scanline - 1;
						unsigned char *srcptr = (unsigned char *) buffer - 1;
						for (int i = 0; i < width; i++)
						{
							uint8 b = *(++srcptr);
							uint8 g = *(++srcptr);
							uint8 r = *(++srcptr);
							++srcptr;	// Alpha isn't used...
							*(++destptr) = r;
							*(++destptr) = g;
							*(++destptr) = b;
						}
						break;
					}
					case B_GRAY8:
						// This is fine by itself.
						// The extra copy step is actually a bit of a waste of time.  Oh well.
						memcpy (scanline, buffer, width);
						break;
					default:
						fprintf (stderr, "JPEGTranslator: Unsupported color_space\n");
						return B_ERROR;
				}
				//compress it
				jpeg_write_scanlines (&cinfo, row_pointer, 1);
			}
			jpeg_finish_compress (&cinfo);
			jpeg_destroy_compress (&cinfo);
			delete [] scanline; //? 
			delete [] buffer;
			return B_NO_ERROR;
		}
		else if (outType == B_TRANSLATOR_BITMAP || !outType)
		// "Translate" it to B_TRANSLATOR_BITMAP (simple copy)
		{
			ssize_t number_read;
			inSource->Seek (0L, SEEK_SET);
			char buffer[BUF_SIZE];
			do
			{
				number_read = inSource->Read (buffer, BUF_SIZE);
				outDestination->Write (buffer, number_read);
			} while (number_read > 0);
			return B_NO_ERROR;
		}
		else
			return B_NO_TRANSLATOR;
	}
	// Nope.  Try if it's a JPEG file then.
	cinfo.err = jpeg_std_error (&jerr.pub);
	jerr.pub.error_exit = my_error_exit;
	if (setjmp (jerr.setjmp_buffer))
	{
		jpeg_destroy_decompress (&cinfo);
		if (status == 1)
		{
			// We got here during an attempt to read the JPEG header.
			// printf ("Huu\n");
			return B_NO_TRANSLATOR;
		}
		else
		{
			// printf ("Hoo!\n");
			return B_ERROR;
		}
	}

	jpeg_create_decompress (&cinfo);
	jpeg_BPositionIO_src (&cinfo, inSource);
	status = 1;	// Reading header
	inSource->Seek (0L, SEEK_SET);
	jpeg_read_header (&cinfo, true);
	
	// We have a JPEG file in *inSource!
	status = 2;
	if (outType == B_TRANSLATOR_BITMAP || !outType)
	// Translate it to a B_TRANSLATOR_BITMAP (decompress)
	{
		cinfo.out_color_space = JCS_RGB;	// We want RGB output.  JPEGs _can_ be in a different color space.
		jpeg_start_decompress (&cinfo);
		int width = cinfo.output_width;
		int height = cinfo.output_height;
		int row_stride = cinfo.output_width*cinfo.output_components;
		JSAMPARRAY buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
		uint32 *tbuffer = new uint32[width];

		TranslatorBitmap tbitmap;
		tbitmap.magic    = B_TRANSLATOR_BITMAP;
		tbitmap.bounds   = BRect (0, 0, width - 1, height - 1);
		tbitmap.rowBytes = width*4;
		tbitmap.colors   = B_RGB_32_BIT;
		tbitmap.dataSize = height*width*4;
		
		// tbitmap.bounds.PrintToStream();
		swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
		swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
		swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
		swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
		swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				
		// write out the TranslatorBitmap header
		outDestination->Write (&tbitmap, sizeof (TranslatorBitmap));
		
		// read the scanlines
		while (cinfo.output_scanline < cinfo.output_height)
		{
			// printf ("Reading scanline %d... ", cinfo.output_scanline); fflush (stdout);
			jpeg_read_scanlines (&cinfo, buffer, 1);
		
			// Expand those RGB triples into BGRA
			unsigned char *bufptr = ((unsigned char *) buffer[0]) - 1;
			for (int i = 0; i < width; i++)
			{
				uint32 pixel;
				#if defined (__POWERPC__)
					pixel = 255 | (uint32 (*(++bufptr)) << 16) | (uint32 (*(++bufptr)) << 24) | (uint32 (*(++bufptr)) << 8); // ???
				#else
					pixel = (255 << 24) | (*(++bufptr) << 16) | (*(++bufptr) << 8) | (*(++bufptr));
				#endif
				tbuffer[i] = pixel;
			}
			
			// Write the new scanline out to the TranslatorBitmap stream
			// printf ("Writing\n");
			outDestination->Write (tbuffer, width*4);
		}
		// printf ("Done.\n");
		jpeg_finish_decompress (&cinfo);
		jpeg_destroy_decompress (&cinfo);
		delete [] tbuffer;
		return B_NO_ERROR;
	}
	else if (outType == JPEG_CODE)
	// "Translate" it to JPEG (i.e. simple copy)
	{
		ssize_t number_read;
		inSource->Seek (0L, SEEK_SET);	// Horrible!  Most has already been read.
										// Anyway, this situation shouldn't occur often anyway.
		char buffer[BUF_SIZE];
		do
		{
			number_read = inSource->Read (buffer, BUF_SIZE);
			outDestination->Write (buffer, number_read);
		} while (number_read > 0);
		return B_OK;
	}
	else
		return B_NO_TRANSLATOR;
}

class MySlider : public BSlider
{
public:
			 MySlider (BRect rect, const char *name, const char *label, BMessage *message, int32 min, int32 max, thumb_style ThumbStyle = B_TRIANGLE_THUMB)
			 : BSlider (rect, name, label, message, min, max, ThumbStyle)
			 {
			 	text = new char[32];
			 }
			~MySlider ()
			 {
			 	delete [] text;
			 }
virtual char *UpdateText (void) const
			 {
			 	sprintf (text, "%ld", Value());
			 	return text;
			 }
private:
char *text;
};

class JPEGConfigView : public BView
{
public:
			 JPEGConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
//			 	printf ("JPEGConfigView ctor\n");
				SetViewColor (ui_color (B_PANEL_BACKGROUND_COLOR));
				g_settings_lock.Lock();
				qualityS = new MySlider (BRect (8, 54, 192, 106), "qualityS", "Compression", new BMessage ('qsld'), 0, 100);
				qualityS->SetValue (g_settings.quality);
				qualityS->SetLimitLabels ("Compactness", "Quality");
				qualityS->SetHashMarks (B_HASH_MARKS_BOTTOM);
				qualityS->SetHashMarkCount (10);
				AddChild (qualityS);
				smoothingS = new MySlider (BRect (8, 110, 192, 162), "smoothingS", "Smoothing", new BMessage ('ssld'), 0, 100);
				smoothingS->SetValue (g_settings.smoothing);
				smoothingS->SetLimitLabels ("None", "Full");
				smoothingS->SetHashMarks (B_HASH_MARKS_BOTTOM);
				smoothingS->SetHashMarkCount (10);
				AddChild (smoothingS);
				colorspacePU = new BPopUpMenu ("Color Space");
				colorspacePU->AddItem (new BMenuItem ("YCbCr", new BMessage ('cRGB')));
				colorspacePU->AddItem (new BMenuItem ("Greyscale", new BMessage ('cGRY')));
				colorspaceMF = new BMenuField (BRect (8, 166, 192, 186), "colorspaceMF", "Color Space", colorspacePU);
				colorspaceMF->SetDivider (70);
				if (g_settings.color_space == JCS_YCbCr)
					colorspacePU->ItemAt(0)->SetMarked (true);
				else
					colorspacePU->ItemAt(1)->SetMarked (true);
				AddChild (colorspaceMF);
				dctmethodPU = new BPopUpMenu ("DCT Method");
				dctmethodPU->AddItem (new BMenuItem ("Integer (Accurate)", new BMessage ('dcIA')));
				dctmethodPU->AddItem (new BMenuItem ("Integer (Fast)", new BMessage ('dcIF')));
				dctmethodPU->AddItem (new BMenuItem ("Floating Point", new BMessage ('dcFP')));
				dctmethodMF = new BMenuField (BRect (8, 190, 192, 210), "dctmethodMF", "DCT Method", dctmethodPU);
				dctmethodMF->SetDivider (70);
				switch (g_settings.dct_method)
				{
				case JDCT_ISLOW:
					dctmethodPU->ItemAt(0)->SetMarked (true);
					break;
				case JDCT_IFAST:
					dctmethodPU->ItemAt(1)->SetMarked (true);
					break;
				case JDCT_FLOAT:
					dctmethodPU->ItemAt(2)->SetMarked (true);
					break;
				default:
					fprintf (stderr, "JPEGTranslator: Illegal DCT Method specified\n");
				}
				AddChild (dctmethodMF);
				progressiveCB = new BCheckBox (BRect (8, 214, 192, 230), "progressiveCB", "Progressive JPEG File", new BMessage ('prog'));
				progressiveCB->SetValue (g_settings.progressive ? B_CONTROL_ON : B_CONTROL_OFF);
				AddChild (progressiveCB);
				g_settings_lock.Unlock();
			 }

virtual		~JPEGConfigView ()
			 {
//			 	printf ("JPEGConfigView dtor\n");
//			 	g_prefsloader.Save();
			 }

virtual void Draw (BRect update)
			 {
			 	SetFont (be_bold_font);
			 	SetDrawingMode (B_OP_OVER);
			 	DrawString ("JPEG Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 1998 - 2000 ∑ Sum Software", BPoint (8, 48));
			 	BView::Draw (update);
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
			 	BMessenger messenger (this);
			 	for (int i = 0; i < dctmethodPU->CountItems(); i++)
			 	{
			 		BMenuItem *item = dctmethodPU->ItemAt (i);
			 		if (item)
			 			item->SetTarget (messenger);
			 	}
				for (int i = 0; i < colorspacePU->CountItems(); i++)
			 	{
			 		BMenuItem *item = colorspacePU->ItemAt (i);
			 		if (item)
			 			item->SetTarget (messenger);
			 	}
			 	qualityS->SetTarget (messenger);
			 	smoothingS->SetTarget (messenger);
			 	progressiveCB->SetTarget (messenger);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
			 	case 'qsld':
			 		g_settings_lock.Lock();
					g_settings.quality = qualityS->Value();
			 		g_settings_lock.Unlock();
			 		break;
			 		
				case 'ssld':
					g_settings_lock.Lock();
					g_settings.smoothing = smoothingS->Value();
					g_settings_lock.Unlock();
					break;
			 		
			 	case 'cRGB':
			 		g_settings_lock.Lock();
			 		g_settings.color_space = JCS_YCbCr;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'cGRY':
			 		g_settings_lock.Lock();
			 		g_settings.color_space = JCS_GRAYSCALE;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'dcIA':
			 		g_settings_lock.Lock();
			 		g_settings.dct_method = JDCT_ISLOW;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'dcIF':
			 		g_settings_lock.Lock();
			 		g_settings.dct_method = JDCT_IFAST;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'dcFP':
			 		g_settings_lock.Lock();
			 		g_settings.dct_method = JDCT_FLOAT;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 		
			 	case 'prog':
			 		g_settings_lock.Lock();
			 		g_settings.progressive = (progressiveCB->Value() == B_CONTROL_ON);
			 		g_settings_lock.Unlock();
			 		break;
			 		
				default:
					inherited::MessageReceived (msg);
				}
				g_settings_lock.Lock();
				g_settings.settings_touched = true;
				g_settings_lock.Unlock();
			 }
			 
private:
typedef BView inherited;
BCheckBox	*progressiveCB;
BSlider		*qualityS;
BSlider		*smoothingS;
BPopUpMenu	*colorspacePU;
BMenuField	*colorspaceMF;
BPopUpMenu	*dctmethodPU;
BMenuField	*dctmethodMF;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new JPEGConfigView (BRect (0, 0, 200, 236), "JPEGTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}


char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "JPEG Translator";

int32 translatorVersion = 26;

translation_format inputFormats[] = {
	{ JPEG_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/jpeg", "JPEG format" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ JPEG_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/jpeg", "JPEG Format" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
