#include <Locker.h>
#include <Autolock.h>
#include <Path.h>
#include <FindDirectory.h>
#include <View.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <CheckBox.h>
#include <MenuField.h>
#include <StringView.h>
#include <Alert.h>
#include <stdio.h>
#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <ctype.h>
#include "png.h"

#define BUF_SIZE 4096
#define PNG_CODE 'PNG '
// I think...

#   define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)

#if defined (__POWERPC__)
#  define COLOR_MASK	0xFFFFFF00
#  define ALPHA_MASK	0x000000FF
#  define RED_MASK		0x0000FF00
#  define GREEN_MASK	0x00FF0000
#  define BLUE_MASK		0xFF000000
#  define BLUE(x)     ((x) >> 24)
#  define GREEN(x)   (((x) >> 16) & 0xFF)
#  define RED(x)     (((x) >>  8) & 0xFF)
#  define ALPHA(x)    ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(b) << 24) & 0xFF000000) | ((int(g) << 16) & 0xFF0000) | ((int(r) << 8) & 0xFF00) | (int(a) & 0xFF))
#else	// IA
#  define COLOR_MASK	0x00FFFFFF
#  define ALPHA_MASK	0xFF000000
#  define RED_MASK		0x00FF0000
#  define GREEN_MASK	0x0000FF00
#  define BLUE_MASK		0x000000FF
#  define ALPHA(x)    ((x) >> 24)
#  define RED(x)     (((x) >> 16) & 0xFF)
#  define GREEN(x)   (((x) >>  8) & 0xFF)
#  define BLUE(x)     ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(a) << 24) & 0xFF000000) | ((int(r) << 16) & 0xFF0000) | ((int(g) << 8) & 0xFF00) | (int(b) & 0xFF))
#endif

void read_data_fn (png_structp png_ptr, png_bytep data, png_uint_32 length);
void write_data_fn (png_structp png_ptr, png_bytep data, png_uint_32 length);
void output_flush_fn (png_structp png_ptr);
void info_callback (png_structp png_ptr, png_infop info);
void row_callback (png_structp png_ptr, png_bytep new_row, png_uint_32 row_num, int pass);
void end_callback (png_structp png_ptr, png_infop info);
void read_transform_fn (png_structp png_ptr, png_row_infop row_info, png_bytep data);

typedef struct
{
	BPositionIO *out;
	uint32		 rowBytes;
	png_bytep	 image;
} my_progressive_data;

struct png_settings
{
	BPoint  origin;
	int32	z_compression;// = 6; //Z_BEST_COMPRESSION;	// settable
	int32	p_interlace;//   = /*PNG_INTERLACE_ADAM7*/ PNG_INTERLACE_NONE;	// settable
	int32	p_progressive;	// Progressive decoding	// settable
	int32	warn;	// warn about broken PNGs // settable
	int32	verbose;	// output PNG info to stdout	// settable
	bool	settings_touched;
};
BLocker g_settings_lock ("PNG Settings Lock");
BLocker g_lock ("PNG Translator Lock");
png_settings g_settings;

static int32 p_compression = PNG_COMPRESSION_TYPE_DEFAULT;	// Don't change
static int32 p_filter = PNG_NO_FILTERS;
static int32 p_filter_type = PNG_FILTER_TYPE_DEFAULT;		// Don't change

#define PROGRESSIVE_BUFFER_SIZE 1024
// The amount of data read at a time.  Bigger is faster, but smaller means smoother
// operation when data is streamed in (for example in Net+).

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Compression Level [0..9] = %ld\n", g_settings.z_compression);
	printf ("Interlace [0..1] = %ld (%s)\n", g_settings.p_interlace, (g_settings.p_interlace == PNG_INTERLACE_ADAM7) ? "Adam7" : "None");
	printf ("Progressive Decoding = %s\n", g_settings.p_progressive ? "On" : "Off");
	printf ("Warn about broken PNGs = %s\n", g_settings.warn ? "Yes" : "No");
	printf ("Verbose operation = %s\n", g_settings.verbose ? "Yes" : "No");
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
			 PrefsLoader ()
			 {
			 	// printf ("PrefsLoader... "); fflush (stdout);
			 	BAutolock lock (&g_settings_lock);
			 	// defaults:
				g_settings.origin = B_ORIGIN;
				g_settings.z_compression = 6;
				g_settings.p_interlace = PNG_INTERLACE_NONE;
				g_settings.p_progressive = 0;
				g_settings.warn = 1;
				g_settings.verbose = 0;
				g_settings.settings_touched = false;
				
				BPath path;
				if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
				{
					path.SetTo ("/tmp");
				}
				path.Append ("PNGTranslator_Settings");
				FILE *fp = fopen (path.Path(), "r");
				if (fp)
				{
					char line[1024];
					char name[32];
					char *ptr;
					while (true)
					{
						line[0] = 0;
						fgets (line, 1024, fp);
						if (!line[0])
							break;
						ptr = line;
						while (isspace (*ptr))
							ptr++;
						if (*ptr == '#' || !*ptr)	// Comment line
							continue;
						if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
						{
							fprintf (stderr, "Strange PNGTranslator settings line: %s\n", name);
						}
						else
						{
							if (!strcmp (name, "window_origin"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
									fprintf (stderr, "Illegal window origin in PNG Translator settings: %s\n", ptr);
							}
							else if (!strcmp (name, "compression_level"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.z_compression) != 1) || (g_settings.z_compression < 0) || (g_settings.z_compression > 9))
								{
									fprintf (stderr, "Illegal compression level in PNG Translator settings: %s\n", ptr);
									g_settings.z_compression = 6;
								}
							}
							else if (!strcmp (name, "interlace"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.p_interlace) != 1) || (g_settings.p_interlace < 0) || (g_settings.p_interlace > 1))
								{
									fprintf (stderr, "Illegal interlace setting in PNG Translator settings: %s\n", ptr);
									g_settings.p_interlace = PNG_INTERLACE_NONE;
								}
							}
							else if (!strcmp (name, "progressive_decoding"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.p_progressive) != 1) || (g_settings.p_progressive < 0) || (g_settings.p_progressive > 1))
								{
									fprintf (stderr, "Illegal progresive decoding setting in PNG Translator settings: %s\n", ptr);
									g_settings.p_progressive = 0;
								}
							}
							else if (!strcmp (name, "verbose_decoding"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.verbose) != 1) || (g_settings.verbose < 0) || (g_settings.verbose > 1))
								{
									fprintf (stderr, "Illegal verbose decoding setting in PNG Translator settings: %s\n", ptr);
									g_settings.verbose = 0;
								}
							}
							else if (!strcmp (name, "broken_png_warning"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.warn) != 1) || (g_settings.warn < 0) || (g_settings.warn > 1))
								{
									fprintf (stderr, "Illegal broken_png_warning setting in PNG Translator settings: %s\n", ptr);
									g_settings.warn = 1;
								}
							}
							else
							{
								fprintf (stderr, "Unknown PNGTranslator setting: %s\n", name);
							}
						}
					}
					fclose (fp);
				}
				// printf ("Done.\n");
			 }
			 
			~PrefsLoader ()
			 {
			 	// printf ("~PrefsLoader... "); fflush (stdout);
			 	BAutolock lock (&g_settings_lock);
			 	if (g_settings.settings_touched)
			 	{
			 		BPath path;
			 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
			 		{
			 			path.SetTo ("/tmp");
			 		}
			 		path.Append ("PNGTranslator_Settings");
			 		FILE *fp = fopen (path.Path(), "w");
			 		if (fp)
			 		{
			 			fprintf (fp, "# PNGTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
			 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
			 			fprintf (fp, "compression_level = %ld\n", g_settings.z_compression);
			 			fprintf (fp, "interlace = %ld\n", g_settings.p_interlace);
			 			fprintf (fp, "verbose_decoding = %ld\n", g_settings.verbose);
			 			fprintf (fp, "progressive_decoding = %ld\n", g_settings.p_progressive);
			 			fprintf (fp, "broken_png_warning = %ld\n", g_settings.warn);
			 			fclose (fp);
			 		}
			 	}
			 	// printf ("Done.\n");
			 }
};

PrefsLoader g_prefsloader;

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage */*ioExtension*/, translator_info *outInfo, uint32 outType)
{
//	BAutolock lock (&g_lock);
	inSource->Seek (0L, SEEK_SET);
	unsigned char header[8];
	size_t number_read = 0;
	number_read = inSource->Read (header, 8L);
	if (number_read)
	{
		if (png_check_sig (header, number_read))
		// We have a PNG file in *inSource!
		{
			if (outType == PNG_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= PNG_CODE;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;	// PNG is lossless, this should do.
				outInfo->capability	= 0.1;	// Let's be modest for now.
				strcpy (outInfo->MIME, "image/png");
				strcpy (outInfo->name, "PNG Image");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == PNG_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= B_TRANSLATOR_BITMAP;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;
				outInfo->capability	= 0.1;
				strcpy (outInfo->MIME, "image/x-be-bitmap");
				strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage *ioExtension, uint32 outType, BPositionIO *outDestination)
{
//	BAutolock lock (&g_lock);
	inSource->Seek (0L, SEEK_SET);
	unsigned char header[8];
	ssize_t number_read = 0;
	number_read = inSource->Read (header, 8L);
	if (number_read)
	{
		if (png_check_sig (header, number_read))
		// We have a PNG file in *inSource!
		{
			if (outType == PNG_CODE)
			// "translate" it to PNG...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// decode the PNG to a bitmap
			{
				bool keep8 = false;
				int32 ext_cs;
//				if (ioExtension)
//					ioExtension->PrintToStream();
				if (ioExtension) 
				{
					ioExtension->FindInt32 (B_TRANSLATOR_EXT_BITMAP_COLOR_SPACE, &ext_cs);
					if (ext_cs == B_CMAP8)
						keep8 = true;
					// printf ("ioExt found; keep8 = %s\n", keep8 ? "true" : "false");
				}
				png_structp png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, 
					NULL, NULL, NULL);
				if (!png_ptr)
				{
					fprintf (stderr, "PNGTranslator: Problems initializing png_ptr\n");
					return B_ERROR;
				}
				png_infop info_ptr = png_create_info_struct (png_ptr);
				if (!info_ptr)
				{
					png_destroy_read_struct (&png_ptr, (png_infopp)NULL, 
						(png_infopp)NULL);
					fprintf (stderr, "PNGTranslator: Problems initializing info_ptr\n");
					return B_ERROR;
				}

				g_settings_lock.Lock();
				if (g_settings.p_progressive == 0)
				{
					g_settings_lock.Unlock();
					png_infop end_info = png_create_info_struct (png_ptr);
					if (!end_info)
					{
						png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
						fprintf (stderr, "PNGTranslator: Problems initializing end_ptr\n");
						return B_ERROR;
					}
					// Set I/O
					png_set_read_fn (png_ptr, inSource, read_data_fn);
	
					png_set_sig_bytes (png_ptr, number_read);
					png_read_info (png_ptr, info_ptr);
					png_uint_32 width, height;
					int bit_depth, color_type, interlace_type, compression_type, filter_type;
					png_get_IHDR (png_ptr, info_ptr, &width, &height, &bit_depth,
						&color_type, &interlace_type, &compression_type, &filter_type);
					
					// Set up all the transformations to massage libPNG into providing us
					// with a 32bit BGRA bitmap (so we have as little as possible left to do
					// by ourselves...
					
					g_settings_lock.Lock();
					if (g_settings.verbose)
					{
						printf ("width = %ld, height = %ld\n", width, height);
						printf ("bit_depth = %d\ncolor_type = %d\ninterlace_type = %d\ncompression_type = %d\nfiller_type = %d\n",
							bit_depth, color_type, interlace_type, compression_type, filter_type);
						if (keep8)
							printf ("8 bit color space requested\n");
					}
					g_settings_lock.Unlock();
	
					if (color_type == PNG_COLOR_TYPE_PALETTE && bit_depth <= 8 && !keep8)
					{
						png_set_palette_to_rgb (png_ptr);
						// png_set_read_user_transform_fn (png_ptr, read_transform_fn);
//						printf ("Palette set\n");
					}
					if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
						png_set_gray_1_2_4_to_8 (png_ptr);
					if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
						png_set_tRNS_to_alpha (png_ptr);

					png_set_strip_16 (png_ptr);
					png_set_packing (png_ptr);

					if (color_type == PNG_COLOR_TYPE_RGB || color_type == PNG_COLOR_TYPE_RGB_ALPHA || color_type == PNG_COLOR_TYPE_PALETTE)
						png_set_bgr (png_ptr);
	//				if (bit_depth == 8 && color_type == PNG_COLOR_TYPE_RGB)
						png_set_filler (png_ptr, 255, PNG_FILLER_AFTER);	// Full opacity if no alpha channel
					if (color_type == PNG_COLOR_TYPE_GRAY || color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
						png_set_gray_to_rgb (png_ptr);
	
	//				png_color_16 my_background;
	//				png_color_16p image_background;
	//				if (png_get_bKGD (png_ptr, info_ptr, &image_background))
	//					png_set_background (png_ptr, image_background, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
	//				else
	//					png_set_background (png_ptr, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);
	
					// TODO: Add gamma correction too!
					
	//				int number_of_passes = 1;
	//				if (interlace_type == PNG_INTERLACE_ADAM7)
	//					number_of_passes = png_set_interlace_handling (png_ptr);
	
					png_read_update_info (png_ptr, info_ptr);
					
					// Okay, allocate the bitmap...
					unsigned char *image;
					uint32 uwidth;
					if (keep8)
						uwidth = 4*int ((width + 3)/4);
					else
						uwidth = width*4;
						
					uint32 datasize = uwidth*height;
						
					TranslatorBitmap tbitmap;
					tbitmap.magic      = B_TRANSLATOR_BITMAP;
					tbitmap.bounds     = BRect (0, 0, width - 1, height - 1);
					tbitmap.rowBytes   = uwidth;
					if (keep8)
						tbitmap.colors = B_CMAP8;
					else
						tbitmap.colors = B_RGB_32_BIT;
					tbitmap.dataSize   = datasize;

//					tbitmap.bounds.PrintToStream();
					swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
					swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
					swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
					swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
					swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
					
					int32 err;
					err = outDestination->SetSize (sizeof (TranslatorBitmap) + uwidth*height);
					if (err < 0)
					{
						fprintf (stderr, "PNGTranslatore: SetSize returns %ld\n", err);
						return err;
					}

					image = new unsigned char[datasize];
					
					png_bytep *row_pointers = new png_bytep[height];
					for (uint32 i = 0; i < height; i++)
						row_pointers[i] = image + i*uwidth;
	
					// And now read the whole thing in one go...
					png_read_image (png_ptr, row_pointers);
					
					// This was commented out because it doesn't really seem
					// necessary, and the BeOS supplied PNG translator on Maui
					// incorrectly terminated PNGs (? they fail pngtest in any case;
					// "Too many IDAT's found").
					
					if (setjmp (png_jmpbuf (png_ptr)))
					{
						BAutolock lock (&g_settings_lock);
						if (g_settings.warn)
							(new BAlert ("Error", 
										 "PNGTranslator warning:\nThis PNG Image is broken (it is not terminated correctly).\nI will try to read it anyway, but even if I succeed, please re-save it.", 
										 "Proceed"))->Go();
			
						goto there;
					}
					png_read_end (png_ptr, end_info);
there:
					err = outDestination->Write (&tbitmap, sizeof (TranslatorBitmap));
					if (err < 0)
					{
						fprintf (stderr, "PNGTranslator: Write (header) returns %ld\n", err);
						delete [] row_pointers;
						delete [] image;
						return err;
					}
					err = outDestination->Write (image, datasize);
					if (err < 0)
					{
						fprintf (stderr, "PNGTranslator: Write (data) returns %ld\n", err);
						delete [] row_pointers;
						delete [] image;
						return err;
					}
	
					// There could be text info in there that we might want to use
					// in the future.  For now, just:
					png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);
					delete [] row_pointers;
					delete [] image;
				}
				else
				{
					g_settings_lock.Unlock();
					my_progressive_data data;
					data.out = outDestination;
					data.rowBytes = 0;
					data.image = NULL;
					
					png_set_progressive_read_fn (png_ptr, (void *) &data, info_callback, row_callback, end_callback);
					png_set_sig_bytes (png_ptr, number_read);
	
					png_byte buffer[PROGRESSIVE_BUFFER_SIZE];
					ssize_t len = inSource->Read (buffer, PROGRESSIVE_BUFFER_SIZE);
					while (len > 0)
					{
						png_process_data (png_ptr, info_ptr, buffer, len);
						len = inSource->Read (buffer, PROGRESSIVE_BUFFER_SIZE);
					}
					// outDestination->Write (data.image, data.rowBytes*25);
					delete [] data.image;
				}
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == PNG_CODE)
			// encode the bitmap to PNG...
			{
				png_uint_32 width, height;
				int bit_depth, color_type, interlace_type, compression_type, filter_type;
				
				// Get info on the bitmap
				TranslatorBitmap tbitmap;
				inSource->Seek (0L, SEEK_SET);
				if (inSource->Read (&tbitmap, sizeof (TranslatorBitmap)) < sizeof (TranslatorBitmap))
				{
					fprintf (stderr, "PNGTranslator: Couldn't read header from inStream\n");
					return B_ERROR;
				}
				swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				
				BRect bounds = tbitmap.bounds;
				width  = bounds.IntegerWidth() + 1;
				height = bounds.IntegerHeight() + 1;
				switch (tbitmap.colors)
				{
					case B_RGB32:
					case B_RGBA32:
						bit_depth  = 8;
						color_type = PNG_COLOR_TYPE_RGB_ALPHA;
						break;
					case B_CMAP8:
						bit_depth  = 8;
						color_type = PNG_COLOR_TYPE_PALETTE;
						break;
					case B_GRAY1:
						bit_depth  = 1;
						color_type = PNG_COLOR_TYPE_GRAY;
						break;
					default:
						fprintf (stderr, "PNGTranslator: Unsupported color_space\n");
						return B_ERROR;
				}
				compression_type = p_compression;
				g_settings_lock.Lock();
				interlace_type   = g_settings.p_interlace;
				g_settings_lock.Unlock();
				filter_type		 = p_filter_type;
				
				png_structp png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 
					NULL, NULL, NULL);
				if (!png_ptr)
				{
					fprintf (stderr, "PNGTranslator: Problems initializing png_ptr\n");
					return B_ERROR;
				}
				png_infop info_ptr = png_create_info_struct (png_ptr);
				if (!info_ptr)
				{
					png_destroy_write_struct (&png_ptr, (png_infopp)NULL);
					fprintf (stderr, "PNGTranslator: Problems initializing info_ptr\n");
					return B_ERROR;
				}
				png_set_write_fn (png_ptr, outDestination, write_data_fn, output_flush_fn);
				
//				printf ("width = %d, height = %d\n", width, height);
//				printf ("bit_depth = %d\ncolor_type = %d\ninterlace_type = %d\ncompression_type = %d\nfilter_type = %d\n",
//					bit_depth, color_type, interlace_type, compression_type, filter_type);
				png_set_filter (png_ptr, 0, p_filter);
				g_settings_lock.Lock();
				png_set_compression_level (png_ptr, g_settings.z_compression);
				g_settings_lock.Unlock();
				png_set_IHDR (png_ptr, info_ptr, width, height, bit_depth, color_type, 
					interlace_type, compression_type, filter_type);
					
				if (color_type == PNG_COLOR_TYPE_PALETTE)
				{
					const color_map *cmap = system_colors();
					const rgb_color *spalette = cmap->color_list;
					png_color palette[256];
					for (int i = 0; i < 256; i++)
					{
						palette[i].red   = spalette[i].red;
						palette[i].green = spalette[i].green;
						palette[i].blue  = spalette[i].blue;
					}
					png_set_PLTE (png_ptr, info_ptr, palette, 256);
				}
				
				// png_set_filler (png_ptr, 255, PNG_FILLER_AFTER);
				png_write_info (png_ptr, info_ptr);
				if (tbitmap.colors == B_GRAY1)
					png_set_packing (png_ptr);
				png_set_bgr (png_ptr);
				png_set_invert_mono (png_ptr);

				////////////////////////////////
				// What follows here is very inefficient.
				// We read in the entire bitmap and copy it to the PNG stream.
				// We really should support doing this row by row,
				// potentially saving a lot of memory.  Interlacing makes this kind-of
				// difficult, but we could at least supply a special case for that.

				// Okay, allocate the bitmap...
				// printf ("Allocating image: %d\n", tbitmap.dataSize);
				unsigned char *image = new unsigned char[tbitmap.dataSize];
				png_bytep *row_pointers = new png_bytep[height];
				for (uint32 i = 0; i < height; i++)
					row_pointers[i] = image + i*tbitmap.rowBytes;

				// And now read the whole thing in our bitmap...
				// printf ("Read: %d bytes\n", inSource->Read (image, tbitmap.dataSize));
				inSource->Read (image, tbitmap.dataSize);
				
				// And write it out to the PNG stream.
				png_write_image (png_ptr, row_pointers);
				png_write_end (png_ptr, info_ptr);

				png_destroy_write_struct (&png_ptr, &info_ptr);
				delete [] row_pointers;
				delete [] image;
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// "translate" the bitmap...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

class PNGConfigView : public BView
{
public:
			 PNGConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
				SetViewColor (220, 220, 220, 255);
				
			 	compressionPU = new BPopUpMenu ("Compression");
				compressionPU->AddItem (new BMenuItem ("0 (none)", new BMessage ('pPUc'), '0'));
				for (int32 i = 1; i < 9; i++)
				{
					char label[16];	// quite safe
					sprintf (label, "%ld", i);
					compressionPU->AddItem (new BMenuItem (label, new BMessage ('pPUc'), label[0]));
				}
				compressionPU->AddItem (new BMenuItem ("9 (highest)", new BMessage ('pPUc'), '9'));
			 	compressionMF = new BMenuField (BRect (8, 70, 192, 90), "compressionMF", "Compression:", compressionPU);
			 	compressionMF->SetDivider (80);
			 	g_settings_lock.Lock();
			 	compressionPU->ItemAt(g_settings.z_compression)->SetMarked (true);
				AddChild (compressionMF);
				interlacePU = new BPopUpMenu ("Interlace");
				interlacePU->AddItem (new BMenuItem ("None", new BMessage ('pPUi'), 'N'));
				interlacePU->AddItem (new BMenuItem ("Adam7", new BMessage ('pPUi'), 'A'));
				interlaceMF = new BMenuField (BRect (8, 94, 192, 114), "interlaceMF", "Interlace:", interlacePU);
				interlaceMF->SetDivider (80);
				interlacePU->ItemAt(g_settings.p_interlace)->SetMarked (true);
				AddChild (interlaceMF);
				progressiveCB = new BCheckBox (BRect (8, 118, 192, 136), "progressiveCB", "Progressive Decoding", new BMessage ('prgD'));
				progressiveCB->SetValue (g_settings.p_progressive);
				AddChild (progressiveCB);
				warnCB = new BCheckBox (BRect (8, 138, 192, 156), "warnCB", "Warn about broken PNGs", new BMessage ('warn'));
				warnCB->SetValue (g_settings.warn);
				AddChild (warnCB);
				verboseCB = new BCheckBox (BRect (8, 158, 192, 176), "verboseCB", "Verbose Decoding (stdout)", new BMessage ('verb'));
				verboseCB->SetValue (g_settings.verbose);
				AddChild (verboseCB);
				g_settings_lock.Unlock();
			 }

virtual		~PNGConfigView ()
			 {
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
			 	BMessenger messenger (this);
			 	for (int i = 0; i < compressionPU->CountItems(); i++)
			 	{
			 		BMenuItem *item = compressionPU->ItemAt (i);
			 		if (item)
			 			item->SetTarget (messenger);
			 	}
				for (int i = 0; i < interlacePU->CountItems(); i++)
			 	{
			 		BMenuItem *item = interlacePU->ItemAt (i);
			 		if (item)
			 			item->SetTarget (messenger);
			 	}
				progressiveCB->SetTarget (this);
				warnCB->SetTarget (this);
				warnCB->SetEnabled (!progressiveCB->Value());
				verboseCB->SetTarget (this);
			 }
virtual void Draw (BRect update)
			 {
			 	SetFont (be_bold_font);
				SetDrawingMode (B_OP_OVER);
			 	DrawString ("PNG Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 1998 - 2001 ∑ Sum Software", BPoint (8, 48));
			 	BView::Draw (update);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
			 	case 'pPUi':
			 		// Change this when PNG defines more than just "none" or "adam7"
			 		g_settings_lock.Lock();
			 		g_settings.p_interlace = interlacePU->IndexOf (interlacePU->FindMarked());
			 		g_settings_lock.Unlock();
			 		break;
			 		
				case 'pPUc':
					g_settings_lock.Lock();
					g_settings.z_compression = compressionPU->IndexOf (compressionPU->FindMarked());
					g_settings_lock.Unlock();
					break;
			 	
			 	case 'prgD':
					g_settings_lock.Lock();
					g_settings.p_progressive = progressiveCB->Value();
					g_settings_lock.Unlock();
					warnCB->SetEnabled (!progressiveCB->Value());
			 		break;
			 	
			 	case 'warn':
			 		g_settings_lock.Lock();
			 		g_settings.warn = warnCB->Value();
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'verb':
			 		g_settings_lock.Lock();
			 		g_settings.verbose = verboseCB->Value();
			 		g_settings_lock.Unlock();
			 		break;
			 		
				default:
					inherited::MessageReceived (msg);
				}
				g_settings_lock.Lock();
				g_settings.settings_touched = true;
				g_settings_lock.Unlock();
			 }
			 
private:
typedef BView inherited;
BPopUpMenu	*compressionPU;
BMenuField	*compressionMF;
BPopUpMenu	*interlacePU;
BMenuField	*interlaceMF;
BCheckBox	*progressiveCB;
BCheckBox	*warnCB;
BCheckBox	*verboseCB;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new PNGConfigView (BRect (0, 0, 200, 180), "PNGTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}

void read_data_fn (png_structp png_ptr, png_bytep data, png_uint_32 length)
{
	BPositionIO *stream = (BPositionIO *) png_get_io_ptr (png_ptr);
//	printf ("png_Read: %d bytes\n", stream->Read (data, length));
	stream->Read (data, length);
}

void write_data_fn (png_structp png_ptr, png_bytep data, png_uint_32 length)
{
	BPositionIO *stream = (BPositionIO *) png_get_io_ptr (png_ptr);
	//printf ("png_Write: %d bytes\n", stream->Write (data, length));
	stream->Write (data, length);
}

void output_flush_fn (png_structp png_ptr)
{
//	BPositionIO *stream = (BPositionIO *) png_get_io_ptr (png_ptr);
//	printf ("Flushing... Ooh...\n");
//	stream->Flush();
	// BPositionIO's don't have Flush() members...  
	// Let's just skip the flushing and keep our fingers crossed...
}

void info_callback (png_structp png_ptr, png_infop info_ptr)
{
	my_progressive_data *data = (my_progressive_data *) png_get_progressive_ptr (png_ptr);
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type, compression_type, filter_type;
	png_get_IHDR (png_ptr, info_ptr, &width, &height, &bit_depth,
		&color_type, &interlace_type, &compression_type, &filter_type);
	
	// Set up all the transformations to massage libPNG into providing us
	// with a 32bit BGRA bitmap (so we have as little as possible left to do
	// by ourselves...
	
	g_settings_lock.Lock();
	if (g_settings.verbose)
	{
		printf ("Progressive decoding, width = %ld, height = %ld\n", width, height);
		printf ("bit_depth = %d\ncolor_type = %d\ninterlace_type = %d\ncompression_type = %d\nfiller_type = %d\n",
			bit_depth, color_type, interlace_type, compression_type, filter_type);
	}
	g_settings_lock.Unlock();

	if (color_type == PNG_COLOR_TYPE_PALETTE && bit_depth <= 8)
	{
		png_set_palette_to_rgb (png_ptr);
		// png_set_read_user_transform_fn (png_ptr, read_transform_fn);
//		printf ("Palette set\n");
	}
	if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
		png_set_gray_1_2_4_to_8 (png_ptr);
	if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_tRNS_to_alpha (png_ptr);

	png_set_strip_16 (png_ptr);
	png_set_packing (png_ptr);

	if (color_type == PNG_COLOR_TYPE_RGB || color_type == PNG_COLOR_TYPE_RGB_ALPHA || color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_bgr (png_ptr);
//	if (bit_depth == 8 && color_type == PNG_COLOR_TYPE_RGB)
		png_set_filler (png_ptr, 255, PNG_FILLER_AFTER);	// Full opacity if no alpha channel
	if (color_type == PNG_COLOR_TYPE_GRAY || color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		png_set_gray_to_rgb (png_ptr);

//	png_color_16 my_background;
//	png_color_16p image_background;
//	if (png_get_bKGD (png_ptr, info_ptr, &image_background))
//		png_set_background (png_ptr, image_background, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
//	else
//		png_set_background (png_ptr, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);

	// TODO: Add gamma correction too!
	
	int number_of_passes = 1;
	if (interlace_type == PNG_INTERLACE_ADAM7)
		number_of_passes = png_set_interlace_handling (png_ptr);

	png_read_update_info (png_ptr, info_ptr);
	
	TranslatorBitmap tbitmap;
	tbitmap.magic    = B_TRANSLATOR_BITMAP;
	tbitmap.bounds   = BRect (0, 0, width - 1, height - 1);
	tbitmap.rowBytes = width*4;
	tbitmap.colors   = B_RGB_32_BIT;
	tbitmap.dataSize = height*width*4;
	
	data->image = new png_byte[height*width*4];
	
//	tbitmap.bounds.PrintToStream();
	swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
	swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
	
	int32 err;
	err = data->out->SetSize (sizeof (TranslatorBitmap) + width*height*4);
	if (err < 0)
	{
		fprintf (stderr, "PNGTranslatore: SetSize returns %ld\n", err);
	}
	err = data->out->Write (&tbitmap, sizeof (TranslatorBitmap));
	if (err < 0)
	{
		fprintf (stderr, "PNGTranslator: Write (header) returns %ld\n", err);
	}
	data->rowBytes = width*4;
}

void row_callback (png_structp png_ptr, png_bytep new_row, png_uint_32 row_num, int pass)
{
	my_progressive_data *data = (my_progressive_data *) png_get_progressive_ptr (png_ptr);
	png_bytep old_row = data->image + data->rowBytes*row_num;

	png_progressive_combine_row (png_ptr, old_row, new_row);

	if (new_row)
		data->out->WriteAt (sizeof (TranslatorBitmap) + row_num*data->rowBytes, old_row, data->rowBytes);
}

void end_callback (png_structp png_ptr, png_infop info)
{
}

void read_transform_fn (png_structp png_ptr, png_row_infop row_info, png_bytep data)
{
	uint32 *pixels = (uint32 *) data;
	for (uint32 i = 0; i < row_info->width; i++)
	{
		uint32 p = *pixels;
		*(pixels++) = PIXEL (BLUE (p), GREEN (p), RED (p), ALPHA (p));
	}
}

char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "PNG Translator";

int32 translatorVersion = 40;

translation_format inputFormats[] = {
	{ PNG_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/png", "Portable Network Graphics format (PNG)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ PNG_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/png", "Portable Network Graphics format (PNG)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
