#include <Locker.h>
#include <Path.h>
#include <FindDirectory.h>
#include <View.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <CheckBox.h>
#include <MenuField.h>
#include <StringView.h>
#include <Alert.h>
#include <stdio.h>
#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <ctype.h>
#include "png.h"

#define BUF_SIZE 4096
#define PSD_CODE '8BPS'

#if defined (__POWERPC__)
#  define COLOR_MASK	0xFFFFFF00
#  define ALPHA_MASK	0x000000FF
#  define RED_MASK		0x0000FF00
#  define GREEN_MASK	0x00FF0000
#  define BLUE_MASK		0xFF000000
#  define BLUE(x)     ((x) >> 24)
#  define GREEN(x)   (((x) >> 16) & 0xFF)
#  define RED(x)     (((x) >>  8) & 0xFF)
#  define ALPHA(x)    ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(b) << 24) & 0xFF000000) | ((int(g) << 16) & 0xFF0000) | ((int(r) << 8) & 0xFF00) | (int(a) & 0xFF))
#else	// IA
#  define COLOR_MASK	0x00FFFFFF
#  define ALPHA_MASK	0xFF000000
#  define RED_MASK		0x00FF0000
#  define GREEN_MASK	0x0000FF00
#  define BLUE_MASK		0x000000FF
#  define ALPHA(x)    ((x) >> 24)
#  define RED(x)     (((x) >> 16) & 0xFF)
#  define GREEN(x)   (((x) >>  8) & 0xFF)
#  define BLUE(x)     ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(a) << 24) & 0xFF000000) | ((int(r) << 16) & 0xFF0000) | ((int(g) << 8) & 0xFF00) | (int(b) & 0xFF))
#endif

enum psd_mode 
{
	mBitmap			= 0,
	mGrayscale,
	mIndexed,
	mRGB,
	mCMYK,
	mMultichannel	= 7,
	mDuotone,
	mLab
};

enum compression_mode
{
	mRawData	= 0,
	mRLE
};

struct psd_settings
{
	BPoint  origin;
	uint32	verbose;
	bool	settings_touched;
};
BLocker g_settings_lock ("PSD Settings Lock");
psd_settings g_settings;

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Verbose operation = %s\n", g_settings.verbose ? "yes" : "no");
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
			 PrefsLoader ()
			 {
			 	// printf ("PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	// defaults:
				g_settings.origin = B_ORIGIN;
				g_settings.settings_touched = false;
				
				BPath path;
				if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
				{
					path.SetTo ("/tmp");
				}
				path.Append ("PSDTranslator_Settings");
				FILE *fp = fopen (path.Path(), "r");
				if (fp)
				{
					char line[1024];
					char name[32];
					char *ptr;
					while (true)
					{
						line[0] = 0;
						fgets (line, 1024, fp);
						if (!line[0])
							break;
						ptr = line;
						while (isspace (*ptr))
							ptr++;
						if (*ptr == '#' || !*ptr)	// Comment line
							continue;
						if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
						{
							fprintf (stderr, "Strange PSDTranslator settings line: %s\n", name);
						}
						else
						{
							if (!strcmp (name, "window_origin"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
									fprintf (stderr, "Illegal window origin in PSD Translator settings: %s\n", ptr);
							}
							else if (!strcmp (name, "verbose_decoding"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.verbose) != 1) || (g_settings.verbose < 0) || (g_settings.verbose > 1))
								{
									fprintf (stderr, "Illegal verbose decoding setting in PNG Translator settings: %s\n", ptr);
									g_settings.verbose = 0;
								}
							}
							else
							{
								fprintf (stderr, "Unknown PSDTranslator setting: %s\n", name);
							}
						}
					}
					fclose (fp);
				}
				g_settings_lock.Unlock();
				// printf ("Done.\n");
			 }
			 
			~PrefsLoader ()
			 {
			 	// printf ("~PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	if (g_settings.settings_touched)
			 	{
			 		BPath path;
			 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
			 		{
			 			path.SetTo ("/tmp");
			 		}
			 		path.Append ("PSDTranslator_Settings");
			 		FILE *fp = fopen (path.Path(), "w");
			 		if (fp)
			 		{
			 			fprintf (fp, "# PNGTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
			 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
			 			fprintf (fp, "verbose_decoding = %ld\n", g_settings.verbose ? 1L : 0L);
			 			fclose (fp);
			 		}
			 	}
			 	g_settings_lock.Unlock();
			 	// printf ("Done.\n");
			 }
};

PrefsLoader g_prefsloader;

bool psd_check_sig (const unsigned char *header)
{
	return (header[0] == '8' && header[1] == 'B' && header[2] == 'P' && header[3] == 'S'
			&& header[4] == 0 && header[5] == 1);
}

status_t rle_decompress (const unsigned char *in, unsigned char *out, size_t insize, int32 outwidth)
{
	off_t offset_in = 0;
	off_t offset_out = 0;
	while (offset_in <= insize && offset_out < outwidth)
	{
		unsigned char byte = in[offset_in++];
		if (byte > 128)
		{
			int count = 257 - byte;
			unsigned char val = in[offset_in++];
			for (int i = 0; i < count; i++)
		 		out[offset_out++] = val;
		}
		else
		{
			int count = byte + 1;
			for (int i = 0; i < count; i++)
				out[offset_out++] = in[offset_in++];
		}
	}
	return B_OK;
}

status_t de_interleave_rgb (unsigned char **data, unsigned char *out, uint32 rows, uint32 columns, uint16 channels)
{
	// data is [planes*rows][columns]
	for (uint32 j = 0; j < rows; j++)
	{
		unsigned char *rout = out + j*columns*4;
		for (uint32 i = 0; i < columns; i++)
		{
			rout[i*4 + 2] = data[j][i];	// R
			rout[i*4 + 1] = data[rows + j][i];	// G
			rout[i*4] = data[2*rows + j][i];	// B
			if (channels == 3)	// no alpha channel present
				rout[i*4 + 3] = 255;
			else if (channels == 4)
				rout[i*4 + 3] = data[3*rows + j][i];
			else
				return B_ERROR;
		}
	}
	return B_OK;
}

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage */*ioExtension*/, translator_info *outInfo, uint32 outType)
{
	inSource->Seek (0L, SEEK_SET);
	unsigned char header[6];
	size_t number_read = 0;
	number_read = inSource->Read (header, 6L);
	if (number_read)
	{
		if (psd_check_sig (header))
		// We have a PSD 2.5 file in *inSource!
		{
			if (outType == PSD_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= PSD_CODE;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;	// PSD is lossless, this should do.
				outInfo->capability	= 0.1;	// Let's be modest for now.
				strcpy (outInfo->MIME, "image/x-photoshop");
				strcpy (outInfo->name, "Photoshop Image");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == PSD_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= B_TRANSLATOR_BITMAP;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;
				outInfo->capability	= 0.1;
				strcpy (outInfo->MIME, "image/x-be-bitmap");
				strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage *ioExtension, uint32 outType, BPositionIO *outDestination)
{
	inSource->Seek (0L, SEEK_SET);
	unsigned char header[6];
	ssize_t number_read = 0;
	bool keep8 = false;
	number_read = inSource->Read (header, 6L);
	if (number_read)
	{
		if (psd_check_sig (header))
		// We have a PSD file in *inSource!
		{
			if (outType == PSD_CODE)
			// "translate" it to PSD...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// decode the PSD to a bitmap
			{
				int32 ext_cs;
				if (ioExtension) 
				{
					ioExtension->FindInt32 (B_TRANSLATOR_EXT_BITMAP_COLOR_SPACE, &ext_cs);
					if (ext_cs == B_CMAP8)
						keep8 = true;
					// printf ("ioExt found; keep8 = %s\n", keep8 ? "true" : "false");
				}
				uint16 channels;
				uint32 rows, columns;
				uint16 depth;
				uint16 mode;
				uint32 mode_data;

				char roh[24];
				if (inSource->Read (roh, 24L) != 24)
				{
					fprintf (stderr, "PSD Translator: Problems reading header\n");
					return B_ERROR;
				}
				
				// skip 6 "reserved" bytes
				channels  = ntohs (*((uint16 *) (roh + 6)));
				rows      = ntohl (*((uint32 *) (roh + 8)));
				columns   = ntohl (*((uint32 *) (roh + 12)));
				depth     = ntohs (*((uint16 *) (roh + 16)));
				mode      = ntohs (*((uint16 *) (roh + 18)));
				mode_data = ntohl (*((uint32 *) (roh + 20)));

				char *mode_data_buffer = new char [mode_data];
				if (inSource->Read (mode_data_buffer, mode_data) != mode_data)
				{
					fprintf (stderr, "PSD Translator: Problems reading mode data\n");
					delete [] mode_data_buffer;
					return B_ERROR;
				}
				
				uint32 imres;
				if (inSource->Read (roh, 4L) != 4)
				{
					fprintf (stderr, "PSD Translator: Problems reading header (image resources data)\n");
					delete [] mode_data_buffer;
					return B_ERROR;
				}
				imres = ntohl (*((uint32 *) roh));
				
				// Skip image resources block
				char *imres_buffer = new char [imres];
				if (inSource->Read (imres_buffer, imres) != imres)
				{
					fprintf (stderr, "PSD Translator: Problems skipping image resources\n");
					delete [] mode_data_buffer;
					delete [] imres_buffer;
					return B_ERROR;
				}
				delete [] imres_buffer;
				
				uint32 reserved;
				if (inSource->Read (roh, 4L) != 4)
				{
					fprintf (stderr, "PSD Translator: Problems reading reserved block\n");
					return B_ERROR;
				}
				reserved = ntohl (*((uint32 *) roh));
				
				// Skip the reserved data
				char *resdata_buffer = new char [reserved];
				if (inSource->Read (resdata_buffer, reserved) != reserved)
				{
					fprintf (stderr, "PSD Translator: Problems skipping reserved data\n");
					delete [] mode_data_buffer;
					delete [] resdata_buffer;
					return B_ERROR;
				}
				delete [] resdata_buffer;
				
				uint16 compression;
				if (inSource->Read (roh, 2L) != 2)
				{
					fprintf (stderr, "PSD Translator: Problems reading compression block\n");
					return B_ERROR;
				}
				compression = ntohs (*((uint16 *) roh));
				

				g_settings_lock.Lock();
				if (g_settings.verbose)
				{
					printf ("width = %ld, height = %ld\n", columns, rows);
					printf ("channels = %d, depth = %d, mode = %d, compression = %d\n", channels, depth, mode, compression);
					if (keep8)
						printf ("8 bit color space requested\n");
				}
				g_settings_lock.Unlock();
	
				uint32 i, j, k;
				uint16 *scan_counts = 0;
				if (compression == mRLE)
				{
					scan_counts = new uint16 [rows*channels];
					if (inSource->Read (scan_counts, rows*channels*2) != rows*channels*2)
					{
						fprintf (stderr, "PSD Translator: Problems reading byte counts\n");
						delete [] mode_data_buffer;
						delete [] scan_counts;
						return B_ERROR;
					}
					for (i = 0; i < rows*channels; i++)
						swap_data (B_INT16_TYPE, &scan_counts[i], sizeof (uint16), B_SWAP_BENDIAN_TO_HOST);
				}
				
				// prepare scanlines
				unsigned char **scanlines = new unsigned char * [rows*channels];
				unsigned char **decom = new unsigned char * [rows*channels];
				for (i = 0; i < rows*channels; i++)
				{
					if (compression == mRLE)
					{
						scanlines[i] = new unsigned char [scan_counts[i]];
						decom[i] = new unsigned char [columns];
						if (inSource->Read (scanlines[i], scan_counts[i]) != scan_counts[i])
						{
							fprintf (stderr, "PSD Translator: Problems reading scan line data %ld\n", i);
							return B_ERROR;
						}
					}
					else
					{
						scanlines[i] = new unsigned char [columns];
						decom[i] = 0;
						if (inSource->Read (scanlines[i], columns) != columns)
						{
							fprintf (stderr, "PSD Translator: Problems reading scan line data %ld\n", i);
							return B_ERROR;
						}
					}
				}

				// Okay, allocate the bitmap...
				unsigned char *image;
				uint32 uwidth;
				if (keep8)
					uwidth = 4*int ((columns + 3)/4);
				else
					uwidth = columns*4;
					
				image = new unsigned char[uwidth*rows];
				
				// de-interleave (and decompress if necessary)
				switch (mode)
				{
				case mBitmap:
				case mGrayscale:
				case mCMYK:
				case mMultichannel:
				case mDuotone:
				case mLab:
					fprintf (stderr, "PSDTranslator: Unsupported color space in PSD file\n");
					delete [] image;
					return B_ERROR;
					break;
				case mRGB:
				{
					if (compression == mRLE)
					{
//						for (j = 220; j < 22; j++)
//						{
//							printf ("[%ld]:", j);
//							for (int i = 0; i < scan_counts[j]; i++)
//								printf ("%d ", scanlines[j][i]);
//							printf ("\n");
//						}
	
						for (k = 0; k < channels; k++)
							for (j = 0; j < rows; j++)
								rle_decompress (scanlines[k*rows + j], decom[k*rows + j], scan_counts[k*rows + j], columns);
						de_interleave_rgb (decom, image, rows, columns, channels);
					}
					else
					{
						de_interleave_rgb (scanlines, image, rows, columns, channels);
					}
					break;
				}
				}

				TranslatorBitmap tbitmap;
				tbitmap.magic      = B_TRANSLATOR_BITMAP;
				tbitmap.bounds     = BRect (0, 0, columns - 1, rows - 1);
				tbitmap.rowBytes   = uwidth;
				if (keep8)
					tbitmap.colors = B_CMAP8;
				else
					tbitmap.colors = B_RGB_32_BIT;
				tbitmap.dataSize   = rows*uwidth;
				
//					tbitmap.bounds.PrintToStream();
				swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				
				// OK, we have the image in a bitmap now, let's write it out!
				int32 err = outDestination->SetSize (sizeof (TranslatorBitmap) + uwidth*rows);
				if (err < 0)
				{
					fprintf (stderr, "PSDTranslatore: SetSize returns %ld\n", err);
					return err;
				}
				err = outDestination->Write (&tbitmap, sizeof (TranslatorBitmap));
				if (err < 0)
				{
					fprintf (stderr, "PSDTranslator: Write (header) returns %ld\n", err);
					return err;
				}
				err = outDestination->Write (image, uwidth*rows);
				if (err < 0)
				{
					fprintf (stderr, "PSDTranslator: Write (data) returns %ld\n", err);
					return err;
				}

				for (i = 0; i < rows*channels; i++)
					delete [] decom[i];
				delete [] decom;
				for (i = 0; i < rows*channels; i++)
					delete [] scanlines[i];
				delete [] scanlines;
				delete [] scan_counts;
				delete [] mode_data_buffer;
				delete [] image;
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == PSD_CODE)
			// encode the bitmap to PSD...
			{
				// Get info on the bitmap
				TranslatorBitmap tbitmap;
				inSource->Seek (0L, SEEK_SET);
				if (inSource->Read (&tbitmap, sizeof (TranslatorBitmap)) < sizeof (TranslatorBitmap))
				{
					fprintf (stderr, "PSDTranslator: Couldn't read header from inStream\n");
					return B_ERROR;
				}
				swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				
				uint32 rows, columns;
				
				BRect bounds = tbitmap.bounds;
				columns = bounds.IntegerWidth() + 1;
				rows    = bounds.IntegerHeight() + 1;
				uint16 bit_depth = 0;
				switch (tbitmap.colors)
				{
					case B_RGB32:
					case B_RGBA32:
						bit_depth  = 8;
						break;
//					case B_CMAP8:
//						bit_depth  = 8;
//						break;
//					case B_GRAY1:
//						bit_depth  = 1;
//						break;
					default:
						fprintf (stderr, "PSDTranslator: Unsupported color_space\n");
						return B_ERROR;
				}

				// Construct and write out the PSD header
				char header[40];
				
				header[0] = '8';
				header[1] = 'B';
				header[2] = 'P';
				header[3] = 'S';													// Signature	
				*((uint16 *) (header + 4)) = B_HOST_TO_BENDIAN_INT16 (1);			// Version
				header[6] = header[7] = header[8] = header[9] = header[10] = header[11] = 0;	// Reserved
				*((uint16 *) (header + 12)) = B_HOST_TO_BENDIAN_INT16 (4);			// Channels
				*((uint32 *) (header + 14)) = B_HOST_TO_BENDIAN_INT32 (rows);
				*((uint32 *) (header + 18)) = B_HOST_TO_BENDIAN_INT32 (columns);
				*((uint16 *) (header + 22)) = B_HOST_TO_BENDIAN_INT16 (8);			// Depth
				*((uint16 *) (header + 24)) = B_HOST_TO_BENDIAN_INT16 (mRGB);		// Mode
				*((uint32 *) (header + 26)) = B_HOST_TO_BENDIAN_INT32 (0);			// Mode Data
				*((uint32 *) (header + 30)) = B_HOST_TO_BENDIAN_INT32 (0);			// Image Resources
				*((uint32 *) (header + 34)) = B_HOST_TO_BENDIAN_INT32 (0);			// Reserved Data
				*((uint32 *) (header + 38)) = B_HOST_TO_BENDIAN_INT16 (mRawData);	// Compression

				if (outDestination->Write (header, 40L) != 40L)
				{
					fprintf (stderr, "PSD Translator: Couldn't write out the header\n");
					return B_ERROR;
				}

				// Okay, allocate the bitmap...
				uint32 lsize = tbitmap.dataSize/4;
				printf ("Allocating image: %ld\n", tbitmap.dataSize);
				uint32 *image = new uint32 [lsize];

				// And now read the whole thing in our bitmap...
				// printf ("Read: %d bytes\n", inSource->Read (image, tbitmap.dataSize));
				if (inSource->Read (image, tbitmap.dataSize) != tbitmap.dataSize)
				{
					fprintf (stderr, "PSD Translator: Couldn't read the image data\n");
					delete [] image;
					return B_ERROR;
				}
				
				// And write it out to the PSD stream.  Planar data, no padding.
				// at the expense of memory, make use of plane buffers (for speed)
				uint32 *ip = image;
				char *plane = new char [lsize];
				char *pp = plane;
				for (uint32 i = 0; i < lsize; i++)
				{
					uint32 pixel = *ip++;
					*pp++ = RED(pixel);
				}
				if (outDestination->Write (plane, lsize) != lsize)
				{
					fprintf (stderr, "PSD Translator: Error writing bitmap data (red plane)\n");
					delete [] image;
					return B_ERROR;
				}
				ip = image;
				pp = plane;
				for (uint32 i = 0; i < lsize; i++)
				{
					uint32 pixel = *ip++;
					*pp++ = GREEN(pixel);
				}
				if (outDestination->Write (plane, lsize) != lsize)
				{
					fprintf (stderr, "PSD Translator: Error writing bitmap data (green plane)\n");
					delete [] image;
					return B_ERROR;
				}
				ip = image;
				pp = plane;
				for (uint32 i = 0; i < lsize; i++)
				{
					uint32 pixel = *ip++;
					*pp++ = BLUE(pixel);
				}
				if (outDestination->Write (plane, lsize) != lsize)
				{
					fprintf (stderr, "PSD Translator: Error writing bitmap data (blue plane)\n");
					delete [] image;
					return B_ERROR;
				}
				ip = image;
				pp = plane;
				for (uint32 i = 0; i < lsize; i++)
				{
					uint32 pixel = *ip++;
					*pp++ = ALPHA(pixel);
				}
				if (outDestination->Write (plane, lsize) != lsize)
				{
					fprintf (stderr, "PSD Translator: Error writing bitmap data (alpha plane)\n");
					delete [] image;
					return B_ERROR;
				}
				
				delete [] image;
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// "translate" the bitmap...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

class PSDConfigView : public BView
{
public:
			 PSDConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
				SetViewColor (220, 220, 220, 255);
				verboseCB = new BCheckBox (BRect (8, 158, 192, 176), "verboseCB", "Verbose Decoding (stdout)", new BMessage ('verb'));
				verboseCB->SetValue (g_settings.verbose);
				AddChild (verboseCB);
				g_settings_lock.Unlock();
			 }

virtual		~PSDConfigView ()
			 {
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
			 	BMessenger messenger (this);
				verboseCB->SetTarget (this);
			 }
virtual void Draw (BRect update)
			 {
			 	SetDrawingMode (B_OP_OVER);
			 	SetFont (be_bold_font);
			 	DrawString ("PSD Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 2000 ∑ Sum Software", BPoint (8, 48));
			 	BView::Draw (update);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
			 	case 'verb':
			 		g_settings_lock.Lock();
			 		g_settings.verbose = verboseCB->Value();
			 		g_settings_lock.Unlock();
			 		break;
			 		
				default:
					inherited::MessageReceived (msg);
				}
				g_settings_lock.Lock();
				g_settings.settings_touched = true;
				g_settings_lock.Unlock();
			 }
			 
private:
typedef BView inherited;
BCheckBox	*verboseCB;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new PSDConfigView (BRect (0, 0, 200, 180), "PSDTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}
char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "PSD Translator";

int32 translatorVersion = 1;

translation_format inputFormats[] = {
	{ PSD_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-photoshop", "Adobe Photoshop Image format (PSD)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ PSD_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-photoshop", "Adobe Photoshop Image format (PSD)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
