#include <Locker.h>
#include <Autolock.h>
#include <Path.h>
#include <FindDirectory.h>
#include <View.h>
#include <CheckBox.h>
#include <StringView.h>
#include <PopUpMenu.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <Slider.h>
#include <Alert.h>
#include <stdio.h>
#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <ctype.h>
#include "tiff.h"
#include "tiffio.h"

#define BUF_SIZE 4096
#define TIFF_CODE 'TIFF'

#if defined (__POWERPC__)
#  define COLOR_MASK	0xFFFFFF00
#  define ALPHA_MASK	0x000000FF
#  define RED_MASK		0x0000FF00
#  define GREEN_MASK	0x00FF0000
#  define BLUE_MASK		0xFF000000
#  define BLUE(x)     ((x) >> 24)
#  define GREEN(x)   (((x) >> 16) & 0xFF)
#  define RED(x)     (((x) >>  8) & 0xFF)
#  define ALPHA(x)    ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(b) << 24) & 0xFF000000) | ((int(g) << 16) & 0xFF0000) | ((int(r) << 8) & 0xFF00) | (int(a) & 0xFF))
#else	// IA
#  define COLOR_MASK	0x00FFFFFF
#  define ALPHA_MASK	0xFF000000
#  define RED_MASK		0x00FF0000
#  define GREEN_MASK	0x0000FF00
#  define BLUE_MASK		0x000000FF
#  define ALPHA(x)    ((x) >> 24)
#  define RED(x)     (((x) >> 16) & 0xFF)
#  define GREEN(x)   (((x) >>  8) & 0xFF)
#  define BLUE(x)     ((x)        & 0xFF)
#  define PIXEL(r,g,b,a) (((int(a) << 24) & 0xFF000000) | ((int(r) << 16) & 0xFF0000) | ((int(g) << 8) & 0xFF00) | (int(b) & 0xFF))
#endif

struct tiff_settings
{
	BPoint  origin;
	int32	verbose;
	int32	compression;
	int32	jpeg_quality;
	int32	jpeg_colormode;
	int32	photometric;
	int32	xres;
	int32	yres;
	bool	settings_touched;
};
BLocker g_settings_lock ("TIFF Settings Lock");
tiff_settings g_settings;

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Verbose decoding = %ld\n", g_settings.verbose);
	printf ("Compression = %ld\n", g_settings.compression);
	printf ("JPEG Quality = %ld\n", g_settings.jpeg_quality);
	printf ("JPEG Color Mode = %ld\n", g_settings.jpeg_colormode);
	printf ("Photometric = %ld\n", g_settings.photometric);
	printf ("Resolution = (%ld x %ld)\n", g_settings.xres, g_settings.yres);
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
			 PrefsLoader ()
			 {
			 	// printf ("PrefsLoader... "); fflush (stdout);
			 	BAutolock lock (&g_settings_lock);
			 	// defaults:
				g_settings.origin = B_ORIGIN;
				g_settings.verbose = 0;
				g_settings.compression = COMPRESSION_PACKBITS;
				g_settings.jpeg_quality = 75;
				g_settings.jpeg_colormode = JPEGCOLORMODE_RGB;
				g_settings.photometric = PHOTOMETRIC_RGB;
				g_settings.xres = 72;
				g_settings.yres = 72;
				g_settings.settings_touched = false;
				
				BPath path;
				if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
				{
					path.SetTo ("/tmp");
				}
				path.Append ("TIFFTranslator_Settings");
				FILE *fp = fopen (path.Path(), "r");
				if (fp)
				{
					char line[1024];
					char name[32];
					char *ptr;
					while (true)
					{
						line[0] = 0;
						fgets (line, 1024, fp);
						if (!line[0])
							break;
						ptr = line;
						while (isspace (*ptr))
							ptr++;
						if (*ptr == '#' || !*ptr)	// Comment line
							continue;
						if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
						{
							fprintf (stderr, "Strange TIFFTranslator settings line: %s\n", name);
						}
						else
						{
							if (!strcmp (name, "window_origin"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
									fprintf (stderr, "Illegal window origin in TIFF Translator settings: %s\n", ptr);
							}
							else if (!strcmp (name, "verbose_decoding"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.verbose) != 1) || (g_settings.verbose < 0) || (g_settings.verbose > 1))
								{
									fprintf (stderr, "Illegal verbose decoding setting in TIFF Translator settings: %s\n", ptr);
									g_settings.verbose = 0;
								}
							}
							else if (!strcmp (name, "compression"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.compression) != 1) || 
								   ((g_settings.compression != COMPRESSION_NONE) && (g_settings.compression != COMPRESSION_PACKBITS) && (g_settings.compression != COMPRESSION_JPEG) && (g_settings.compression != COMPRESSION_DEFLATE)))
								{
									fprintf (stderr, "Invalid compression setting in TIFF Translator settings: %s\n", ptr);
									g_settings.compression = COMPRESSION_NONE;
								}
							}
							else if (!strcmp (name, "jpeg_quality"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.jpeg_quality) != 1) || 
								   ((g_settings.jpeg_quality < 0) || (g_settings.jpeg_quality > 100)))
								{
									fprintf (stderr, "Invalid jpeg_quality setting in TIFF Translator settings: %s\n", ptr);
									g_settings.jpeg_quality = 75;
								}
							}
							else if (!strcmp (name, "x_resolution"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.xres) != 1) || 
								   (g_settings.xres <= 0))
								{
									fprintf (stderr, "Invalid x_resolution setting in TIFF Translator settings: %s\n", ptr);
									g_settings.xres = 72;
								}
							}
							else if (!strcmp (name, "y_resolution"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.yres) != 1) || 
								   (g_settings.yres <= 0))
								{
									fprintf (stderr, "Invalid y_resolution setting in TIFF Translator settings: %s\n", ptr);
									g_settings.yres = 72;
								}
							}
							else
							{
								fprintf (stderr, "Unknown TIFFTranslator setting: %s\n", name);
							}
						}
					}
					fclose (fp);
				}
				// printf ("Done.\n");
			 }
			 
			~PrefsLoader ()
			 {
			 	// printf ("~PrefsLoader... "); fflush (stdout);
			 	BAutolock lock (&g_settings_lock);
			 	if (g_settings.settings_touched)
			 	{
			 		BPath path;
			 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
			 		{
			 			path.SetTo ("/tmp");
			 		}
			 		path.Append ("TIFFTranslator_Settings");
			 		FILE *fp = fopen (path.Path(), "w");
			 		if (fp)
			 		{
			 			fprintf (fp, "# PNGTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
			 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
			 			fprintf (fp, "verbose_decoding = %ld\n", g_settings.verbose);
			 			fprintf (fp, "compression = %ld\n", g_settings.compression);
			 			fclose (fp);
			 		}
			 	}
			 	// printf ("Done.\n");
			 }
};

PrefsLoader g_prefsloader;

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage */*ioExtension*/, translator_info *outInfo, uint32 outType)
{
//	BAutolock lock (&g_lock);
	inSource->Seek (0L, SEEK_SET);
	unsigned char header[4];
	size_t number_read = 0;
	number_read = inSource->Read (header, 4L);
	if (number_read)
	{
		uint16 magic = *((uint16 *)header);
		if ((magic == TIFF_BIGENDIAN && header[3] == 42)	// 42 = "version" (but not really)
		 || (magic == TIFF_LITTLEENDIAN && header[2] == 42))
		// We have a TIFF file in *inSource!
		{
			if (outType == TIFF_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= TIFF_CODE;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;	// TIFF can be lossless, this should do.
				outInfo->capability	= 0.1;	// Let's be modest for now.
				strcpy (outInfo->MIME, "image/tiff");
				strcpy (outInfo->name, "TIFF Image");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == TIFF_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= B_TRANSLATOR_BITMAP;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;
				outInfo->capability	= 0.1;
				strcpy (outInfo->MIME, "image/x-be-bitmap");
				strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

tsize_t StreamRead (thandle_t handle, tdata_t data, tsize_t size)
{
	BPositionIO *stream = (BPositionIO *) handle;
	return stream->Read (data, size);
}

tsize_t StreamWrite (thandle_t handle, tdata_t data, tsize_t size)
{
	BPositionIO *stream = (BPositionIO *) handle;
	return stream->Write (data, size);
}

toff_t StreamSeek (thandle_t handle, toff_t off, int whence)
{
	BPositionIO *stream = (BPositionIO *) handle;
	return stream->Seek (off, whence);
}

int StreamClose (thandle_t handle)
{
	return B_OK;	// don't close the stream ourselves
}

toff_t StreamSize (thandle_t handle)
{
	BPositionIO *stream = (BPositionIO *) handle;
	off_t cur = stream->Position();
	off_t size = stream->Seek (0, SEEK_END);
	stream->Seek (cur, SEEK_SET);
	return size;
}

int StreamMap (thandle_t, tdata_t *, toff_t *)
{
	return 0;
}

void StreamUnmap (thandle_t, tdata_t, toff_t)
{
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage *ioExtension, uint32 outType, BPositionIO *outDestination)
{
//	BAutolock lock (&g_lock);
	g_settings_lock.Lock();
	int32 verbose = g_settings.verbose;
	g_settings_lock.Unlock();
	inSource->Seek (0L, SEEK_SET);
	unsigned char header[4];
	size_t number_read = 0;
	number_read = inSource->Read (header, 4L);
	if (number_read)
	{
		uint16 magic = *((uint16 *)header);
		if ((magic == TIFF_BIGENDIAN && header[3] == 42)	// 42 = "version" (but not really)
		 || (magic == TIFF_LITTLEENDIAN && header[2] == 42))
		// We have a TIFF file in *inSource!
		{
			if (outType == TIFF_CODE)
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// decode the TIFF to a bitmap
			{
				inSource->Seek (0L, SEEK_SET);
				TIFF *tif = TIFFClientOpen ("TIFFTranslator", "r", (thandle_t) inSource, StreamRead, StreamWrite, StreamSeek, StreamClose, StreamSize, StreamMap, StreamUnmap);
				if (!tif)
				{
					fprintf (stderr, "TIFFTranslator: Couldn't open stream as TIFF\n");
					return B_NO_TRANSLATOR;
				}
				uint32 width, height;
				uint32 xres, yres, unit;
				TIFFGetField (tif, TIFFTAG_IMAGEWIDTH, &width);
				TIFFGetField (tif, TIFFTAG_IMAGELENGTH, &height);
				TIFFGetField (tif, TIFFTAG_XRESOLUTION, &xres);
				TIFFGetField (tif, TIFFTAG_YRESOLUTION, &yres);
				TIFFGetField (tif, TIFFTAG_RESOLUTIONUNIT, &unit);
				if (unit == RESUNIT_CENTIMETER)
				{
					xres = int (xres*2.54 + 0.5);
					yres = int (yres*2.54 + 0.5);
				}
//				if (xres <= 0) xres = 72;
//				if (yres <= 0) yres = 72;
				
				uint32 *image;
				uint32 uwidth = width*4;
					
				uint32 datasize = uwidth*height;
				if (verbose)
				{
					printf ("Found a %ld x %ld image\n", width, height);
					printf ("Resolution: %ld x %ld dpi\n", xres, yres);
				}
					
				TranslatorBitmap tbitmap;
				tbitmap.magic      = B_TRANSLATOR_BITMAP;
				tbitmap.bounds     = BRect (0, 0, width - 1, height - 1);
				tbitmap.rowBytes   = uwidth;
				tbitmap.colors	   = B_RGB_32_BIT;
				tbitmap.dataSize   = datasize;

//				tbitmap.bounds.PrintToStream();
				swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				
				int32 err;
				err = outDestination->SetSize (sizeof (TranslatorBitmap) + uwidth*height);
				if (err < 0)
				{
					fprintf (stderr, "TIFFTranslatore: SetSize returns %ld\n", err);
					return err;
				}

				image = new uint32[width*height];
				
				// And now read the whole thing in one go...
				if (!TIFFReadRGBAImage (tif, width, height, image, 0))
				{
					fprintf (stderr, "TIFFTranslator: Couldn't read RGBA image\n");
					delete [] image;
					return err;
				}
				// The image is upside down now, and not in the correct BGRA order.
				// Switch blue and red channels here, flip image during final output.
				for (uint32 i = 0; i < width*height; i++)
				{
					uint32 pixel = image[i];
					image[i] = PIXEL (BLUE(pixel), GREEN(pixel), RED(pixel), ALPHA(pixel));
				}

				err = outDestination->Write (&tbitmap, sizeof (TranslatorBitmap));
				if (err < 0)
				{
					fprintf (stderr, "TIFFTranslator: Write (header) returns %ld\n", err);
					delete [] image;
					return err;
				}
				for (int32 y = height - 1; y >= 0; y--)
				{
					err = outDestination->Write (image + y*width, uwidth);
					if (err < 0)
					{
						fprintf (stderr, "TIFFTranslator: Write (data) returns %ld\n", err);
						delete [] image;
						return err;
					}
				}

				delete [] image;
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == TIFF_CODE)
			// encode the bitmap to TIFF
			{
				// Open the TIFF stream for writing
				TIFF *out = TIFFClientOpen ("TIFFTranslator", "w", (thandle_t) outDestination, StreamRead, StreamWrite, StreamSeek, StreamClose, StreamSize, StreamMap, StreamUnmap);
				if (!out)
				{
					fprintf (stderr, "TIFFTranslator: Couldn't open stream for writing\n");
					return B_NO_TRANSLATOR;
				}

				// Get info on the bitmap
				TranslatorBitmap tbitmap;
				inSource->Seek (0L, SEEK_SET);
				if (inSource->Read (&tbitmap, sizeof (TranslatorBitmap)) < sizeof (TranslatorBitmap))
				{
					fprintf (stderr, "TIFFTranslator: Couldn't read header from inStream\n");
					return B_ERROR;
				}
				swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				
				BRect bounds = tbitmap.bounds;
				uint32 width  = bounds.IntegerWidth() + 1;
				uint32 height = bounds.IntegerHeight() + 1;
				switch (tbitmap.colors)
				{
					case B_RGB32:
					case B_RGBA32:
						break;
					default:
						fprintf (stderr, "TIFFTranslator: Unsupported color_space\n");
						return B_ERROR;
				}
				
				g_settings_lock.Lock();
				// And write it out to the TIFF file.
				TIFFSetField(out, TIFFTAG_IMAGEWIDTH,  width);
				TIFFSetField(out, TIFFTAG_IMAGELENGTH, height);
				TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
				TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 4);
				TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
				TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
				TIFFSetField(out, TIFFTAG_PHOTOMETRIC, g_settings.photometric);
				TIFFSetField(out, TIFFTAG_COMPRESSION, g_settings.compression);
				switch (g_settings.compression) {
				case COMPRESSION_JPEG:
					TIFFSetField(out, TIFFTAG_JPEGQUALITY, g_settings.jpeg_quality);
					TIFFSetField(out, TIFFTAG_JPEGCOLORMODE, g_settings.jpeg_colormode);
					break;
				case COMPRESSION_LZW:
				case COMPRESSION_DEFLATE:
//					if (predictor != 0)
//						TIFFSetField(out, TIFFTAG_PREDICTOR, predictor);
					break;
				}
				int32 linebytes = 4*width;
				unsigned char *buf;
				if (TIFFScanlineSize(out) > linebytes)
					buf = (unsigned char *)_TIFFmalloc(linebytes);
				else
					buf = (unsigned char *)_TIFFmalloc(TIFFScanlineSize(out));
				TIFFSetField(out, TIFFTAG_ROWSPERSTRIP,
				    TIFFDefaultStripSize(out, height));
				TIFFSetField(out, TIFFTAG_XRESOLUTION, g_settings.xres);
				TIFFSetField(out, TIFFTAG_YRESOLUTION, g_settings.yres);
				TIFFSetField(out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
				g_settings_lock.Unlock();

				for (int32 row = 0; row < height; row++) {
					if (inSource->Read (buf, linebytes) != linebytes)
					{
						fprintf (stderr, "TIFFTranslator: Couldn't read row %ld\n", row);
						break;
					}
					// massage pixels in right format: swap red and blue channels
					for (int32 i = 0; i < width; i++)
					{
						uint32 pixel = ((uint32 *) buf)[i];
						((uint32 *) buf)[i] = PIXEL (BLUE(pixel), GREEN(pixel), RED(pixel), ALPHA(pixel));
					}
					if (TIFFWriteScanline(out, buf, row, 0) < 0)
						break;
				}
				TIFFClose(out);
				if (buf) _TIFFfree (buf);

				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// just copy the bitmap
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

class MySlider : public BSlider
{
public:
			 MySlider (BRect rect, const char *name, const char *label, BMessage *message, int32 min, int32 max, thumb_style ThumbStyle = B_TRIANGLE_THUMB)
			 : BSlider (rect, name, label, message, min, max, ThumbStyle)
			 {
			 	text = new char[32];
			 }
			~MySlider ()
			 {
			 	delete [] text;
			 }
virtual char *UpdateText (void) const
			 {
			 	sprintf (text, "%ld", Value());
			 	return text;
			 }
private:
char *text;
};

class TIFFConfigView : public BView
{
public:
			 TIFFConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
				SetViewColor (ui_color (B_PANEL_BACKGROUND_COLOR));
				
				g_settings_lock.Lock();
				compressionPU = new BPopUpMenu ("Compression");
				compressionPU->AddItem (new BMenuItem ("None", new BMessage ('pCno')));
				compressionPU->AddItem (new BMenuItem ("PackBits", new BMessage ('pCpb')));
				compressionPU->AddItem (new BMenuItem ("JPEG", new BMessage ('pCjp')));
				compressionPU->AddItem (new BMenuItem ("Deflate", new BMessage ('pCdf')));
				qualityS = new MySlider (BRect (8, 94, 192, 146), "qualityS", "JPEG Compression", new BMessage ('qsld'), 0, 100);
				qualityS->SetValue (g_settings.jpeg_quality);
				qualityS->SetLimitLabels ("Compactness", "Quality");
				qualityS->SetHashMarks (B_HASH_MARKS_BOTTOM);
				qualityS->SetHashMarkCount (10);
				switch (g_settings.compression)
				{
					case COMPRESSION_NONE:
						compressionPU->ItemAt(0)->SetMarked (true);
						qualityS->SetEnabled (false);
						break;
					case COMPRESSION_PACKBITS:
						compressionPU->ItemAt(1)->SetMarked (true);
						qualityS->SetEnabled (false);
						break;
					case COMPRESSION_JPEG:
						compressionPU->ItemAt(2)->SetMarked (true);
						qualityS->SetEnabled (true);
						break;
					case COMPRESSION_DEFLATE:
						compressionPU->ItemAt(3)->SetMarked (true);
						qualityS->SetEnabled (false);
						break;
					default:
						break;
				}
				compressionMF = new BMenuField (BRect (8, 70, 192, 90), "compressionMF", "Compression", compressionPU);
				AddChild (compressionMF);
				AddChild (qualityS);
				verboseCB = new BCheckBox (BRect (8, 150, 192, 168), "verboseCB", "Verbose Decoding (stdout)", new BMessage ('verb'));
				verboseCB->SetValue (g_settings.verbose);
				AddChild (verboseCB);
				g_settings_lock.Unlock();
			 }

virtual		~TIFFConfigView ()
			 {
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
				verboseCB->SetTarget (this);
			 	BMessenger messenger (this);
			 	for (int i = 0; i < compressionPU->CountItems(); i++)
			 	{
			 		BMenuItem *item = compressionPU->ItemAt (i);
			 		if (item)
			 			item->SetTarget (messenger);
			 	}
			 	qualityS->SetTarget (messenger);
			 }
virtual void Draw (BRect update)
			 {
			 	SetFont (be_bold_font);
				SetDrawingMode (B_OP_OVER);
			 	DrawString ("TIFF Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 2001 ∑ Sum Software", BPoint (8, 48));
			 	BView::Draw (update);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
			 	case 'verb':
			 		g_settings_lock.Lock();
			 		g_settings.verbose = verboseCB->Value();
					g_settings.settings_touched = true;
			 		g_settings_lock.Unlock();
			 		break;
			 		
			 	case 'qsld':
			 		g_settings_lock.Lock();
					g_settings.jpeg_quality = qualityS->Value();
					g_settings.settings_touched = true;
			 		g_settings_lock.Unlock();
			 		break;

				case 'pCno':
					g_settings_lock.Lock();
					g_settings.compression = COMPRESSION_NONE;
					g_settings.settings_touched = true;
					g_settings_lock.Unlock();
					qualityS->SetEnabled (false);
					break;

				case 'pCpb':
					g_settings_lock.Lock();
					g_settings.compression = COMPRESSION_PACKBITS;
					g_settings.settings_touched = true;
					g_settings_lock.Unlock();
					qualityS->SetEnabled (false);
					break;

				case 'pCjp':
					g_settings_lock.Lock();
					g_settings.compression = COMPRESSION_JPEG;
					g_settings.settings_touched = true;
					g_settings_lock.Unlock();
					qualityS->SetEnabled (true);
					break;

				case 'pCdf':
					g_settings_lock.Lock();
					g_settings.compression = COMPRESSION_DEFLATE;
					g_settings.settings_touched = true;
					g_settings_lock.Unlock();
					qualityS->SetEnabled (false);
					break;

				default:
					inherited::MessageReceived (msg);
				}
			 }
			 
private:
typedef BView inherited;
BCheckBox	*verboseCB;
BPopUpMenu	*compressionPU;
BMenuField	*compressionMF;
MySlider	*qualityS;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new TIFFConfigView (BRect (0, 0, 200, 172), "TIFFTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}

char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "TIFF Translator";

int32 translatorVersion = 10;

translation_format inputFormats[] = {
	{ TIFF_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/tiff", "Tagged Image File Format (TIFF)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ TIFF_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/tiff", "Tagged Image File Format (TIFF)" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
